/****************************************************************************
**  LinkTest                                                               **
*****************************************************************************
**  Copyright (c) 2008-2022                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/
#include "linktest.h"
#include "config.h"
#include "compiler.h"
#include "cmdline.h"
#include "error.h"
#include "vcluster.h"
#include "format_print.h"
#include <cstdlib>
#include <cstdio>
#include <string>
#include <vector>
#include <stdexcept>
#include <cassert>
#include <cinttypes>

#include <map>
#include "vector_bool.h"

static void print_cmdline_usage(const std::string& prog);

static struct linktest_args cmdline_args;

struct Argument{
    enum ArgumentType{
        kInteger = 1,
        kString  = 2,
        kBoolean = 3,
        kFloat   = 4,
        kCustom  = 5
    };

    ArgumentType    type;
    int             nargs;
    std::string     optshort;
    std::string     optlong;
    std::string     defaultval;
    bool            required;
    std::string     help;

    void*           p;

    std::string     opt() const;
};

auto linktest_all_cmdline_argdefs = {
    Argument{
        .type       = Argument::kBoolean,
        .nargs      = 0,
        .optshort   = "h",
        .optlong    = "help",
        .defaultval = "",
        .required   = false,
        .help       = "Print this help message and exit",
        .p          = nullptr
    },
    Argument{
        .type       = Argument::kBoolean,
        .nargs      = 0,
        .optshort   = "v",
        .optlong    = "version",
        .defaultval = "",
        .required   = false,
        .help       = "Print LinkTest version and exit",
        .p          = nullptr
    },
    Argument{
        .type       = Argument::kString,
        .nargs      = 1,
        .optshort   = "m",
        .optlong    = "mode",
        .defaultval = "",
        .required   = false,
        .help       = "Transport Layer to be used [REQUIRED]*",
        .p          = &cmdline_args.virtual_cluster_implementation
    },
    Argument{
        .type       = Argument::kInteger,
        .nargs      = 1,
        .optshort   = "w",
        .optlong    = "num-warmup-messages",
        .defaultval = "",
        .required   = true,
        .help       = "Number of warm-up messages [REQUIRED]",
        .p          = &cmdline_args.num_warmup_msg
    },
    Argument{
        .type       = Argument::kInteger,
        .nargs      = 1,
        .optshort   = "n",
        .optlong    = "num-messages",
        .defaultval = "",
        .required   = true,
        .help       = "Number of messages [REQUIRED]",
        .p          = &cmdline_args.num_msg
    },
    Argument{
        .type       = Argument::kInteger,
        .nargs      = 1,
        .optshort   = "s",
        .optlong    = "size-messages",
        .defaultval = "",
        .required   = true,
        .help       = "Message size in bytes [REQUIRED]",
        .p          = &cmdline_args.len_msg
    },
    Argument{
        .type       = Argument::kString,
        .nargs      = 1,
        .optshort   = "o",
        .optlong    = "output",
        .defaultval = "pingpong_results_bin.sion",
        .required   = false,
        .help       = "output file name",
        .p          = &cmdline_args.output
    },
    Argument{
        .type       = Argument::kBoolean,
        .nargs      = 0,
        .optshort   = "",
        .optlong    = "no-sion-file",
        .defaultval = "0",
        .required   = false,
        .help       = "Do not write data to sion file",
        .p          = &cmdline_args.do_nosion
    },
    Argument{
        .type       = Argument::kBoolean,
        .nargs      = 0,
        .optshort   = "",
        .optlong    = "parallel-sion-file",
        .defaultval = "0",
        .required   = false,
        .help       = "Write data SION file in parallel",
        .p          = &cmdline_args.do_sion_par
    },
    Argument{
        .type       = Argument::kInteger,
        .nargs      = 1,
        .optshort   = "",
        .optlong    = "num-slowest",
        .defaultval = "10",
        .required   = false,
        .help       = "Number of slowest pairs to be retested",
        .p          = &cmdline_args.max_stest
    },
    Argument{
        .type       = Argument::kInteger,
        .nargs      = 1,
        .optshort   = "",
        .optlong    = "min-iterations",
        .defaultval = "1",
        .required   = false,
        .help       = "LinkTest repeats for at least <min_iterations>",
        .p          = &cmdline_args.min_iterations
    },
    Argument{
         .type       = Argument::kFloat,
         .nargs      = 1,
         .optshort   = "",
         .optlong    = "min-runtime",
         .defaultval = "0.0",
         .required   = false,
         .help       = "LinkTest runs for at least <min_runtime> seconds communication time",
         .p          = &cmdline_args.min_walltime
    },
    Argument{
        .type       = Argument::kString,
        .nargs      = 1,
        .optshort   = "",
        .optlong    = "memory-buffer-allocator",
        .defaultval = "DEFAULT",
        .required   = false,
        .help       = "Allocator type for memory buffers",
        .p          = &cmdline_args.memory_buffer_allocator
    },
    Argument{
        .type       = Argument::kBoolean,
        .nargs      = 0,
        .optshort   = "",
        .optlong    = "all-to-all",
        .defaultval = "0",
        .required   = false,
        .help       = "Additionally perform MPI all-to-all tests",
        .p          = &cmdline_args.do_alltoall
    },
    Argument{
        .type       = Argument::kBoolean,
        .nargs      = 0,
        .optshort   = "",
        .optlong    = "unidirectional",
        .defaultval = "0",
        .required   = false,
        .help       = "Perform unidirectional tests",
        .p          = &cmdline_args.do_unidir
    },
    Argument{
        .type       = Argument::kBoolean,
        .nargs      = 0,
        .optshort   = "",
        .optlong    = "bidirectional",
        .defaultval = "0",
        .required   = false,
        .help       = "Perform bidirectional tests",
        .p          = &cmdline_args.do_bidir
    },
    Argument{
        .type       = Argument::kBoolean,
        .nargs      = 0,
        .optshort   = "",
        .optlong    = "bisection",
        .defaultval = "0",
        .required   = false,
        .help       = "Test between bisecting halves",
        .p          = &cmdline_args.do_bisection
    },
    Argument{
        .type       = Argument::kBoolean,
        .nargs      = 0,
        .optshort   = "",
        .optlong    = "serial-tests",
        .defaultval = "0",
        .required   = false,
        .help       = "Serialize tests",
        .p          = &cmdline_args.do_serial
    },
    Argument{
        .type       = Argument::kBoolean,
        .nargs      = 0,
        .optshort   = "",
        .optlong    = "randomize-steps",
        .defaultval = "0",
        .required   = false,
        .help       = "Randomize step execution order",
        .p          = &cmdline_args.do_mix
    },
    Argument{
        .type       = Argument::kInteger,
        .nargs      = 1,
        .optshort   = "",
        .optlong    = "seed-randomize-steps",
        .defaultval = "9876543210",
        .required   = false,
        .help       = "Seed for step randomization",
        .p          = &cmdline_args.seed_randomize_steps
    },
    Argument{
        .type       = Argument::kBoolean,
        .nargs      = 0,
        .optshort   = "",
        .optlong    = "use-gpu-memory",
        .defaultval = "0",
        .required   = false,
        .help       = "Use GPU memory to store message buffers",
        .p          = &cmdline_args.do_use_gpus
    },
    Argument{
        .type       = Argument::kCustom,
        .nargs      = 1,
        .optshort   = "",
        .optlong    = "use-gpu-memory-on-task",
        .defaultval = "[]",
        .required   = false,
        .help       = "Use GPU memory only on task ids marked with 1",
        .p          = &cmdline_args.use_gpus_on_task
    },
    Argument{
        .type       = Argument::kBoolean,
        .nargs      = 0,
        .optshort   = "",
        .optlong    = "multi-buffer",
        .defaultval = "0",
        .required   = false,
        .help       = "Use multiple send and receive buffers",
        .p          = &cmdline_args.use_multi_buf
    },
    Argument{
        .type       = Argument::kInteger,
        .nargs      = 1,
        .optshort   = "",
        .optlong    = "num-multi-buffer",
        .defaultval = "0",
        .required   = false,
        .help       = "Number of buffers when using multiple buffers",
        .p          = &cmdline_args.num_multi_buf
    },
    Argument{
        .type       = Argument::kBoolean,
        .nargs      = 0,
        .optshort   = "",
        .optlong    = "randomize-buffers",
        .defaultval = "0",
        .required   = false,
        .help       = "Randomize buffers",
        .p          = &cmdline_args.randomize_buffers
    },
    Argument{
        .type       = Argument::kInteger,
        .nargs      = 1,
        .optshort   = "",
        .optlong    = "seed-randomize-buffers",
        .defaultval = "1234567890",
        .required   = false,
        .help       = "Seed for buffer randomization",
        .p          = &cmdline_args.buf_mt_seed
    },
    Argument{
        .type       = Argument::kBoolean,
        .nargs      = 0,
        .optshort   = "",
        .optlong    = "check-buffers",
        .defaultval = "0",
        .required   = false,
        .help       = "Check buffers after timing kernel",
        .p          = &cmdline_args.check_buffers
    },
    Argument{
        .type       = Argument::kInteger,
        .nargs      = 1,
        .optshort   = "",
        .optlong    = "num-randomize-tasks",
        .defaultval = "0",
        .required   = false,
        .help       = "Use num different randomly assigned task IDs for communication scheme",
        .p          = &cmdline_args.num_randomize_tasks
    },
    Argument{
        .type       = Argument::kInteger,
        .nargs      = 1,
        .optshort   = "",
        .optlong    = "seed-randomize-tasks",
        .defaultval = "29309775",
        .required   = false,
        .help       = "Seed for task randomization",
        .p          = &cmdline_args.seed_randomize_tasks
    },
    Argument{
        .type       = Argument::kBoolean,
        .nargs      = 0,
        .optshort   = "",
        .optlong    = "group-processes-by-hostname",
        .defaultval = "0",
        .required   = false,
        .help       = "Group processes by hostnames and ensure that all processes from one group communicate with all processes from another, i.e. the communication pattern is applied to groups",
        .p          = &cmdline_args.do_group_processes_by_hostname
    }
};

int atouint32_t(const char* const string, uint32_t* const ptr){
    unsigned int i;

    if(isdigit(string[0])==0) return 1;
    const size_t stringLen=strlen(string);
    if(stringLen>10) return 1;
    for(i=1;i<stringLen;i++) if(isdigit(string[i])==0) return 1;
    if(stringLen==10){
        uint32_t a;
        sscanf(&string[0],"%1" SCNu32,&a);
        if(a>4) return 1;
        if(a==4){
            sscanf(&string[1],"%" SCNu32 ,&a);
            if(a>294967295) return 1;
            goto a;
        }
        goto a;
    }
a:  i=sscanf(string,"%" SCNu32,ptr);
    if(i!=1) return 1;
    return 0;
}

int atouint64_t(const char* const string, std::uint64_t* const ptr){
    unsigned int i;

    if(isdigit(string[0])==0) return 1;
    const size_t stringLen=strlen(string);
    if(stringLen>20) return 1;
    for(i=1;i<stringLen;i++) if(isdigit(string[i])==0) return 1;
    if(stringLen==20){
        std::uint64_t a;
        sscanf(&string[0],"%1" SCNu64,&a);
        if(a>1) return 1;
        if(a==1){
            sscanf(&string[1],"%" SCNu64,&a);
            if(a>8446744073709551615) return 1;
            goto b;
        }
        goto b;
    }
b:  i=sscanf(string,"%" SCNu64,ptr);
    if(i!=1) return 1;
    return 0;
}

std::string Argument::opt() const{
    std::string s = "";

    if (optshort.length() > 0)
        s += ("-" + optshort);
    if ((optshort.length() > 0) and (optlong.length() > 0))
        s += "/";
    if (optlong.length() > 0)
        s += ("--" + optlong);

    return s;
}

/* Post-processing of command line arguments */

const struct linktest_args* fixup_cmdline_args(const std::uint64_t size){
    // The maximum number of serial retests of slow connections depends on the type of test that is exectued.
    if(cmdline_args.do_bisection){if(cmdline_args.max_stest>size*size/2)   cmdline_args.max_stest=size*size/2  ;}
    else{                         if(cmdline_args.max_stest>size*(size-1)) cmdline_args.max_stest=size*(size-1);}

    if(cmdline_args.min_iterations > 1 || cmdline_args.min_walltime > 0.0) {
            cmdline_args.do_nosion = 1;
    }

    if(cmdline_args.virtual_cluster_implementation.size() > linktest_args::VClusterImplmaxCharacters) {
            warn("mode parameter value longer than 16 characters, cannot be saved to sionfile correctly");
    }
    cmdline_args.virtual_cluster_implementation.resize(linktest_args::VClusterImplmaxCharacters); // important for writing to sionfile

    if(cmdline_args.use_multi_buf){
        if(!cmdline_args.do_unidir){
            error("Use of multiple receive buffers only supported for uni-directional test.");
        }
        if(!cmdline_args.num_multi_buf){
            cmdline_args.num_multi_buf=cmdline_args.num_msg;
        }else if(cmdline_args.num_multi_buf==1){
            warn("Only one multiple buffer requested. Turning of use of multiple buffers for better performance.");
            cmdline_args.use_multi_buf=0;
            cmdline_args.num_multi_buf=0;
        }else{
            auto M=((cmdline_args.num_msg>cmdline_args.num_warmup_msg)?cmdline_args.num_msg:cmdline_args.num_warmup_msg);
            if(cmdline_args.num_multi_buf>M){
                warn("Number of multiple buffers exceeds number of messages to be sent, reducing number of multiple buffers to number of messages.");
                cmdline_args.num_multi_buf=M;
            }
        }
    }

    //Zero Mersenne Twister seeds cause the PRNGs to generate zeros as output
    if(cmdline_args.randomize_buffers){
        if(cmdline_args.buf_mt_seed==0){
            warn("Zero Mersenne Twister seed will result in the buffers being filled with zeros.");
        }
    }else{
        cmdline_args.buf_mt_seed=0;
    }

    //Warn users of statistically correlated permutations when too many tasks are used
    if(cmdline_args.num_randomize_tasks>1){
        if(size>624){
            warn("The used Mersenne Twister PRNG is only suitable for generating task permutations up to 622 tasks. For 624 tasks and above permutations are statistically correlated.");
        }
    }

     return(&cmdline_args);
}

static bool special_cmdline_args(const std::string& prog, const std::vector<std::string>& args){
    for(auto& arg: args) {
        if (arg == "-h" or arg == "--help") {
            print_cmdline_usage(prog);
            return true;
        }
        if (arg == "-v" or arg == "--version") {
            print_linktest_version();
            return true;
        }
    }

    return false;
}

static bool arg_match(const std::string& arg, const Argument& argdef){
    if ((argdef.optshort.length() > 0) and (arg == "-" + argdef.optshort)) {
        return true;
    }
    if ((argdef.optlong.length() > 0)  and (arg == "--" + argdef.optlong)) {
        return true;
    }

    return false;
}

static int report_unknown_cmdline_args(const std::vector<std::string>& args){
    for (auto& arg: args) {
        if ('-' != arg[0]) {
            continue;
        }

        bool match = false;
        for(auto& argdef: linktest_all_cmdline_argdefs) {
            if (arg_match(arg, argdef)) {
                match = true;
                break;
            }
        }

        if (not match) {
            error("Unknown argument %s skipped", arg.c_str());
        }
    }

    return SUCCESS;
}

// Report a missing required argument via `error()` message.
static void fail_missing_cmdline_arg(const Argument& argdef){
    std::string err = "Required argument " + argdef.opt() + " missing";
    error(err.c_str());
}

/* Find a command line argument that matches the `argdef` and return a const iterator to
 * the location. We return a const iterator to enable subsequent logic to access command
 * line arguments following the match.
 */
std::vector<std::string>::const_iterator find_cmdline_arg_match(const std::vector<std::string>& args, const Argument& argdef){
    auto match = args.end();
    for(auto arg = args.begin(); arg != args.end(); ++arg) {
        if (arg_match((*arg), argdef)) {
            match = arg;
        }
    }

        return match;
}

template<typename T>
void assign_argument_value(const Argument& argdef, const T& val)
{
    if (nullptr == argdef.p) {
        /* Avoid false positive `error()` messages for the special arguments that may show
         * up here due to the surrounding logic.
         */
        if (("h" != argdef.optshort) and ("v" != argdef.optshort)) {
            error("Trying to assign value for invalid argument %s", argdef.opt().c_str());
        }
    } else {
            *reinterpret_cast<T*>(argdef.p) = val;
    }
}

static int handle_cmdline_arg_match(const std::string* val, const Argument& argdef){
    switch (argdef.nargs) {
        case 0:
            switch (argdef.type) {
                case Argument::kBoolean:
                    assign_argument_value<std::int8_t>(argdef, 1);
                    #ifdef DEBUG_CMDLINE
                        printf("DEBUG: %s->%" PRId8 "\n",argdef.optlong.c_str(),*(std::int8_t*)argdef.p);
                    #endif
                    break;
                default:
                    throw std::logic_error("Unsupported comand line argument: " + argdef.optlong + " " + *val);
            }
            break;
        case 1:
            if (nullptr == val) {
                error("Argument %s requires a value", argdef.opt().c_str());
                return -1;
            }
            switch (argdef.type) {
                case Argument::kInteger:
                    std::uint64_t tmp;
                    if(atouint64_t(val->c_str(),&tmp)!=0) throw std::logic_error("Failed to parse kInteger!");
                    assign_argument_value<std::uint64_t>(argdef, tmp);
                    #ifdef DEBUG_CMDLINE
                        printf("DEBUG: %s->%" PRIu64 "\n",argdef.optlong.c_str(),*(std::uint64_t*)argdef.p);
                    #endif
                    break;
                case Argument::kString:
                    assign_argument_value<std::string>(argdef, *val);
                    #ifdef DEBUG_CMDLINE
                        printf("DEBUG: %s->%s\n",argdef.optlong.c_str(),(*(std::string*)argdef.p).c_str());
                    #endif
                    break;
                case Argument::kFloat:
                    assign_argument_value<double>(argdef, atof(val->c_str()));
                    #ifdef DEBUG_CMDLINE
                        printf("DEBUG: %s->%f\n",argdef.optlong.c_str(),*(double*)argdef.p);
                    #endif
                    break;
                case Argument::kCustom:
                    if (argdef.optlong == "use-gpu-memory-on-task") {
                        *reinterpret_cast<std::vector<bool>*>(argdef.p) = VectorBool::fromString(*val);
                        #ifdef DEBUG_CMDLINE
                        printf("DEBUG: %s->%f\n",argdef.optlong.c_str(),*(double*)argdef.p);
                        #endif
                        break;
                    }
                default:
                    throw std::logic_error("Unsupported comand line argument: " + argdef.optlong + " " + *val);
            }
            break;
        default:
            throw std::logic_error("Support for more than one values for an argument is not implemented");
        }

        return SUCCESS;
}

static int handle_cmdline_arg_match(const std::vector<std::string>& args,
                                    std::vector<std::string>::const_iterator match,
                                    const Argument& argdef)
{
        const std::string* val = nullptr;
        if ((match + 1) != args.end())
                val = &(*(match + 1));

        return handle_cmdline_arg_match(val, argdef);
}

static int handle_cmdline_arg_default(const std::string& val,
                                      const Argument& argdef)
{
        /* boolean default values need to be treated because presence of the argument
         * on the command line means true whereas default values can be either true or
         * false.
         */
        if (Argument::kBoolean == argdef.type) {
                assign_argument_value<std::int8_t>(argdef, atoi(argdef.defaultval.c_str()));
                return SUCCESS;
        } else {
                return handle_cmdline_arg_match(&val, argdef);
        }
}


const struct linktest_args* parse_cmdline_args(int argc, char **argv){
    std::vector<std::string> args;
    for (auto i = 1; argv[i]; ++i) {
        args.push_back(argv[i]);
    }

    cmdline_args.special_args=special_cmdline_args(argv[0], args);
    if(cmdline_args.special_args) {
        return &cmdline_args;
    } 

    for(auto& argdef: linktest_all_cmdline_argdefs) {
        auto match = find_cmdline_arg_match(args, argdef);
        if (match != args.end()) {
            auto err = handle_cmdline_arg_match(args, match, argdef);
            if (unlikely(err))
                return nullptr;
        } else {
            if (argdef.required) {
                fail_missing_cmdline_arg(argdef);
                return nullptr;
            }

            auto err = handle_cmdline_arg_default(argdef.defaultval, argdef);
            if (unlikely(err))
                return nullptr;
        }
    }

    #if HAVE_VCLUSTER_CUDA != 1
    if(cmdline_args.do_use_gpus || cmdline_args.virtual_cluster_implementation=="cuda"){
        fatal("Requested using GPUs, but compiled without CUDA support.");
    }
    #endif

    report_unknown_cmdline_args(args);

    if(cmdline_args.do_bidir&&cmdline_args.do_unidir){
        fatal("Cannot perform a bidirectional test at the same time as a uni-directional test.");
    }

    return &cmdline_args;
}

const struct linktest_args *bcast_cmdline_args(VirtualCluster* cl,
                                               int root, const struct
                                               linktest_args* args)
{
        struct linktest_args *tmp;

        if (root == cl->rank()) {
                /* bcast will not modify the arguments so this cast should
                 * be fine.
                 */
                tmp = (struct linktest_args* )args;
        } else {
                tmp = &cmdline_args;
        }

        /* This code is designed to deal with different padding on different
         * processing elements but not with different endianess.
         */

        cl->bcast(root,&tmp->do_alltoall                   ,1);
        cl->bcast(root,&tmp->do_bidir                      ,1);
        cl->bcast(root,&tmp->do_unidir                     ,1);
        cl->bcast(root,&tmp->do_mix                        ,1);
        cl->bcast(root,&tmp->do_serial                     ,1);
        cl->bcast(root,&tmp->do_nosion                     ,1);
        cl->bcast(root,&tmp->do_sion_par                   ,1);
        cl->bcast(root,&tmp->num_msg                       ,1);
        cl->bcast(root,&tmp->num_warmup_msg                ,1);
        cl->bcast(root,&tmp->len_msg                       ,1);
        cl->bcast(root,&tmp->min_walltime                  ,1);
        cl->bcast(root,&tmp->min_iterations                ,1);
        cl->bcast(root,&tmp->max_stest                     ,1);
        cl->bcast(root,&tmp->do_use_gpus                   ,1);
        cl->bcast(root,&tmp->do_bisection                  ,1);
        cl->bcast(root,&tmp->do_group_processes_by_hostname,1);
        cl->bcast(root,&tmp->use_multi_buf                 ,1);
        cl->bcast(root,&tmp->num_multi_buf                 ,1);
        cl->bcast(root,&tmp->randomize_buffers             ,1);
        cl->bcast(root,&tmp->buf_mt_seed                   ,1);
        cl->bcast(root,&tmp->check_buffers                 ,1);
        cl->bcast(root,&tmp->seed_randomize_steps          ,1);
        cl->bcast(root,&tmp->seed_randomize_tasks          ,1);

        auto bcastString = [&](std::string& str){
            std::uint64_t str_size = str.size(); // only root has correct value
            cl->bcast(root, &str_size, 1);
            if(cl->rank() != root) {
                str.resize(str_size);
            }
            cl->bcast(root, str.data(), str.size());
        };

        bcastString(tmp->output);
        bcastString(tmp->virtual_cluster_implementation);

        return(tmp);
}

struct linktest_args *localize_cmdline_args(int rank, struct linktest_args* args) {

    std::map<std::string,Alloc::Type> alloc_map={{"DEFAULT"                      ,Alloc::Type::Default           },
                                                {"aligned-memory_allocator"      ,Alloc::Type::AlignedMalloc     },
                                                {"pinned-memory-map_allocator"   ,Alloc::Type::PinnedMmap        },
                                                {"POSIX_aligned-memory_allocator",Alloc::Type::POSIXAlignedMalloc},
                                                {"CUDA_memory_allocator"         ,Alloc::Type::CUDA              }};
    try{
        cmdline_args.alloc_typ=alloc_map.at(cmdline_args.memory_buffer_allocator);
    }catch(const std::out_of_range& oor) {
        fatal("Unknown memory-allocator type: %s.",cmdline_args.memory_buffer_allocator);
    }
    
    if(cmdline_args.virtual_cluster_implementation=="cuda"){
        args->alloc_typ = Alloc::Type::CUDA;
    }

    if(args->do_use_gpus) {
        bool isGpuRank = (std::size_t)rank >= args->use_gpus_on_task.size() || args->use_gpus_on_task.at(rank);
        args->alloc_typ = isGpuRank ? Alloc::Type::CUDA : Alloc::Type::Default;
    }

    if(cmdline_args.alloc_typ==Alloc::Type::Default){
        cmdline_args.alloc_typ=Alloc::Type::POSIXAlignedMalloc;
    }

    return args;
}

void print_cmdline_usage(const std::string& prog)
{
        std::fprintf(stderr,
                "Version : %d.%d.%d (%s)\n"
                "Usage   : %s [options]\n\n"
                "Possible options (default values in parenthesis):\n\n", 
                VERSION_MAJOR,
                VERSION_MINOR,
                VERSION_PATCH,
                GIT_HASH,
                prog.c_str()
        );

        for(auto& argdef: linktest_all_cmdline_argdefs) {
                std::string val = (1 == argdef.nargs) ? "VAL" : "";

                char buf1[256];
                std::snprintf(buf1, sizeof(buf1), "%s %s", argdef.opt().c_str(), val.c_str());

                char buf2[256];
                if ((not argdef.required) and (argdef.defaultval.length() > 0)) {
                        std::snprintf(buf2, sizeof(buf2), "(%s)", argdef.defaultval.c_str());
                } else {
                        buf2[0] = 0;
                }

                std::fprintf(stderr, " %-32s%-72s%32s\n", buf1, argdef.help.c_str(), buf2);
        }

        std::string modeList = "[";
        if(VirtualCluster::impls.size() > 0) {
            for(const auto& name: VirtualCluster::impls) {
                modeList = modeList + name + ", ";
            }
            modeList.erase(modeList.size()-2);
        }
        modeList += "]";
        std::fprintf(stderr,
                "\n\n"
                "* This build supports %s. \n"
                "  Alternatively to --mode, the transport layer can be defined by using linktest.LAYER\n"
                "  or setting environment variable LINKTEST_VCLUSTER_IMPL\n", modeList.c_str());
}

void print_cmdline_args(const struct linktest_args* args){
    auto getPatternDescription = [args]() {
        if(args->do_bisection){
            if(args->do_bidir ) return "Bidirectional End-to-End Bisection";
            if(args->do_unidir) return "Unidirectional End-to-End Bisection";
            return "Semidirectional End-to-End Bisection";
        }else{
            if(args->do_bidir ) return "Bidirectional End-to-End";
            if(args->do_unidir) return "Unidirectional End-to-End";
            return "Semidirectional End-to-End";
        }
    };

    auto getMemoryBufferType = [args](){
        switch(args->alloc_typ){
            case(Alloc::Type::Invalid):
            case(Alloc::Type::Default):
                return "Invalid";
            case(Alloc::Type::AlignedMalloc):
                return "aligned_alloc";
            case(Alloc::Type::PinnedMmap):
                return "mmap";
            case(Alloc::Type::POSIXAlignedMalloc):
                return "posix_memalign";
            case(Alloc::Type::CUDA):
                #if HAVE_VCLUSTER_CUDA == 1
                    return "CUDA";
                #else
                    return "No CUDA";
                #endif
            default:
                return "Unknown";
        }
    };

    auto getProtocol = [args]() {
        if(args->do_nosion)   return "No";
        if(args->do_sion_par) return "Yes, parallel";
        else                  return "Yes, funnelled";
    };

    auto getYesNo = [](int boolean) {
        return boolean ? "Yes" : "No";
    };

    std::printf("\n\n");
    std::printf("LinkTest: %d.%d.%d (%s)\n", VERSION_MAJOR, VERSION_MINOR, VERSION_PATCH, GIT_HASH);
    std::printf(HLINE);
    std::printf("Virtual-cluster implementation:     %-s\n"             ,args->virtual_cluster_implementation.c_str());
    std::printf("Message length:                     %-" PRIu64 " B\n"  ,args->len_msg);
    std::printf("Number of messages:                 %-" PRIu64 "\n"    ,args->num_msg);
    std::printf("Number of warm-up messages:         %-" PRIu64 "\n"    ,args->num_warmup_msg);
    std::printf("Communication pattern:              %-s\n"             ,getPatternDescription());
    std::printf("Memory-buffer type:                 %-s\n"             ,getMemoryBufferType());
    std::printf("Combine tasks on nodes?             %-s\n"             ,getYesNo(args->do_group_processes_by_hostname));
    std::printf("Additionally test all-to-all?       %-s\n"             ,getYesNo(args->do_alltoall));
    std::printf("Use GPU memory?                     %-s\n"             ,getYesNo(args->do_use_gpus));
    std::printf("Use GPU memory on tasks             %-s\n"             ,VectorBool::toString(args->use_gpus_on_task).c_str());
    std::printf("Randomize test order?               %-s\n"             ,getYesNo(args->do_mix));
    std::printf("Perform serial tests?               %-s\n"             ,getYesNo(args->do_serial));
    std::printf("Maximum number of serial retests?   %-" PRIu64 "\n"    ,args->max_stest);
    std::printf("Check buffers?                      %-s\n"             ,getYesNo(args->check_buffers));
    std::printf("Randomize buffers?                  %-s\n"             ,getYesNo(args->randomize_buffers));
    if(args->randomize_buffers){
        std::printf("Buffer MT seed:                     %-" PRIu64 "\n",args->buf_mt_seed);
    }
    std::printf("Use multiple message buffers?       %-s\n"     ,getYesNo(args->use_multi_buf));
    if(args->use_multi_buf){
        std::printf("Number of multiple message buffers: %-" PRIu64 "\n",args->num_multi_buf);
    }
    std::printf("Randomize test order?               %-s\n"             ,getYesNo(args->do_mix));
    if(args->do_mix){
        std::printf("Seed for step randomization:        %-" PRIu64 "\n" ,args->seed_randomize_steps);
    }
    std::printf("Number of communication schemes     %-" PRIu64 "\n" ,args->num_randomize_tasks);
    if(args->num_randomize_tasks){
        std::printf("Seed for random comm. schemes:      %-" PRIu64 "\n" ,args->seed_randomize_tasks);
    }
    std::printf("Write output SION file?             %-s\n"          ,getProtocol());
    if(args->do_nosion==0) {
        std::printf("Output SION filename:               %-s\n"          ,args->output.c_str());
    }
    std::printf(HLINE);
}
