#include "format_units.h"

#include <cstdio>
#include <math.h>
#include <memory>

const char* const SI_PREFIX[]={"y","z","a","f","p","n","u","m"," ","k","M","G","T","P","E","Z","Y"};
const char* const IEC_PREFIX[]={"  ","ki","Mi","Gi","Ti","Pi","Ei","Zi","Yi"};

namespace UnitPrefix{

std::unique_ptr<char[]> IEC_prefix(double value, unsigned int width){
    /* Formats the double-precision value "value" to a null-
     * terminated string with width "width" plus an additional
     * three characters for a space and the two-character IEC
     * prefix. Inside the IEC prefix range we print at least
     * four digits plus the sign and decimal separator. Outside
     * the IEC range we use exponential E notation with a
     * three-digit signed exponent. Note that at most 64-bit
     * doubles are supported. Minimum width should be at least
     * 9 for exponential notation and 7 for prefix-based
     * notation.
     *
     * INPUTS:
     *         value: The double precision value to be
     *                formatted.
     *         width: The unsigned integer number of digits to
     *                be printed. This includes the sign and
     *                decimal separator. Additionally a space
     *                and the correct prefix are appended.
     *                This value should at a minimum exceed 5
     *                for SI-prefix notation or 7 when using
     *                exponential notation in lieu of the 
     *                "value" being outside the supported SI
     *                range. Shorter widths are expanded as
     *                necessary.
     *
     * OUTPUT: A unique pointer to an allocated char array
     *         containing the formatted "value".
     */
    
    /* Determine IEC-Prefix & Normalize Value*/
    std::unique_ptr<char[]> c_string(new char[width+4]); //4 = 3 white spaces, or 1 white space and 2 IEC prefix, & 1 NULL terminator
    if(value==(double)0.){
        std::sprintf(c_string.get(),"%*.*f   ",width,width-6,0.0);
        return(c_string);
    }
    int    floor_exponent_2_int   =(int)floor(log2(abs(value)));
    if(floor_exponent_2_int<0||floor_exponent_2_int>80){
        double floor_exponent_10_double=floor(log10(abs(value)));
        value/=pow((double)10.0,floor_exponent_10_double);
        // We use E-Notation
        if(width<9){ //We have to extend the width to make sure there is enough room for the exponent
            width=9;
        }
        std::unique_ptr<char[]> c_string(new char[width+4]); //4 = 3 white space & 1 NULL terminator
        width-=5;
        if(floor_exponent_2_int<0){
            std::sprintf(c_string.get(),"%*.*fe-%03d   ",width,width-3,value,-(int)floor_exponent_10_double);
        }else{
            std::sprintf(c_string.get(),"%*.*fe+%03d   ",width,width-3,value, (int)floor_exponent_10_double);
        }
        return(c_string);
    }else{
        int IEC_exponent=(floor_exponent_2_int/10)*10;
        value/=pow((double)2.0,(double)IEC_exponent);
        if(width<7){ //We have to extend the width to make sure there is enough room for the required digits
            width=7;
        }
        std::sprintf(c_string.get(),"%*.*f %s",width,width-6,value,IEC_PREFIX[IEC_exponent/10]);
        return(c_string);
    }

    return(NULL); //This line should not be reached under normal execution
}

std::unique_ptr<char[]> SI_prefix(double value, unsigned int width){
    /* Formats the double-precision value "value" to a null-
     * terminated string with width "width" plus an additional
     * three characters for a space and the two-character IEC
     * prefix. Inside the IEC prefix range we print at least
     * four digits plus the sign and decimal separator. Outside
     * the IEC range we use exponential E notation with a
     * three-digit signed exponent. Note that at most 64-bit
     * doubles are supported. Minimum width should be at least
     * 9 for exponential notation and 6 for prefix-based
     * notation.
     *
     * INPUTS:
     *         value: The double precision value to be
     *                formatted.
     *         width: The unsigned integer number of digits to
     *                be printed. This includes the sign and
     *                decimal separator. Additionally a space
     *                and the correct prefix are appended.
     *                This value should at a minimum exceed 5
     *                for SI-prefix notation or 6 when using
     *                exponential notation in lieu of the 
     *                "value" being outside the supported SI
     *                range. Shorter widths are expanded as
     *                necessary.
     *
     * OUTPUT: A unique pointer to an allocated char array
     *         containing the formatted "value".
     */

    /* Determine SI-Prefix & Normalize Value*/
    std::unique_ptr<char[]> c_string(new char[width+3]); //3 = 2 white spaces, or 1 white space & 1 SI prefix, & 1 NULL terminator
    if(value==(double)0.){
        std::sprintf(c_string.get(),"%*.*f  ",width,width-5,0.0);
        return(c_string);
    }
    double floor_exponent_10_double=log10(abs(value));
    if(floor_exponent_10_double<0) floor_exponent_10_double-=2.0;
    floor_exponent_10_double=floor(floor_exponent_10_double);
    int    floor_exponent_10_int   =(int)floor_exponent_10_double;
    if(floor_exponent_10_int<-27||floor_exponent_10_int>24){
        if(floor_exponent_10_int<0) floor_exponent_10_int+=2;
        value/=pow((double)10.0,(double)floor_exponent_10_int);
        // We use E-Notation
        if(width<9){ //We have to extend the width to make sure there is enough room for the exponent
            width=9;
        }
        width-=5;
        if(floor_exponent_10_int<0){
            std::sprintf(c_string.get(),"%*.*fe-%03d  ",width,width-3,value,-floor_exponent_10_int);
        }else{
            std::sprintf(c_string.get(),"%*.*fe+%03d  ",width,width-3,value, floor_exponent_10_int);
        }
        return(c_string);
    }else{
        int SI_exponent=(floor_exponent_10_int/3)*3;
        value/=pow((double)10.0,(double)SI_exponent);
        if(width<6){ //We have to extend the width to make sure there is enough room for the required digits
            width=6;
        }
        std::sprintf(c_string.get(),"%*.*f %s",width,width-5,value,SI_PREFIX[SI_exponent/3+8]);
        return(c_string);
    }

    return(NULL); //This line should not be reached under normal execution
}

}
