/****************************************************************************
**  LinkTest                                                               **
*****************************************************************************
**  Copyright (c) 2008-2022                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/
#include "config.h"
#include "compiler.h"
#include "error.h"
#include "environ.h"
#include "ibverbs_ctx.h"

#include <stdexcept>

namespace linktest
{
namespace ibverbs
{

HCASelectorWithFunction::HCASelectorWithFunction(MatchFunction f)
: f_(f)
{
}

bool HCASelectorWithFunction::match(int dev_id, int sorted_dev_id,
                    const std::string& name, int port)
{
    return f_(dev_id, sorted_dev_id, name, port);
}

void Context::all_ports_in_active_state_(struct ibv_device* dev,
                     std::vector<int>& ports)
{
    struct ibv_device_attr dattr;
    struct ibv_port_attr pattr;
    int port;

    if (unlikely(!dev)) {
        throw std::runtime_error("== is nullptr");
    }

    auto ctx = ibv_open_device(dev);
    if (unlikely(!ctx)) {
        error("ibv_open_device() failed");
        return;
    }

    EXEC_IFFAIL(ibv_query_device(ctx, &dattr), error("ibv_query_device() failed"); return);

    for (port = 1; port <= dattr.phys_port_cnt; ++port) {
        if (unlikely( ibv_query_port(ctx, port, &pattr)))
            continue;

        if ((IBV_PORT_ACTIVE == pattr.state) and
            (pattr.lid > 0)) {
            ports.push_back(port);
        }
    }

    ibv_close_device(ctx);
}

static inline std::string device_key_string(ibv_device* dev, int p)
{
    return std::string(ibv_get_device_name(dev)) + ":" + std::to_string(p);
}

void Context::query_all_active_hcas_(ActiveHCAMap& m)
{
    for (auto i = 0; devs[i]; ++i) {
        std::vector<int> ports;
        all_ports_in_active_state_(devs[i], ports);

        for(auto p: ports) {
            m.insert({device_key_string(devs[i], p), {i, p}});
        }
    }
}

int Context::find_and_open_hca_(const ActiveHCAMap& m, HCASelector& selector)
{
    devnum  = -1;
    portnum_ = -1;

    auto i = 0;
    for(auto& j: m) {
        auto k = j.first.find_first_of(":");

        if (selector.match(j.second.first, i, 
                   j.first.substr(0, k),
                   j.second.second)) {
            devnum = j.second.first;
            portnum_ =  j.second.second;
            break;
        }

        ++i;
    }

    if (unlikely((devnum < 0) || (portnum_ < 0))) {
        error("Could not get hold of a matching/working HCA");
        return ERROR;
    }

    return SUCCESS;
}

Context::Context(HCASelector& selector)
: devs(nullptr), ctx(nullptr)
{
    devs = ibv_get_device_list(NULL);
    if (unlikely(!devs)) {
        error("ibv_get_device_list returned NULL.");
        throw;
    }

    std::map<std::string, std::pair<int, int>> m;
    query_all_active_hcas_(m);

    auto ret = find_and_open_hca_(m, selector);
    if (unlikely(ret))
        goto fail;

    ctx = ibv_open_device(devs[devnum]);
    if (unlikely(!ctx)) {
        error("ibv_open_device failed.");
        goto fail;
    }

    return;

fail:
    if (ctx)
        ibv_close_device(ctx);
    if (devs)
        ibv_free_device_list(devs);

    throw;
}

Context::~Context()
{
    auto ret = ibv_close_device(ctx);
    if (unlikely(ret)) {
        warn("ibv_close_device failed.");
    }

    ibv_free_device_list(devs);
}

}
}
