/****************************************************************************
**  LinkTest                                                               **
*****************************************************************************
**  Copyright (c) 2008-2022                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/
#ifndef LINKTEST_IBVERBS_CTX_H
#define LINKTEST_IBVERBS_CTX_H

#include <vector>
#include <map>
#include <string>
#include <functional>

#include <infiniband/verbs.h>


namespace linktest
{
namespace ibverbs
{

/* Base class for HCA selection logic
 * Implementations of `HCASelector` provide a `match` function that is used
 * to find the right HCA to use for communication.
 */
class HCASelector
{

public:
    virtual bool match(int dev_id, int sorted_dev_id,
               const std::string& name, int port) = 0;

};

class HCASelectorWithFunction : public HCASelector
{

private:
    using MatchFunction = 
            std::function<bool(int, int, const std::string&, int)>;

public:
    HCASelectorWithFunction(MatchFunction f);

public:
    virtual bool match(int dev_id, int sorted_dev_id,
               const std::string& name, int port);

private:
    MatchFunction   f_;

};

// Wrapper around an `ibv_context` 
class Context
{

public:
    Context(HCASelector& selector);
    ~Context();

private:
    using ActiveHCAMap = std::map<std::string, std::pair<int, int>>;
    // Get a list of all active HCAs and ports
    void query_all_active_hcas_(ActiveHCAMap& m);

    // Retrieve the list of all active ports of the device that are usable for
    // LinkTest. Ethernet ports are skipped (identified by a zero LID).
    void all_ports_in_active_state_(struct ibv_device* dev,
                    std::vector<int>& ports);

    int find_and_open_hca_(const ActiveHCAMap& m, HCASelector& selector);

public:
    constexpr int port_number() const;

public:
    struct ibv_device**     devs;
    int                     devnum;

    struct ibv_context*     ctx;

private:
    int                     portnum_;

};

constexpr int Context::port_number() const
{
    return portnum_;
}

}
}
#endif
