/****************************************************************************
**  LinkTest                                                               **
*****************************************************************************
**  Copyright (c) 2008-2022                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/
#include "config.h"
#include "compiler.h"
#include "error.h"
#include "ibverbs_qp.h"
#include "ibverbs_ctx.h"
#include "ibverbs_cq.h"
#include "ibverbs_pd.h"

#include <utility>
#include <stdexcept>


namespace linktest
{
namespace ibverbs
{

QueuePair::QueuePair(Context* ctx, CompletionQueue* cq, ProtectionDomain* pd)
: ctx_(ctx), cq_(cq), pd_(pd), qp_(nullptr)
{
    create_ibv_qp_();
}

QueuePair::QueuePair(QueuePair&& other)
: ctx_(std::exchange(other.ctx_, nullptr)),
  cq_(std::exchange(other.cq_, nullptr)),
  pd_(std::exchange(other.pd_, nullptr)),
  qp_(std::exchange(other.qp_, nullptr))
{
}

QueuePair::~QueuePair()
{
    if (qp_) {
        auto err = ibv_destroy_qp(qp_);
        if (unlikely(err)) {
            warn("ibv_destroy_qp failed.");
            // Do not throw an exception
        }
    }
}

void QueuePair::create_ibv_qp_()
{
    struct ibv_qp_init_attr iattr = {
        .send_cq = cq_->cq,
        .recv_cq = cq_->cq,
        .cap = {
            // We will not have more than one WR posted at a time.
            .max_send_wr     = kDefault_max_send_wr,
            .max_recv_wr     = kDefault_max_recv_wr,
            .max_send_sge    = kDefault_max_send_sge,
            .max_recv_sge    = kDefault_max_recv_sge,
            .max_inline_data = kDefault_max_inline_data
        },
        .qp_type = IBV_QPT_RC,
        .sq_sig_all = 0         // Each send WR will specify whether or not
                    // to create a CQE.
    };

    qp_ = ibv_create_qp(pd_->pd, &iattr);
    if (unlikely(!qp_)) {
        error("ibv_create_qp() failed");
        throw std::runtime_error("ibv_create_qp() failed");
    }
}

int QueuePair::post_empty_recv_wr()
{
    ibv_recv_wr wr = {
        .wr_id          = 0,
        .next           = nullptr,
        .sg_list        = nullptr,
        .num_sge        = 0
    };
    ibv_recv_wr* bad_wr;

    EXEC_IFFAIL(ibv_post_recv(qp_, &wr, &bad_wr), error("ibv_post_recv failed."); return ERROR);

    return SUCCESS;
}

int QueuePair::move_to_init(std::uint8_t portnum)
{
    struct ibv_qp_attr attr = {
        .qp_state        = IBV_QPS_INIT,
        .qp_access_flags = IBV_ACCESS_REMOTE_WRITE,
        .pkey_index      = kDefault_pkey_index,
        .port_num        = portnum
    };

    auto ret = ibv_modify_qp(qp_, &attr,
                 IBV_QP_STATE |
                 IBV_QP_PKEY_INDEX |
                 IBV_QP_PORT |
                 IBV_QP_ACCESS_FLAGS);
    if (unlikely(ret)) {
        error("Moving queue pair to init state failed.");
        return ERROR;
    }

    return SUCCESS;
}

int QueuePair::move_to_rtr(const RemoteAddressInfo& other, std::uint8_t sl, std::uint8_t portnum)
{
    struct ibv_qp_attr attr = {
        .qp_state       = IBV_QPS_RTR,
        .path_mtu       = IBV_MTU_512,  /* recommended value */
        .rq_psn         = static_cast<std::uint32_t>(other.psn),
        .dest_qp_num    = static_cast<std::uint32_t>(other.qpn),
        /* AH (address handle) attributes */
        .ah_attr        = {
            .dlid           = static_cast<std::uint16_t>(other.lid),
            .sl             = sl,
            .src_path_bits  = kDefault_ah_src_path_bits,
            .is_global      = kDefault_ah_is_global,
            .port_num       = portnum
        },
        .max_dest_rd_atomic     = kDefault_max_dest_rd_atomic,
        .min_rnr_timer          = kDefault_min_rnr_timer
    };

    auto ret = ibv_modify_qp(qp_, &attr,
                 IBV_QP_STATE |
                 IBV_QP_PATH_MTU |
                 IBV_QP_DEST_QPN |
                 IBV_QP_RQ_PSN |
                 IBV_QP_MAX_DEST_RD_ATOMIC |
                 IBV_QP_MIN_RNR_TIMER |
                 IBV_QP_AV);
    if (unlikely(ret)) {
        error("Moving queue pair to ready-to-receive state failed.");
        return ERROR;
    }

    return SUCCESS;
}

int QueuePair::move_to_rts()
{
    struct ibv_qp_attr attr = {
        .qp_state       = IBV_QPS_RTS,
        .sq_psn         = kDefault_sq_psn,
        .max_rd_atomic  = kDefault_max_rd_atomic,
        .timeout        = kDefault_timeout,
        .retry_cnt      = kDefault_retry_cnt,
        .rnr_retry      = kDefault_rnr_retry
    };

    auto ret = ibv_modify_qp(qp_, &attr,
                 IBV_QP_STATE |
                 IBV_QP_TIMEOUT |
                 IBV_QP_RETRY_CNT |
                 IBV_QP_RNR_RETRY |
                 IBV_QP_SQ_PSN |
                 IBV_QP_MAX_QP_RD_ATOMIC);
    if (unlikely(ret)) {
        error("Moving queue pair to ready-to-send state failed.");
        return ERROR;
    }

    return SUCCESS;
}

int QueuePair::post_send(struct ibv_send_wr* wr, struct ibv_send_wr** bwr)
{
    EXEC_IFFAIL(ibv_post_send(qp_, wr, bwr), error("ibv_post_send() failed."); return ERROR);

    return SUCCESS;
}


}
}
