/****************************************************************************
**  LinkTest                                                               **
*****************************************************************************
**  Copyright (c) 2008-2022                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/
#ifndef LINKTEST_IBVERBS_QP_H
#define LINKTEST_IBVERBS_QP_H

#include <infiniband/verbs.h>


namespace linktest
{
namespace ibverbs
{

class Context;
class CompletionQueue;
class ProtectionDomain;


/* Address information for exchange between communication
 * partner. struct address stores all information needed
 * to set up a queue pair (and more).
 */
class RemoteAddressInfo
{

public:
    int     lid;
    int     qpn;
    int     psn;
    /* We are skipping the ibv_gid since IB routing
     * is not implemented by any vendor currently
     * and thus the GID is not important for our
     * unicast traffic.
     */
};


// An InfiniBand queue pair
class QueuePair
{

public:
    QueuePair(Context* ctx, CompletionQueue* cq, ProtectionDomain* pd);

    QueuePair(const QueuePair&) = delete;
    QueuePair(QueuePair&& other);

public:
    ~QueuePair();

public:
    // Queue pair number
    constexpr std::uint32_t      number() const;

public:
    int     post_empty_recv_wr();

    // Move the queue pair from `RESET` to `INIT` state.
    int     move_to_init(std::uint8_t portnum);
    // Move the queue pair from `INIT` to `RTR` state.
    int     move_to_rtr(const RemoteAddressInfo& other, std::uint8_t sl, std::uint8_t portnum);
    // Move the queue pair to the `RTS` state. This should done after the
    // queue pair has been successfully moved into `RTR`state.
    int     move_to_rts();

public:
    // Post a send to the queue pair
    int     post_send(struct ibv_send_wr* wr, struct ibv_send_wr** bwr);

private:
    Context*                ctx_;
    CompletionQueue*        cq_;
    ProtectionDomain*       pd_;

private:
    struct ibv_qp*          qp_;

private:
    void    create_ibv_qp_();

private:
    enum {
        kDefault_max_send_wr = 16,
        kDefault_max_recv_wr = 16,
        kDefault_max_send_sge = 1,
        kDefault_max_recv_sge = 1,
        kDefault_max_inline_data = LINKTEST_IBVERBS_MAX_INLINE_SZ_,
        kDefault_pkey_index = 0,
        kDefault_ah_src_path_bits = 0,  // Use base LID (0 == LMC)
        kDefault_ah_is_global = 0,      // GRH attributes are not valid
                        // (since they are not set).
                        // We stay within one subnet.
        kDefault_max_dest_rd_atomic = 1,
        kDefault_min_rnr_timer = 12,    // recommended value
        kDefault_sq_psn = 0,
        kDefault_max_rd_atomic = 1,     // Number of outstanding RDMA
                        // reads & atomic operations
                        // on the destination QP
        kDefault_timeout = 14,          // Local ack timeout for
                        // primary path
        kDefault_retry_cnt = 7,         // Retry count
        kDefault_rnr_retry = 7,         // RNR (receiver not ready)
                        // retry
    };

};

constexpr std::uint32_t QueuePair::number() const
{
    return qp_->qp_num;
}


}
}

#endif
