/****************************************************************************
**  LinkTest                                                               **
*****************************************************************************
**  Copyright (c) 2008-2022                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/
#ifndef LINKTEST_MEMORY_H
#define LINKTEST_MEMORY_H

#include <cstddef>
#include "config.h"
#include <cstdint>
#include <unistd.h>
#include <string>

#if HAVE_VCLUSTER_CUDA == 1
	namespace linktest{
		namespace cuda{
			class Allocator;
			class GpuContext;
		}
	}
#endif
namespace Alloc{
    enum Type {
        Invalid           =0,
        Default           =1,
        AlignedMalloc     =2,
        PinnedMmap        =3,
        POSIXAlignedMalloc=4,
        CUDA              =5
    };

    std::string getName(Type alloc_t);
}
namespace AddressSpace{
	enum class ID{
		Invalid = 0,
		/* Local main memory on the stack or heap of the process or any VMA in the
		* process memory
		*/
		Local,
		// A remote location (likely accessible via RDMA but not via load/store).
		Remote,
		// The address space of a locally connected CUDA-programmed device.
		CudaDeviceLocal,
		// The address space of a CUDA-programmed device on a remote system.
		CudaDeviceRemote
	};

	// Map a local address space to the corresponding remote address space.
	ID to_remote(ID localID);
}

// Interface for a memory allocator
class Allocator{
	public:
		/* Information about the address space. All memory allocations the allocator
		 * returns will be in this adress space.
		 */
		[[nodiscard]] virtual AddressSpace::ID address_space_id() const = 0;

		/* Allocate a memory buffer of size len with aligment alignas (must be a 
		 * power of two). The function returns the address of the memory region.
		 */
		virtual void*        alloc(std::size_t len,std::size_t align) = 0;

		// Free a memory buffer.
		virtual int          free(void*  p,std::size_t len) = 0;

		virtual ~Allocator()                            = default;
		Allocator()                                     = default;
		Allocator(const Allocator& copyFrom)            = delete;
		Allocator& operator=(const Allocator& copyFrom) = delete;
		Allocator(Allocator &&)                         = delete;
		Allocator& operator=(Allocator &&)              = delete;
};

// Simple allocator using 
class MallocAllocator : public Allocator {
	public:
		[[nodiscard]] AddressSpace::ID address_space_id() const override;
		void*        alloc(std::size_t len, std::size_t align) override;
		int          free(void* p, std::size_t len) override;

		~MallocAllocator() override                                 = default;
		MallocAllocator()                                           = default;
		MallocAllocator(const MallocAllocator& copyFrom)            = delete;
		MallocAllocator& operator=(const MallocAllocator& copyFrom) = delete;
		MallocAllocator(MallocAllocator &&)                         = delete;
		MallocAllocator& operator=(MallocAllocator &&)              = delete;
};

/* An allocator using mmap() and mlock(), or equivalent, to create
 * a memory-pinned page-aligned memory allocation.
 */
class PinnedMmapAllocator : public Allocator {
	public:
		[[nodiscard]] AddressSpace::ID address_space_id() const override;
		void*        alloc(std::size_t len, std::size_t align) override;
		int          free(void* p, std::size_t len) override;

		~PinnedMmapAllocator() override                                     = default;
		PinnedMmapAllocator()                                               = default;
		PinnedMmapAllocator(const PinnedMmapAllocator& copyFrom)            = delete;
		PinnedMmapAllocator& operator=(const PinnedMmapAllocator& copyFrom) = delete;
		PinnedMmapAllocator(PinnedMmapAllocator &&)                         = delete;
		PinnedMmapAllocator& operator=(PinnedMmapAllocator &&)              = delete;
	private:
		size_t       pgsize_ = sysconf(_SC_PAGESIZE);

};

/* An allocator using POSIX memalign, or
 * equivalent, to allocate a memory region.
 */
class PosixMemAlignedAllocator : public Allocator {
	public:
		[[nodiscard]] AddressSpace::ID address_space_id() const override;
		void*        alloc(std::size_t len, std::size_t align) override;
		int          free(void* p, std::size_t len) override;

		~PosixMemAlignedAllocator() override                                          = default;
		PosixMemAlignedAllocator()                                                    = default;
		PosixMemAlignedAllocator(const PosixMemAlignedAllocator& copyFrom)            = delete;
		PosixMemAlignedAllocator& operator=(const PosixMemAlignedAllocator& copyFrom) = delete;
		PosixMemAlignedAllocator(PosixMemAlignedAllocator &&)                         = delete;
		PosixMemAlignedAllocator& operator=(PosixMemAlignedAllocator &&)              = delete;
	private:
		size_t       pgsize_ = sysconf(_SC_PAGESIZE);
};

#if HAVE_VCLUSTER_CUDA == 1
// A memory allocation on a GPU
class CudaDeviceAllocator : public Allocator{
	public:
		[[nodiscard]] AddressSpace::ID    address_space_id() const override;
		void*           alloc(std::size_t len, std::size_t align) override;
		int             free(void* p, std::size_t len) override;

		~CudaDeviceAllocator() override                                     = default;
		CudaDeviceAllocator()                                               = default;
		CudaDeviceAllocator(linktest::cuda::GpuContext* ctx);
		CudaDeviceAllocator(const CudaDeviceAllocator& copyFrom)            = delete;
		CudaDeviceAllocator& operator=(const CudaDeviceAllocator& copyFrom) = delete;
		CudaDeviceAllocator(CudaDeviceAllocator &&)                         = delete;
		CudaDeviceAllocator& operator=(CudaDeviceAllocator &&)              = delete;
	private:
		linktest::cuda::GpuContext*     ctx_;
};
#endif


// A memory region on some memory adress space.
class MemoryBuffer{
	public:
		MemoryBuffer(Allocator* alloc, std::size_t len, std::size_t align);

		// Create a memory buffer for random initialization using LCE
		MemoryBuffer(Allocator* alloc, std::size_t len, std::size_t align, std::uint_fast64_t buf_mt_seed);

		// Create a memory buffer from a pre-existing 
		MemoryBuffer(void* p, std::size_t len, AddressSpace::ID addr_space_id);

		virtual ~MemoryBuffer();

		MemoryBuffer(const MemoryBuffer& other)             = delete;
		MemoryBuffer& operator=(const MemoryBuffer& other)  = delete;
		MemoryBuffer(MemoryBuffer&& other)                  = default;
		MemoryBuffer& operator=(MemoryBuffer&& other)       = default;

		/* Wrap an existing pointer into a memory buffer. We do not know the allocator
		 * and hence have to ingore it. This is acceptable since the allocator is not
		 * exposed via a public function.
		 */
		static MemoryBuffer wrap(void* p, std::size_t len, AddressSpace::ID addr_space_id);
		template<typename T>
		static MemoryBuffer wrap(T* p, std::size_t len, AddressSpace::ID addr_space_id);

		// Fill buffer with some content
		void fill();

		// Check buffer - returns non-zero if an error is detected
		int check();

		/* Copy from `src` to `dst` memory buffer. Both `MemoryBuffer` instances
		 * must be in the same adress space for this operation to succeed.
		 */
		static void memory_copy(MemoryBuffer& dst, MemoryBuffer& src);

		[[nodiscard]] constexpr AddressSpace::ID   address_space_id() const;
		[[nodiscard]] constexpr void*              p()                const;
		[[nodiscard]] constexpr std::size_t        len()              const;
		[[nodiscard]] constexpr std::size_t        align()            const;
		[[nodiscard]] constexpr bool               use_mt()           const;
		[[nodiscard]] constexpr std::uint_fast64_t buf_mt_seed()      const;
		template<typename T>
		[[nodiscard]] constexpr T*                 pointer()          const;

	private:
		AddressSpace::ID   addr_space_id_; //Address Space
		void*              p_            ; //Pointer to buffer
		std::size_t        len_          ; //Buffer size
		std::size_t        align_        ; //Buffer alignment
		Allocator*         alloc_        ; //Pointer to allocator
		bool               use_mt_       ; //Mersenne Twister 19937 (MT19337) Pseudo-Random-Number Generator (PRNG) if non-zero
		std::uint_fast64_t buf_mt_seed_  ; //MT19337 Seed
};

constexpr AddressSpace::ID   MemoryBuffer::address_space_id() const{return addr_space_id_;}
constexpr void*              MemoryBuffer::p()                const{return p_            ;}
constexpr std::size_t        MemoryBuffer::len()              const{return len_          ;}
constexpr std::size_t        MemoryBuffer::align()            const{return align_        ;}
constexpr bool               MemoryBuffer::use_mt()           const{return use_mt_       ;}
constexpr std::uint_fast64_t MemoryBuffer::buf_mt_seed()      const{return buf_mt_seed_  ;}
template<typename T>
constexpr T*                 MemoryBuffer::pointer()          const{return reinterpret_cast<T*>(p());}
template<typename T>
MemoryBuffer MemoryBuffer::wrap(T* p, std::size_t len, AddressSpace::ID addr_space_id){
    return MemoryBuffer(reinterpret_cast<void*>(p), len*sizeof(T), addr_space_id);
}

#endif
