/****************************************************************************
**  LinkTest                                                               **
*****************************************************************************
**  Copyright (c) 2008-2022                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/
#include "output_sion.h"
#include "config.h"
#include "compiler.h"
#include "linktest.h"
#include "timing.h"
#include "stopwatch.h"
#include "error.h"
#include "cmdline.h"
#include "stats.h"
#include "vcluster_sion_generic_adapter.h"
#include "format_units.h"
#include "serializer.h"
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <type_traits>

#ifdef USE_SION
    #include <mpi.h>
    #ifdef __cplusplus
    extern "C" {
    #endif
    #include "sion.h"
    #ifdef __cplusplus
    }
    #endif
#endif

#ifdef USE_SION
static Serializer getSerializer(VirtualCluster* cl, const struct linktest_args *args, const std::vector<LinktestStats>& statsVec) {
    Serializer s;

    // Binary LinkTest Header
    if(cl->rank() == 0) {
    s.addDataSource(LINKTEST_ID_, LINKTEST_ID_, sizeof(LINKTEST_ID_)-1);
        // Data Version / Code Version / Timestamp  (if root)
        s.addDataSource("Version Major"                       ,&VERSION_MAJOR                                             );
        s.addDataSource("Version Minor"                       ,&VERSION_MINOR                                             );
        s.addDataSource("Version Patch"                       ,&VERSION_PATCH                                             );
        s.addDataSource("Git-Hash"                            , GIT_HASH_                           ,GIT_HASH_LEN_        );
        // Transport Layer
        s.addDataSource("Transport Layer Size"                ,&cl->nameSizeRef()                                         );
        s.addDataSource("Transport Layer"                     , cl->nameRef().c_str()               ,cl->nameSizeRef()    );
        // int8_t options
        s.addDataSource("args->do_alltoall"                   ,&args->do_alltoall                                         );
        s.addDataSource("args->do_bidir"                      ,&args->do_bidir                                            );
        s.addDataSource("args->do_unidir"                     ,&args->do_unidir                                           );
        s.addDataSource("args->do_bisection"                  ,&args->do_bisection                                        );
        s.addDataSource("args->do_mix"                        ,&args->do_mix                                              );
        s.addDataSource("args->do_serial"                     ,&args->do_serial                                           );
        s.addDataSource("args->do_nosion"                     ,&args->do_nosion                                           );
        s.addDataSource("args->do_sion_par"                   ,&args->do_sion_par                                         );
        s.addDataSource("args->do_use_gpus"                   ,&args->do_use_gpus                                         );
        s.addDataSource("args->use_multi_buf"                 ,&args->use_multi_buf                                       );
        s.addDataSource("args->randomize_buffers"             ,&args->randomize_buffers                                   );
        s.addDataSource("args->check_buffers"                 ,&args->check_buffers                                       );
        s.addDataSource("args->do_group_processes_by_hostname",&args->do_group_processes_by_hostname                      );
        s.addDataSource("args->alloc_typ"                     ,&args->alloc_typ                                           ); // enum aka int
        // uint64_t data
        s.addDataSource("args->num_msg"                       ,&args->num_msg                                             );
        s.addDataSource("args->len_msg"                       ,&args->len_msg                                             );
        s.addDataSource("args->num_warmup_msg"                ,&args->num_warmup_msg                                      );
        s.addDataSource("args->max_stest"                     ,&args->max_stest                                           );
        s.addDataSource("args->num_multi_buf"                 ,&args->num_multi_buf                                       );
        s.addDataSource("args->buf_mt_seed"                   ,&args->buf_mt_seed                                         );
        s.addDataSource("args->num_randomize_tasks"           ,&args->num_randomize_tasks                                 );
        s.addDataSource("args->seed_randomize_tasks"          ,&args->seed_randomize_tasks                                );
        s.addDataSource("args->seed_randomize_steps"          ,&args->seed_randomize_steps                                );
        s.addDataSource(END_HEADER_                           , END_HEADER_                         ,sizeof(END_HEADER_)-1);
    }

    s.addDataSource("hostname_len"  ,&cl->hostnameSizeRef()                          );
    s.addDataSource("hostname"      , cl->hostnameRef().c_str(),cl->hostnameSizeRef());
    s.addDataSource("pinned_cpu_num",&cl->cpuIDRef()                                 );

    for(auto& stats : statsVec) {
        // Aggregated metrics (if root)
        if(cl->rank() == 0) {
            s.addDataSource("stats.startTimeString", &stats.startTimeString);
            s.addDataSource("stats.min_time_all"   , &stats.min_time_all   );
            s.addDataSource("stats.max_time_all"   , &stats.max_time_all   );
            s.addDataSource("stats.avg_time_all"   , &stats.avg_time_all   );
            // Aggregated All-to-All Information
            if (args->do_alltoall) {
                s.addDataSource("stats.a2a_min_time_all", &stats.a2a_min_time_all);
                s.addDataSource("stats.a2a_max_time_all", &stats.a2a_max_time_all);
                s.addDataSource("stats.a2a_avg_time_all", &stats.a2a_avg_time_all);
            }
            // Serial Retests
            if(args->max_stest > 0) {
                s.addDataSource("stats.stimings_new"    , stats.stimings_new.data(), stats.stimings_new.size());
                s.addDataSource("stats.stimings_old"    , stats.stimings_old.data(), stats.stimings_old.size());
                s.addDataSource("stats.sfrom"           , stats.sfrom.data()       , stats.sfrom.size()       );
                s.addDataSource("stats.sto"             , stats.sto.data()         , stats.sto.size()         );
            }
            s.addDataSource("stats.EndTimeString"  , &stats.endTimeString  );
        }
        // Raw Matrix
        s.addDataSource("stats.ptimings"     , stats.ptimings.data()     , stats.ptimings.size()     );
        s.addDataSource("stats.accesspattern", stats.accesspattern.data(), stats.accesspattern.size());
        // All to all (if used)
        if (args->do_alltoall) {
            s.addDataSource("stats.a2a_time"       , &stats.a2a_time       );
        }
    }

    s.addDataSource(END_BLOCK_, END_BLOCK_, sizeof(END_BLOCK_)-1);

    return s;
}
#endif

int linktest_output_sion_collect_local_data(VirtualCluster* cl,
                                            const struct linktest_args *args,
                                            const std::vector<LinktestStats>& statsVec,
                                            char **buffer,
                                            long long *sz){
#ifdef USE_SION
    auto serializer = getSerializer(cl, args, statsVec);
    auto numBytesToBeWritten = serializer.getSize();

    char * buf = static_cast<char*>( malloc(sizeof(char) * numBytesToBeWritten) );
    if (unlikely(!buf)) {
        error("cannot allocate buf[%lu] ...", numBytesToBeWritten);
        return ERROR;
    }

    // Write to buffer
    serializer.write(buf, cl->rank());

    *buffer = buf;
    *sz = numBytesToBeWritten;

    if ( cl->rank() < 2 ) {
        std::fprintf(stderr, "[Task %d of %d] Data collected: %lu bytes\n", cl->rank(), cl->size(), numBytesToBeWritten);
    }

    return SUCCESS;
#else
    error("This version of the benchmark was compiled without SIONlib support.");

    *buffer = NULL;
    *sz = 0;

    return ERROR;
#endif
}

#ifdef USE_SION
static int linktest_output_sion_funnelled_root(VirtualCluster* cl,
                          const struct linktest_args *args,
                          const std::vector<LinktestStats>& statsVec,
                          char *buffer,
                          long long sz,
                          double begin){
    int i;
    long long dummy;
    sion_int64 *chunksizes = NULL;
    int *globalranks;
    char filename[256];
    int outsid;
    FILE *outfp;
    
    std::unique_ptr<StopwatchI> rootWatch = Stopwatchfactory::getRootWatch(cl->rank());
    auto printTiming = [rank=cl->rank(), &rootWatch](std::string step_name){
        printTimingIfRoot(rank, step_name, rootWatch->getDuration());
    };

    std::snprintf(filename, sizeof(filename), "%s", args->output.c_str());

    chunksizes = new sion_int64[cl->size()];
    globalranks = new int[cl->size()];

    static_assert(std::is_same<sion_int64, long long>::value);

    EXEC_NOFAIL(cl->gather(0, chunksizes, &sz, 1));
#ifdef DEBUG_OUTPUT
    if(cl->rank()==0){
        debug("[%3d] Chunksize: %d", 0, chunksizes[0]);
        for(int i=1;i<cl->size();i++) debug("[%3d] Chunksize: %d", i, chunksizes[i]);
    }
#endif
    /* The code implicitly assumes that the root process has the biggest chunksize
     * which is true in the current code because we write some of the statistics data
     * only on the root process. This assumption allows us to avoid allocating a
     * buffer for the incoming data from other processing elements.
     */
    for (i = 1; i < cl->size(); ++i) {
        if (unlikely(chunksizes[i] > chunksizes[0])) {
            fatal("chunksizes[%d] > chunksizes[0]", i);
            return ERROR;
        }
    }
    debug("linktest_output_sion_funnelled_root->barrier");
    EXEC_NOFAIL(cl->barrier());

    printTimingIfRoot(cl->rank(), "[sioncollect]", std::chrono::duration<double>(walltime() - begin));

    rootWatch->start();
    for (i = 0; i < cl->size(); ++i)
        globalranks[i] = i;
    auto size = cl->size();
    auto nfiles = 1;
    auto fsblksize = 1;
    std::size_t numItems = sz;
    std::size_t sizeItems = sizeof(char);
    outsid = sion_open(filename, "wb", &size, &nfiles,
               &chunksizes, &fsblksize,
               &globalranks, &outfp);
    if (outsid < 0) {
        fatal("open returned %d", outsid);
        return ERROR;
    }

    sion_seek_fp(outsid, 0, SION_CURRENT_BLK, SION_CURRENT_POS, &outfp);
    sion_fwrite(buffer, sizeItems, numItems, outsid);
    #ifdef DEBUG_OUTPUT
        debug("[%3d] Written %lu Items of size %lu Bytes", cl->rank(), numItems, sizeItems);
    #endif

    for (i = 1; i < cl->size(); ++i) {
        EXEC_IFFAIL(cl->send(i, &dummy, 0), fatal("send failed."); return ERROR);
        EXEC_IFFAIL(cl->recv(i, buffer, chunksizes[i]), fatal("recv failed."));
        sion_seek_fp(outsid, i, SION_CURRENT_BLK, SION_CURRENT_POS, &outfp);
        sion_fwrite(buffer, 1, chunksizes[i], outsid);
    }
    rootWatch->stop();
    printTiming("[sioncollwr]");

    rootWatch->start();
    sion_close(outsid);
    rootWatch->stop();

    sion_int64 sum = 0;
    for (i = 0; i < cl->size(); ++i)
        sum += chunksizes[i];

    printTiming("[sionclose]");
    std::fprintf(stderr,"             %-30s %lld B\n", "[sionwrite]", sum);

    delete[] chunksizes;
    delete[] globalranks;

    return SUCCESS;
}
#endif

int linktest_output_sion_funnelled(VirtualCluster* cl,
                  const struct linktest_args *args,
                  const std::vector<LinktestStats>& statsVec){
#ifdef USE_SION
    char *buffer;
    long long sz;
    double begin;

    begin = walltime();

    EXEC_IFFAIL(linktest_output_sion_collect_local_data(cl, args, statsVec, &buffer, &sz), error("linktest_output_sion_collect_local_data failed."); return ERROR);

    if (0 == cl->rank()) {
        EXEC_IFFAIL(linktest_output_sion_funnelled_root(cl, args, statsVec, buffer, sz, begin), error("linktest_output_sion_funnelled_root failed."); return ERROR);
    } else {
        long long dummy;
        EXEC_IFFAIL(cl->gather(0, &dummy, &sz, 1), fatal("gather failed."));
        EXEC_IFFAIL(cl->barrier(), fatal("barrier failed."));
        EXEC_IFFAIL(cl->recv(0, &dummy, 0), fatal("recv failed."));
        EXEC_IFFAIL(cl->send(0, buffer, sz), fatal("send failed."));
    }

    free(buffer);

    return SUCCESS;
#else
    error("This version of the benchmark was compiled without SIONlib support.");

    return ERROR;
#endif
}

#ifdef USE_SION
auto create_and_register_api(std::string name) {
    char name_c[256];
    std::strncpy(name_c, name.c_str(), 255); //strncpy length must be less than array length
    name_c[255]='\0';
    auto sion_api = sion_generic_create_api(name_c);

    sion_generic_register_create_local_commgroup_cb(sion_api, &VirtualClusterSionGenericAdapter::create_lcg_cb);
    sion_generic_register_free_local_commgroup_cb(sion_api, &VirtualClusterSionGenericAdapter::free_lcg_cb);

    sion_generic_register_barrier_cb(sion_api, &VirtualClusterSionGenericAdapter::barrier_cb);
    sion_generic_register_bcastr_cb(sion_api, &VirtualClusterSionGenericAdapter::bcastr_cb);
    sion_generic_register_gatherr_cb(sion_api, VirtualClusterSionGenericAdapter::gatherr_cb);
    sion_generic_register_scatterr_cb(sion_api, &VirtualClusterSionGenericAdapter::scatterr_cb);

    sion_generic_register_gathervr_cb(sion_api, &VirtualClusterSionGenericAdapter::gathervr_cb);
    sion_generic_register_scattervr_cb(sion_api, &VirtualClusterSionGenericAdapter::scattervr_cb);

    return sion_api;
}
#endif

/* Write in parallel using SIONlib. SIONlib will use the same transport layer as the VirtualCluster
 */
int linktest_output_sion_parallel(VirtualCluster* cl,
                  const struct linktest_args* const args,
                  const std::vector<LinktestStats>& statsVec) {
#ifdef USE_SION
    std::unique_ptr<StopwatchI> rootWatch = Stopwatchfactory::getRootWatch(cl->rank());

    auto printTiming = [rank=cl->rank(), &rootWatch](std::string step_name){
        printTimingIfRoot(rank, step_name, rootWatch->getDuration());    
    };

    auto printIfRoot = [rank=cl->rank()] (const char *__restrict format, auto ...params) {
        if( rank == 0 ) {
            info(format, params...);
        }
    };

    auto sion_api = create_and_register_api(args->virtual_cluster_implementation);
    debug("linktest_output_sion_parallel 1->barrier");
    cl->barrier();

    char* buffer;
    long long sz;
    EXEC_IFFAIL(linktest_output_sion_collect_local_data(cl, args, statsVec, &buffer, &sz), error("linktest_output_sion_collect_local_data failed."); return ERROR);
    debug("linktest_output_sion_parallel 2->barrier");
    cl->barrier();

    auto filename = args->output.c_str();
    int filenumber=0;
    int nfiles = 1;
    int fsblksize = -1;
    sion_int64 chunksize = sz;
    char* newfname;
    FILE* fp;
    int grank=VirtualClusterSionGenericAdapter::global_rank(*cl);
    int lrank=VirtualClusterSionGenericAdapter::local_rank(*cl);
    int gsize=VirtualClusterSionGenericAdapter::global_size(*cl);
    int lsize=VirtualClusterSionGenericAdapter::local_size(*cl);

    rootWatch->start();
    auto file_handle = sion_generic_paropen(
        sion_api, 
        filename, 
        "wb", 
        &chunksize,
        &fsblksize,
        cl, //gcommgroup
        grank, //global rank
        gsize, //global comm size
        &filenumber, //filenumber
        &nfiles, //numfiles
        &lrank, //local rank
        &lsize,//lsoze
        &fp, //fileptr
        &newfname //newfname
    );
    debug("linktest_output_sion_parallel 3->barrier");
    cl->barrier();
    rootWatch->stop();
    printTiming("[sionopen]");

    rootWatch->start();
    auto bwrote = sion_fwrite(buffer, 1, chunksize, file_handle);
    rootWatch->stop();
    printTiming("[sionwrite]");

    if (static_cast<sion_int64>(bwrote) != chunksize) {
        fatal("problems writing data to sionfile on rank %d, buffer[%lu] ...",
              cl->rank(), chunksize);
        return ERROR;
    }
    sion_int64 sum = 0;
    cl->reduce(0, &sum, &chunksize, 1, vcluster_reduce_op::SUM_INT);

    printIfRoot("OUTPUT: writing data to file %-30s (%sB)\n",
            filename, 
            UnitPrefix::IEC_prefix(sum, 9).get());

    rootWatch->start();
    sion_generic_parclose(file_handle);
    rootWatch->stop();
    printTiming("[sionclose]");

    
    return SUCCESS;
#else
    error("This version of the benchmark was compiled without SIONlib support.");

    return ERROR;
#endif
}
