/****************************************************************************
**  LinkTest                                                               **
*****************************************************************************
**  Copyright (c) 2008-2022                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/
#include "slow_pairs.h"
#include "config.h"
#include "compiler.h"
#include "error.h"
#include "vcluster.h"
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <cassert>
#include <cmath>
#include <algorithm>
#include <numeric>
#include <memory>

/* The C++ 2020 Standard implements bit operations which map to processor specific
 * instructions. These, however, are not available in prior versions of C++ so here is
 * a workaround.
 */
#if(__cplusplus<=201703L)
	/*Check if used C++ Standard is 2017 or earlier.
	 *Does not work with C++ Standard 2014 or earlier.
	 *std::size_t is missing from limits.
	 *Does not work with older msvc compilers.
	 *They never updated the macro between versions.
	 */
	#include <climits> //For CHAR_BIT
	namespace std{
		std::size_t bitmask(std::size_t i){
			return (std::size_t)(std::size_t{1}<<i);
		}
		std::size_t countl_zero(std::size_t x){
			std::size_t out=0;
			for(std::size_t i=CHAR_BIT*sizeof(std::size_t)-1;i>=0;--i,++out)if(0!=(x&bitmask(i)))break;
			return out;
		}
		std::size_t countr_zero(std::size_t x){
			std::size_t out=0;
			for(std::size_t i=0;i< CHAR_BIT*sizeof(std::size_t);++i,++out)if(0!=(x&bitmask(i)))break;
			return out;
		}
		std::size_t bit_width(std::size_t x){
			return (std::numeric_limits<std::size_t>::digits-countl_zero(x));
		}
		std::size_t bit_floor(std::size_t x){
			return (x!=0)?std::size_t{1}<<(bit_width(x)-1):0;
		}
	}
#else
	#include <bit>
#endif


static int compare_double(double p, double q){
	if(std::isnan(p)){
		if(std::isnan(q)) return true;
		else              return false;
	}
	if(std::isnan(q))     return true;
	if(q<p)               return true;
	else if(q>p)          return false;
	else                  return false;
}

void recieve_and_merge_data(std::unique_ptr<VirtualCluster> &cl,const std::size_t partner,std::size_t num_largest,std::size_t base_buffer_size,
                            std::vector<int> &slow_from,std::vector<int> &slow_to,std::vector<double> &slow_time){

	const unsigned int rank=cl->rank();
	const unsigned int size=cl->size();

	/* Determine Data Amount */
	std::size_t partner_buffer_size=2*partner-rank>size?size-partner:partner-rank;
	partner_buffer_size*=base_buffer_size;
	if(partner_buffer_size>num_largest) partner_buffer_size=num_largest;

	/* Recieve Data */
	std::vector<int>    alt_from;alt_from.resize(partner_buffer_size);
	std::vector<int>    alt___to;alt___to.resize(partner_buffer_size);
	std::vector<double> alt_time;alt_time.resize(partner_buffer_size);
	cl->recv(partner,alt_from.data(),partner_buffer_size);
	cl->recv(partner,alt___to.data(),partner_buffer_size);
	cl->recv(partner,alt_time.data(),partner_buffer_size);

	/* Merge & Sort */
	//TODO: These are full copies, only need partial copies
	std::vector<int>    old_from(slow_from);
	std::vector<int>    old___to(slow_to);
	std::vector<double> old_time(slow_time);
	std::size_t my______buffer_size=slow_time.size();
	std::size_t my_new__buffer_size=my______buffer_size+partner_buffer_size;
	if(my_new__buffer_size>num_largest)my_new__buffer_size=num_largest;
	slow_from.resize(my_new__buffer_size);
	slow_to.resize(my_new__buffer_size);
	slow_time.resize(my_new__buffer_size);
	std::size_t k=0;
	std::size_t l=0;
	//TODO: Is it more efficient to use two index vectors?
	for(std::size_t j=0;j<my_new__buffer_size;j++){
		if(alt_time[k]>old_time[l]){
			slow_from[j]=alt_from[k];
			slow_to[j]=alt___to[k];
			slow_time[j]=alt_time[k];
			if(++k==partner_buffer_size){
				//Copy the rest of the old buffer into new buffer
				const std::size_t n=my_new__buffer_size-(++j);
				memcpy(&slow_from[j],&old_from[l],n*sizeof(int)   );
				memcpy(&slow_to[j],&old___to[l],n*sizeof(int)   );
				memcpy(&slow_time[j],&old_time[l],n*sizeof(double));
				break;
			}
		}else{
			slow_from[j]=old_from[l];
			slow_to[j]=old___to[l];
			slow_time[j]=old_time[l];
			if(++l==my______buffer_size){
				//Copy the rest of the partner buffer into new buffer
				const std::size_t n=my_new__buffer_size-(++j);
				memcpy(&slow_from[j],&alt_from[k],n*sizeof(int)   );
				memcpy(&slow_to[j],&alt___to[k],n*sizeof(int)   );
				memcpy(&slow_time[j],&alt_time[k],n*sizeof(double));
				break;
			}
		}
	}
}

int find_and_sort_slowest_connections(std::unique_ptr<VirtualCluster> &cl, std::size_t num_slowest,
                                      std::vector<int>       from,std::vector<int>         to,std::vector<double>       time,
                                      std::vector<int> &slow_from,std::vector<int> &slow_to,std::vector<double> &slow_time){

	std::size_t rank=(std::size_t)cl->rank();
	std::size_t max_rank=(std::size_t)(cl->size()-1);

	/* Allocate & Fill From Vector */
	// TODO: Creating a from vector now is not necessary, after the 1st transfer
	//       it is still clear that all the transferred times must have a from
	//       corresponding to the send. Removing this will slightly speed up the
	//       algorithm.
	const std::size_t base_buffer_size=num_slowest>time.size()?time.size():num_slowest;
	slow_from.resize(base_buffer_size);
	std::fill(slow_from.begin(),slow_from.end(),rank);

	/* Resize Other Slow Vectors */
	slow_to.resize(base_buffer_size);
	slow_time.resize(base_buffer_size);

	/* Sort `num_slowest` Largest Times */
	// A permutor is used here as it is faster for sorting three or more arrays
	// in the same fashion. This also applies to partial sorts.
	// TODO: Is this really true?
	std::vector<std::size_t> indices;indices.resize(time.size());
	std::iota(indices.begin(),indices.end(),0);
	std::nth_element(indices.begin(),indices.begin()+base_buffer_size,indices.end(),
	                 [&](std::size_t i,std::size_t j){return compare_double(time[i],time[j]);});
	std::vector<std::size_t> permutor;permutor.resize(base_buffer_size);
	std::partial_sort_copy(indices.begin(),indices.end(),permutor.begin(),permutor.end(),
	                       [&](std::size_t i,std::size_t j){return compare_double(time[i],time[j]);});
	std::transform(permutor.begin(),permutor.end(),slow_time.begin(),[&](std::size_t i){return time[i];});
	std::transform(permutor.begin(),permutor.end(),slow_to.begin(),[&](std::size_t i){return   to[i];});

	/* Begin Merge & Sort */
	std::size_t partner=1;
	if(rank==0){
		std::size_t num_branches=std::bit_width(max_rank);
		for(std::size_t i=0;i<num_branches;i++){
			recieve_and_merge_data(cl,partner,num_slowest,base_buffer_size,slow_from,slow_to,slow_time);
			partner*=2;
		}
	}else{
		std::size_t num_branches=std::countr_zero(rank);
		if(num_branches&&rank>=std::bit_floor(max_rank)){
			std::size_t tmp=std::bit_width(max_rank-rank);
			num_branches=(num_branches>tmp?tmp:num_branches);
		}
		std::size_t partner=1;
		for(std::size_t i=0;i<num_branches;i++){
			recieve_and_merge_data(cl,rank+partner,num_slowest,base_buffer_size,slow_from,slow_to,slow_time);
			partner*=2;
		}
		partner=rank&(rank-1);
		cl->send(partner,slow_from.data(),slow_from.size());
		cl->send(partner,slow_to.data(),slow_to.size());
		cl->send(partner,slow_time.data(),slow_time.size());
		slow_from.resize(num_slowest);
		slow_to.resize(num_slowest);
		slow_time.resize(num_slowest);
	}

	/* Broadcast Slowest Connections */
	cl->bcast(0,slow_from.data(),slow_from.size());
	cl->bcast(0,slow_to.data(),slow_from.size());

	return SUCCESS;
}
