/****************************************************************************
**  LinkTest                                                               **
*****************************************************************************
**  Copyright (c) 2008-2022                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/
#include "utils.h"
#include "config.h"
#include "compiler.h"
#include "error.h"
#include "cmdline.h"
#include "vcluster.h"
#include <cassert>
#include <cstdlib>
#include <cstdio>
#include <cstring>


/* collective_print_many_pes has a zero memory overhead and thus is
 * useful for large core counts even though it might be very slow due
 * to serialization.
 */
static int collective_print_many_pes(VirtualCluster* cl, char* str,
                     int len)
{
    int ret;
    int i;
    int dummy;

    if (0 == cl->rank()) {
        if (likely(std::strlen(str) > 0))
            fputs(str, stderr);

        for (i = 1; i < cl->size(); ++i) {
            ret = cl->send(i, &dummy, 0);
            if (unlikely(ret)) {
                /* We are terminating here to avoid
                 * a deadlock situation. */
                fatal("send failed.");
                return ERROR;
            }

            EXEC_IFFAIL(cl->recv(i, str, len), fatal("recv failed."));

            if (likely(std::strlen(str) > 0))
                fputs(str, stderr);
        }
    } else {
        EXEC_IFFAIL(cl->recv(0, &dummy, 0), fatal("recv failed."));
        EXEC_IFFAIL(cl->send(0, str, len), fatal("send failed."));
    }

    return SUCCESS;
}

/* Fast collective print that requires a buffer whose size increases linearly
 * with the number of processing elements.
 */
static int collective_print_few_pes(VirtualCluster* cl, char* str, int len){
// TODO: Make this work with other compilers by checking for compiler
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wmaybe-uninitialized"
    char* buf;

    if (0 == cl->rank()) {
        buf = new char[len*cl->size()];
    }

    /* Note: If the `new` on rank 0 fails the process will fail and we
     * rely on the resource management system to cleanup the other 
     * processes.
     */

    EXEC_IFFAIL(cl->gather(0, buf, str, len), error("gather failed."); return ERROR);

    if (0 == cl->rank()) {
        for (int i = 0; i < cl->size(); ++i ) {
            str = &buf[len*i];

            if (likely(std::strlen(str) > 0))
                fputs(str, stderr);
        }

        delete[] buf;
    }
#pragma GCC diagnostic pop

    return SUCCESS;
}

int collective_print(VirtualCluster* cl, char* str, int len)
{
    if (cl->size() > 64)
        return collective_print_many_pes(cl, str, len);
    else
        return collective_print_few_pes(cl, str, len);
}

double lat_to_bw(double time, const struct linktest_args* args)
{
    if (unlikely(time < 1e-16))
        return 0.0l;
    else
        return (args->len_msg * 1.0)/(time * 1024.0 * 1024.0);
}

template < class T >
std::ostream& operator<< (std::ostream& os, const std::vector<T>& v) 
{
    os << "[";
    for (auto ii = v.begin(); ii != v.end(); ++ii)
    {
        os << " " << *ii;
    }
    os << "]";
    return os;
}

std::ostream& operator<< (std::ostream& os, const std::vector<int>& v) 
{
    os << "[";
    for (auto ii = v.begin(); ii != v.end(); ++ii)
    {
        os << " " << *ii;
    }
    os << "]";
    return os;
}
