/****************************************************************************
**  LinkTest                                                               **
*****************************************************************************
**  Copyright (c) 2008-2022                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/
#ifndef LINKTEST_VCLUSTER_H
#define LINKTEST_VCLUSTER_H

#include "compiler.h"
#include "memory.h"
#include "memory_multi.h"
#include "stats.h"
#include <string>
#include <memory>

struct linktest_args;

/* A virtual view of our cluster. The VirtualCluster
 * structure is the base class for different derived
 * classes that implement communication and I/O
 * functionality targeted at different hardware
 * architectures and software stacks.
 */
class VirtualCluster{
public:
    VirtualCluster(const std::string name);
    virtual ~VirtualCluster();

    /*! \brief Initialize the communication library.
    *
    * Sets up communication information and prepares the communication layer for data transport.
    * After initialization all other function inkl rank(),size() etc. are expected to work.
    */
    virtual int init() = 0;

    /*! \brief Finalizez the communication library.
    *
    * Cleans up communication information and frees ressources.
    * After finalization all other function inkl rank(),size() etc. are expected to work.
    */
    virtual int finalize() = 0;

    /*! \brief returns the virtual cluster rank
    *
    * Return a unique number representing the calling processing element.
    * The number must be between zero and size-1.
    */
    virtual int rank() = 0;

    /*! \brief returns the virtual cluster size
    *
    * Number of processing elements in this virtual cluster.
    */
    virtual int size() = 0;

    
            std::string            name()           ;
    const   std::string&           nameRef()        ;
            std::int32_t           nameSize()       ;
    const   std::int32_t&          nameSizeRef()    ;
            std::string            hostname()       ;
    const   std::string&           hostnameRef()    ;
            std::int32_t           hostnameSize()   ;
    const   std::int32_t&          hostnameSizeRef();
            std::int32_t           cpuID()          ;
    const   std::int32_t&          cpuIDRef()       ;
                 int               numHostRanks()   ;
                 int               numLocalRanks()  ;
                 int               hostRank()       ;
                 int               localRank()      ;
            std::shared_ptr<int[]> hostLocalRanks() ;

    struct {
        std::vector<std::string> hostForRank;
        std::vector<std::vector<int>> ranksForHost;
    } hostnamesAndRanks;

    /*! \brief send wrapped data in buf to rank dst (communication layer undefined)
    *
    * The data may or may not be routed through the communication layer of this cluster
    * This function should therefore not be used in the kernel
    */
    virtual int send(int dst, MemoryBuffer& buf) = 0;

    /*! \brief receive wrapped data from rank src and store it in buf (communication layer undefined)
    *
    * The data may or may not be routed through the communication layer of this cluster
    * This function should therefore not be used in the kernel
    */
    virtual int recv(int src, MemoryBuffer& buf) = 0;

    /*! \brief wait until all ranks in this virtual cluster have called barrier
    *
    * The synchronization may or may not happen through the communication layer of this cluster
    * This function should therefore not be used in the kernel
    */
    virtual int barrier();
    
    /*! \brief broadcast wrapped data in buf (communication layer undefined)
    *
    * The data may or may not be routed through the communication layer of this cluster
    * This function should therefore not be used in the kernel
    */
    virtual int bcast(int root, MemoryBuffer& buf);
    
    /*! \brief gather wrapped data at rank root  (communication layer undefined)
    *
    * The data may or may not be routed through the communication layer of this cluster
    * This function should therefore not be used in the kernel
    */
    virtual int gather(int root, MemoryBuffer& dst, MemoryBuffer& src);
    
    /*! \brief scatter wrapped data from rank root  (communication layer undefined)
    *
    * The data may or may not be routed through the communication layer of this cluster
    * This function should therefore not be used in the kernel
    */
    virtual int scatter(int root, MemoryBuffer& dst, MemoryBuffer& src);
    
    /*! \brief reduce (fold) wrapped data from at rank root (communication layer undefined)
    *
    * The data may or may not be routed through the communication layer of this cluster
    * This function should therefore not be used in the kernel
    */
    virtual int reduce(int root, MemoryBuffer& dst, MemoryBuffer& src, int op);

    /*! \brief send data in buf to rank dst (communication layer undefined)
    *
    * The data may or may not be routed through the communication layer of this cluster
    * This function should therefore not be used in the kernel
    */
    template<typename T>
    int send(const int dst,const T* const vals,const int len);

    /*! \brief receive data from rank src and store it in buf (communication layer undefined)
    *
    * The data may or may not be routed through the communication layer of this cluster
    * This function should therefore not be used in the kernel
    */
    template<typename T>
    int recv(const int src, T* const vals,const int len);
    
    
    /*! \brief broadcast data in buf (communication layer undefined)
    *
    * The data may or may not be routed through the communication layer of this cluster
    * This function should therefore not be used in the kernel
    */
    template<typename T>
    int bcast(int root, T* vals, int len);
    
    /*! \brief gather data at rank root  (communication layer undefined)
    *
    * The data may or may not be routed through the communication layer of this cluster
    * This function should therefore not be used in the kernel
    */
    template<typename T>
    int gather(int root, T* dst, T* src, int len);
    
    /*! \brief scatter data from rank root  (communication layer undefined)
    *
    * The data may or may not be routed through the communication layer of this cluster
    * This function should therefore not be used in the kernel
    */
    template<typename T>
    int scatter(int root, T* dst, T* src, int len);
    
    /*! \brief reduce (fold) data from at rank root (communication layer undefined)
    *
    * The data may or may not be routed through the communication layer of this cluster
    * This function should therefore not be used in the kernel
    */
    template<typename T>
    int reduce(int root, T* dst, T* src, int len, int op);

    /*! \brief semidirectional kernel
    *
    * Sends a message and receives a message as confirmation
    * from    to
    *   |────►|
    *   | buf |
    *   |◄────|
    */
    virtual int kpingpong(const int from, const int to, 
                          MemoryBuffer& buf, 
                          const int num_msg, double* const timing);

     /*! \brief bidirectional kernel
    *
    * Sends a message and receives a message at te same time.
    * from    to
    *   |◄───►|
    *   | buf |
    */
    virtual int kbipingpong(int from, int to,
                MemoryBuffer& buf1, MemoryBuffer& buf2,
                int num_msg, double* timing);

    /*! \brief unidirectional kernel
    *
    * Sends multiple messages with only one confirmation at the end.
    * from    to
    *   |────►|
    *   | buf |
    *
    * \param doBarrier if true synchronizes all ranks before the time measure starts
    */
    virtual int kUniDir(const int from,const int to,
                        MemoryBuffer& buf1, MemoryBuffer& buf2,
                        const int num_msg, double* const timing,
                        const bool doBarrier);

                        
    virtual int kUniDirMultiBuf(const int from, const int to,
                                MemoryBufferMulti& buf_multi, MemoryBuffer& buf2,
                                const int num_msg, double* const timing,
                                const bool doBarrier);
    virtual int kUniDirLimitedMultiBuf(const int from, const int to,
                                        MemoryBufferMulti& buf_multi, MemoryBuffer& buf2,
                                        const int num_msg, double* const timing,
                                        const bool doBarrier);

    /* Performs all-to-all communication with the buffer <buf> <num_msg> times
        * No warmup
        */
    virtual int kalltoall(MemoryBuffer& buf, const int len_msg, const int num_msg, double* const timing);

    /* Amount of extra memory that the clients should be allocating for the kpingpong
        * kernel. This space may be used by the implementation of the benchmark
        */
    virtual std::uint64_t kpingpong_minimal_buffer_overhead() const;

    // Pingpong kernel with warmup.
    int linktest_kpingpong(int from, int to,
                    MemoryBuffer& buf,
                    const struct linktest_args* args,
                    double* time);


    int linktest_kUniDir(const int from,const int to,
                            MemoryBuffer& buf1, MemoryBuffer& buf2,
                            const struct linktest_args* const args,
                            double* time, const int doBarrier);
    int linktest_kUniDirMultiBuf(const int from,const int to,
                                    MemoryBufferMulti& buf_multi, MemoryBuffer& buf2,
                                    const struct linktest_args* const args,
                                    double* const time, const int doBarrier);
    int linktest_kUniDirLimitedMultiBuf(const int from,const int to,
                                        MemoryBufferMulti& buf_multi, MemoryBuffer& buf2,
                                        const struct linktest_args* const args,
                                        double* const time, const int doBarrier);

    // All-to-all kernel with warmup.
    int linktest_kalltoall(MemoryBuffer& buf,
                    const struct linktest_args* args,
                    double* time);

    // Bidirectional pingpong kernel with warmup.
    int linktest_kbipingpong(int from, int to,
                    MemoryBuffer& buf1,
                    MemoryBuffer& buf2,
                    const struct linktest_args* args,
                    double* time);

    // Determines host ranks and host-local ranks
    void getHostAndLocalRank();

    /* I/O operations.
        *
        * - write_parallel: Write in parallel from all processing elements.
        * - write_funnelled: Funnel the I/O through the first processing element.
        *
        * Implementations must provide one of the two functions and may provide
        * both.
        */
    virtual int write_parallel(const linktest_args* args, const std::vector<LinktestStats>& statsVec);
    virtual int write_funnelled(const linktest_args* args, const std::vector<LinktestStats>& statsVec);

    /** \brief List of supported transport protocol (build at compile time) */
    static const std::vector<std::string> impls;
    /** \return name of the requested transport protocol aka the virtual cluster imlplementation */
    static const std::string& get_vcluster_impl_name(char** argv, const std::string& name);
    /** \return pointer to requested VirtualCluster implementation */
    static VirtualCluster* factory(const std::string& name);

private:
          std::string            name_;
    const std::int32_t           nameSize_;
          std::string            hostname_;
    const std::int32_t           hostnameSize_;
    const std::int32_t           cpuID_;
               int               numHostRanks_;
               int               hostRank_;
               int               localRank_;
               int               numLocalRanks_;
          std::shared_ptr<int[]> hostLocalRanks_;
};

/* VirtualClusterWithHelper 
 * delegates calls to rank(), size(), send() and recv() to another VirtualCluster (the helper)
 * executes calls to benchmark kernels directly
 *
 * Since a full implementation of send()/recv() logic on top of some transport layer is
 * more complicated than the logic required for the implementation of the communication
 * in our kernels it make sense to use a different VirtualCluster for the management 
 * communication than for the actual benchmark.
 */
class VirtualClusterWithHelper : public VirtualCluster {

public:
    VirtualClusterWithHelper(const std::string& name);

    int                    rank()                           override;
    int                    size()                           override;
    int                    numHostRanks()                           ;
    int                    numLocalRanks()                          ;
    int                    hostRank()                               ;
    int                    localRank()                              ;
    std::shared_ptr<int[]> hostLocalRanks()                         ;
    int                    send(int dst, MemoryBuffer& buf) override;
    int                    recv(int src, MemoryBuffer& buf) override;

protected:
    void set_helper_pointer(VirtualCluster* helper);
    VirtualCluster*         helper_;

};


/* Supported reduction operations.
 */
enum vcluster_reduce_op {
    MIN_INT,
    MAX_INT,
    SUM_INT,

    MIN_DOUBLE,
    MAX_DOUBLE,
    SUM_DOUBLE
};

template<typename T>
int VirtualCluster::send(const int dst,const T* const vals,const int len){
    auto tmp = MemoryBuffer::wrap<T>(const_cast<T*>(vals), len, AddressSpace::ID::Local);

    return send(dst, tmp);
}

template<typename T>
int VirtualCluster::recv(const int src, T* const vals,const int len){
    auto tmp = MemoryBuffer::wrap<T>(const_cast<T*>(vals), len, AddressSpace::ID::Local);

    return recv(src, tmp);
}

template<typename T>
int VirtualCluster::bcast(int root, T* vals, int len)
{
    auto tmp = MemoryBuffer::wrap<T>(vals, len, AddressSpace::ID::Local);

    return bcast(root, tmp);
}

template<typename T>
int VirtualCluster::gather(int root, T* dst, T* src, int len)
{
    auto tmp1 = MemoryBuffer::wrap<T>(dst, len, AddressSpace::ID::Local);
    auto tmp2 = MemoryBuffer::wrap<T>(src, len, AddressSpace::ID::Local);

    return gather(root, tmp1, tmp2);
}

template<typename T>
int VirtualCluster::scatter(int root, T* dst, T* src, int len)
{
    auto tmp1 = MemoryBuffer::wrap<T>(dst, len, AddressSpace::ID::Local);
    auto tmp2 = MemoryBuffer::wrap<T>(src, len, AddressSpace::ID::Local);

    return scatter(root, tmp1, tmp2);
}

template<typename T>
int VirtualCluster::reduce(int root, T* dst, T* src, int len, int op)
{
    auto tmp1 = MemoryBuffer::wrap<T>(dst, len, AddressSpace::ID::Local);
    auto tmp2 = MemoryBuffer::wrap<T>(src, len, AddressSpace::ID::Local);

    return reduce(root, tmp1, tmp2, op);
}

#endif
