/****************************************************************************
**  LinkTest                                                               **
*****************************************************************************
**  Copyright (c) 2008-2022                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/
#include "config.h"
#include "compiler.h"
#include "timing.h"
#include "vcluster.h"
#include "vcluster_cuda.h"
#include "vcluster_helper.h"
#include "environ.h"
#include "error.h"
#include "output_sion.h"
#include "pmi.h"
#if HAVE_VCLUSTER_CUDA == 1
#include "gpu_nvidia.h"
#endif
#include <cassert>
#include <stdexcept>
#include <unordered_map>

#undef  CUDA_ENVIRON_PREFIX
#define CUDA_ENVIRON_PREFIX LINKTEST_ENVIRON_PREFIX "CUDA_"


namespace linktest
{
namespace cuda
{

// Environment variables
static struct environ
{
    const char      *helper;
    int             debug;
} env = {
    .helper = "tcp",
    .debug  = 0
};

class Connection
{
public:
    Connection(VirtualClusterImpl* impl_cl, VirtualCluster* comm, int other);
    ~Connection();

    enum {
        kBuf0           = 0,
        kBuf1           = 1,
        kNumBuffers     = 2
    };

    int     setup(MemoryBuffer* buf0, MemoryBuffer* buf1 = nullptr);
    std::pair<CUevent, CUevent>             events;
    /* For simplicity we associate with each connection a pair of
     * local and remote buffers. This is not sufficient for a general-
     * purpose communication library but suffices for the benchmark.
     */
    std::pair<MemoryBuffer*, MemoryBuffer*> buf[kNumBuffers];

private:
    VirtualClusterImpl*     impl_cl_;
    VirtualCluster*         comm_;
    int                     other_;

    struct ExchangeData
    {
        CUipcEventHandle        ev;
        CUipcMemHandle          mem[kNumBuffers];
    };
    int     exchange(ExchangeData* local, ExchangeData* remote);

};

class VirtualClusterImpl
{

public:
    VirtualClusterImpl(VirtualCluster* owner);

    int init();
    int finalize();
    int kpingpong(const int from, const int to, 
                  MemoryBuffer& buf,
                  const int num_msg, double* const timing
                 );
    int kUniDir(const int from, const int to,
                MemoryBuffer& buf1, MemoryBuffer& buf2,
                const int num_msg, double* const timing,
                const bool doBarrier
               );
    int kUniDirMultiBuf(const int from,const int to,
                        MemoryBufferMulti& buf_multi, MemoryBuffer& buf2,
                        const int num_msg, double* const timing,
                        const bool doBarrier
                       );
    int kUniDirLimitedMultiBuf(const int from,const int to,
                               MemoryBufferMulti& buf_multi, MemoryBuffer& buf2,
                               const int num_msg, double* const timing,
                               const bool doBarrier
                              );
    int kbipingpong(const int from, const int to,
                    MemoryBuffer& buf1, MemoryBuffer& buf2,
                    const int num_msg, double* const timing);

private:
    VirtualCluster*                         owner_;
    GpuContext*                             gpuctx_;
    std::unordered_map<int, Connection*>    conns_;
    int kpingpong_poll();
    int kpingpong_isend(Connection& conn, int bufi);
    int kpingpong_send(Connection& conn, int bufi);
    int kpingpong_recv(Connection& conn);
    Connection*     find_or_create_connection(int other, 
                          MemoryBuffer* buf1, 
                          MemoryBuffer* buf2);
    friend class Connection;

};

static int read_environ()
{
    read_environ_str(CUDA_ENVIRON_PREFIX "HELPER", &env.helper);
    read_environ_int(CUDA_ENVIRON_PREFIX "DEBUG", &env.debug);
    return SUCCESS;
}

Connection::Connection(VirtualClusterImpl* impl_cl, VirtualCluster* comm, int other)
: impl_cl_(impl_cl), comm_(comm), other_(other)
{
    for (auto i = 0; i < Connection::kNumBuffers; ++i) {
        buf[i].first  = nullptr;
        buf[i].second = nullptr;
    }
}

Connection::~Connection()
{
    impl_cl_->gpuctx_->execute([this] {
        auto err = cuEventDestroy(events.first);
        if (unlikely(CUDA_SUCCESS != err)) {
            report_cuda_error("cuEventDestroy", err);
            return ERROR;
        }
        err = cuEventDestroy(events.second);
        if (unlikely(CUDA_SUCCESS != err)) {
            report_cuda_error("cuEventDestroy", err);
            return ERROR;
        }

        for (auto i = 0; i < Connection::kNumBuffers; ++i) {
            if (nullptr == buf[i].first)
                continue;

            err = cuIpcCloseMemHandle((CUdeviceptr)buf[i].second->p());
            if (unlikely(CUDA_SUCCESS != err)) {
                report_cuda_error("cuIpcCloseMemHandle", err);
                return ERROR;
            }

            delete buf[i].second;
        }

        return SUCCESS;
    });
}

int Connection::setup(MemoryBuffer* buf0, MemoryBuffer* buf1)
{
    ExchangeData local, remote;

    buf[Connection::kBuf0].first = buf0;
    buf[Connection::kBuf1].first = buf1;

    auto err = impl_cl_->gpuctx_->execute([this, &local] {
        auto err = cuEventCreate(&events.first, 
                     CU_EVENT_BLOCKING_SYNC | 
                     CU_EVENT_DISABLE_TIMING | 
                     CU_EVENT_INTERPROCESS);
        if (unlikely(CUDA_SUCCESS != err)) {
            report_cuda_error("cuEventCreate", err);
            return ERROR;
        }

        err = cuIpcGetEventHandle(&local.ev, events.first);
        if (unlikely(CUDA_SUCCESS != err)) {
            report_cuda_error("cuIpcGetEventHandle", err);
            return ERROR;
        }

        for (auto i = 0; i < Connection::kNumBuffers; ++i) {
            if (nullptr == buf[i].first)
                continue;

            err = cuIpcGetMemHandle(&local.mem[i],(CUdeviceptr)buf[i].first->p());
            if (unlikely(CUDA_SUCCESS != err)) {
                report_cuda_error("cuIpcGetMemHandle", err);
                return ERROR;
            }
        }

        return SUCCESS;
    });
    if (unlikely(err))
        return ERROR;

    EXEC_NOFAIL(exchange(&local, &remote));

    return impl_cl_->gpuctx_->execute([this, &remote] {
        auto err = cuIpcOpenEventHandle(&events.second, remote.ev);
        if (unlikely(CUDA_SUCCESS != err)) {
            report_cuda_error("cuIpcOpenEventHandle", err);
            return ERROR;
        }

        for (auto i = 0; i < Connection::kNumBuffers; ++i) {
            if (nullptr == buf[i].first)
                continue;

            CUdeviceptr p;

            err = cuIpcOpenMemHandle(&p, remote.mem[i],
                         CU_IPC_MEM_LAZY_ENABLE_PEER_ACCESS);
            if (unlikely(CUDA_SUCCESS != err)) {
                report_cuda_error("cuIpcGetMemHandle", err);
                return ERROR;
            }

            buf[i].second = new MemoryBuffer((void*)p,
                                             (CUdeviceptr)buf[i].first->len(),
                                             AddressSpace::ID::CudaDeviceRemote);
        }

        return SUCCESS;
    });
}

int Connection::exchange(ExchangeData* local, ExchangeData* remote)
{
    if (comm_->rank() < other_) {
        EXEC_IFFAIL(comm_->send(other_, local, 1), error("send failed."); return ERROR);

        EXEC_IFFAIL(comm_->recv(other_, remote, 1), error("recv failed."); return ERROR);
    } else {
        EXEC_IFFAIL(comm_->recv(other_, remote, 1), error("recv failed."); return ERROR);

        EXEC_IFFAIL(comm_->send(other_, local, 1), error("send failed."); return ERROR);
    }

    return SUCCESS;
}

VirtualClusterImpl::VirtualClusterImpl(VirtualCluster* owner)
: owner_(owner)
{
}

int VirtualClusterImpl::init()
{
    gpuctx_ = GpuContext::singleton();
    if (!gpuctx_) {
        warn("GpuContext::singleton() returned nullptr. Retrying later.");
    }

    return SUCCESS;
}

int VirtualClusterImpl::finalize()
{
    for(auto& conn: conns_) {
        delete conn.second;
        conn.second = nullptr;
    }

    return SUCCESS;
}

int VirtualClusterImpl::kpingpong(const int from, const int to, MemoryBuffer& buf, 
                                  const int num_msg, double* const timing){
    double tv = 0;

    if ((owner_->rank() != from) and (owner_->rank() != to)) {
        throw std::invalid_argument("Invalid rank");
    }

    /* The `GpuContext` is retrieved in the `init()` function but may not
     * be initialized at this point due to the order of function calls.
     * Retrieve it again in this case. This will be resolved with #7.
     */
    if (unlikely(nullptr == gpuctx_)) {
        gpuctx_ = GpuContext::singleton();
    }
    assert(gpuctx_);

    auto other = (owner_->rank() == from) ? to : from;
    auto& conn = *find_or_create_connection(other, &buf, nullptr);

    if(owner_->rank() == from) {
        tv = walltime();

        for (auto i = 0; i < num_msg; ++i) {
            EXEC_NOFAIL(kpingpong_send(conn, Connection::kBuf0));

            EXEC_NOFAIL(kpingpong_recv(conn));
        }

        tv = walltime() - tv;
        tv = tv / (2.0*num_msg);
    } else if(owner_->rank() == to) {
        for (auto i = 0; i < num_msg; ++i) {
            EXEC_NOFAIL(kpingpong_recv(conn));

            EXEC_NOFAIL(kpingpong_send(conn, Connection::kBuf0));
        }
    } else {
        error("Invalid rank.");
        return ERROR;
    }

    if(timing)
        *timing = tv;

    return SUCCESS;
}

int VirtualClusterImpl::kUniDir(const int from, const int to,
                                MemoryBuffer& buf1, MemoryBuffer& buf2,
                                const int num_msg, double* const timing,
                                const bool doBarrier
                               ){
    fatal("not yet implemented.");
    return ERROR;
}
int VirtualClusterImpl::kUniDirMultiBuf(const int from,const int to,
                                        MemoryBufferMulti& buf_multi, MemoryBuffer& buf2,
                                        const int num_msg, double* const timing,
                                        const bool doBarrier
                                       ){
    fatal("not yet implemented.");
    return ERROR;
}
int VirtualClusterImpl::kUniDirLimitedMultiBuf(const int from,const int to,
                                               MemoryBufferMulti& buf_multi, MemoryBuffer& buf2,
                                               const int num_msg, double* const timing,
                                               const bool doBarrier
                                              ){
    fatal("not yet implemented.");
    return ERROR;
}

int VirtualClusterImpl::kbipingpong(const int from, const int to,
                                    MemoryBuffer& buf1, MemoryBuffer& buf2,
                                    const int num_msg, double* const timing){
    double tv = 0;

    gpuctx_ = GpuContext::singleton();
    assert(gpuctx_);

    if ((owner_->rank() != from) and (owner_->rank() != to)) {
        throw std::invalid_argument("Invalid rank");
    }

    /* The `GpuContext` is retrieved in the `init()` function but may not
     * be initialized at this point due to the order of function calls.
     * Retrieve it again in this case. This will be resolved with #7.
     */
    if (unlikely(nullptr == gpuctx_)) {
        gpuctx_ = GpuContext::singleton();
    }
    assert(gpuctx_);

    auto other = (owner_->rank() == from) ? to : from;
    auto& conn = *find_or_create_connection(other, &buf1, &buf2);

    // Do not touch the buffer here or you are in a race with the remote write.

    if(owner_->rank() == from) {
        tv = walltime();

        for (auto i = 0; i < num_msg; ++i) {
            EXEC_IFFAIL(kpingpong_isend(conn, Connection::kBuf0), fatal("send failed."));

            EXEC_IFFAIL(kpingpong_recv(conn), fatal("recv failed."));

            EXEC_IFFAIL(kpingpong_poll(), fatal("poll failed."));
        }

        tv = walltime() - tv;
        tv = tv / (2.0*num_msg);
    } else if(owner_->rank() == to) {
        for (auto i = 0; i < num_msg; ++i) {
            EXEC_IFFAIL(kpingpong_isend(conn, Connection::kBuf1), fatal("send failed."));

            EXEC_IFFAIL(kpingpong_recv(conn), fatal("recv failed."));

            EXEC_IFFAIL(kpingpong_poll(), fatal("poll failed."));
        }
    }

    if(timing)
        *timing = tv;

    return SUCCESS;
}

int VirtualClusterImpl::kpingpong_isend(Connection& conn, int bufi)
{
    return gpuctx_->execute([&conn, bufi] {
        auto err = cuMemcpyAsync((CUdeviceptr)conn.buf[bufi].second->p(),
                     (CUdeviceptr)conn.buf[bufi].first->p(),
                     conn.buf[bufi].first->len(), 
                     CU_STREAM_PER_THREAD);
        if (unlikely(CUDA_SUCCESS != err)) {
            report_cuda_error("cuStreamSynchronize", err);
            return ERROR;
        }

        err = cuEventRecord(conn.events.first, CU_STREAM_PER_THREAD);
        if (unlikely(CUDA_SUCCESS != err)) {
            report_cuda_error("cuEventRecord", err);
            return ERROR;
        }

        return SUCCESS;
    });
}

int VirtualClusterImpl::kpingpong_send(Connection& conn, int bufi)
{
    EXEC_NOFAIL(kpingpong_isend(conn, bufi));

    EXEC_NOFAIL(kpingpong_poll());

    /* Important: Do not change tail here since other may have already posted
     *            the send and overwritten our buffer.
     */

    return SUCCESS;
}

int VirtualClusterImpl::kpingpong_recv(Connection& conn)
{
    return gpuctx_->execute([&conn] {
        auto err = cuStreamWaitEvent(CU_STREAM_PER_THREAD, 
                         conn.events.second, 0);
        if (unlikely(CUDA_SUCCESS != err)) {
            report_cuda_error("cuEventSynchronize", err);
            return ERROR;
        }

        return SUCCESS;
    });
}

int VirtualClusterImpl::kpingpong_poll()
{
    return gpuctx_->execute([] {
        auto err = cuStreamSynchronize(CU_STREAM_PER_THREAD);
        if (unlikely(CUDA_SUCCESS != err)) {
            report_cuda_error("cuStreamSynchronize", err);
            return ERROR;
        }
        
        return SUCCESS;
    });
}

Connection* VirtualClusterImpl::find_or_create_connection(int other, 
                              MemoryBuffer* buf1, 
                              MemoryBuffer* buf2)
{
    auto p = conns_.find(other);
    if (conns_.end() == p) {
        p = conns_.insert({other, new Connection(this, owner_, other)}).first;
        p->second->setup(buf1, buf2);
    }

    return p->second;
}

}
}

VirtualClusterCUDA::VirtualClusterCUDA(const std::string& name)
: VirtualClusterWithHelper(name), 
  impl_(new linktest::cuda::VirtualClusterImpl(this))
{
}

int VirtualClusterCUDA::init()
{
    EXEC_NOFAIL(linktest::cuda::read_environ());

    set_helper_pointer(VirtualCluster::factory(linktest::cuda::env.helper));
    EXEC_NOFAIL(helper_->init());

    return impl_->init();
}

int VirtualClusterCUDA::finalize()
{
    helper_->finalize(); //Ignore error code
    delete helper_;

    return impl_->finalize();
}

int VirtualClusterCUDA::kpingpong(const int from, const int to, MemoryBuffer& buf,
                                  const int num_msg, double* const timing){
    return impl_->kpingpong(from, to, buf, num_msg, timing);
}

int VirtualClusterCUDA::kUniDir(const int from, const int to,
                                MemoryBuffer &buf1, MemoryBuffer &buf2,
                                const int num_msg, double* const timing,
                                const bool doBarrier
                               ){
    return impl_->kUniDir(from,to,buf1,buf2,num_msg,timing,doBarrier);
}
int VirtualClusterCUDA::kUniDirMultiBuf(const int from, const int to,
                                        MemoryBufferMulti &buf_multi, MemoryBuffer &buf2,
                                        const int num_msg, double* const timing,
                                        const bool doBarrier
                                       ){
    return impl_->kUniDirMultiBuf(from,to,buf_multi,buf2,num_msg,timing,doBarrier);
}
int VirtualClusterCUDA::kUniDirLimitedMultiBuf(const int from, const int to,
                                               MemoryBufferMulti &buf_multi, MemoryBuffer &buf2,
                                               const int num_msg, double* const timing,
                                               const bool doBarrier
                                              ){
    return impl_->kUniDirLimitedMultiBuf(from,to,buf_multi,buf2,num_msg,timing,doBarrier);
}

int VirtualClusterCUDA::kbipingpong(const int from, const int to,
                                    MemoryBuffer& buf1, MemoryBuffer& buf2,
                                    const int num_msg, double* const timing){
    return impl_->kbipingpong(from, to, buf1, buf2, num_msg, timing);
}

int VirtualClusterCUDA::write_parallel(const linktest_args* args, const std::vector<LinktestStats>& statsVec)
{
    fatal("not yet implemented.");
    return ERROR;
}

int VirtualClusterCUDA::write_funnelled(const linktest_args* args, const std::vector<LinktestStats>& statsVec)
{
    return linktest_output_sion_funnelled(this, args, statsVec);
}
