/****************************************************************************
**  LinkTest                                                               **
*****************************************************************************
**  Copyright (c) 2008-2022                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/
#include "vcluster_mpi.h"
#include "format_units.h"
#include "environ.h"
#include "error.h"
#include "timing.h"
#include "stopwatch.h"
#include "output_sion.h"
#include "cmdline.h"
#include <vector>
#include <cstdlib>

#ifdef USE_SION
    #include <cstdio>
    #include "sion.h"
#endif

#undef  MPI_ENVIRON_PREFIX
#define MPI_ENVIRON_PREFIX LINKTEST_ENVIRON_PREFIX "MPI_"

/* Convert the return value of an MPI call to
 * zero (success) and non-zero (error).
 */
static constexpr inline int _mpi_(int ret)
{
    return (MPI_SUCCESS == ret) ? SUCCESS : ERROR;
}

static constexpr int translate_op_to_mpi_type_and_mpi_op(int op,
                           MPI_Datatype *type, int *sz,
                           MPI_Op *oq)
{
    switch(op) {
    case MIN_INT:
    case MAX_INT:
    case SUM_INT:
        *type = MPI_INT;
        *sz = sizeof(int);
        break;
    case MIN_DOUBLE:
    case MAX_DOUBLE:
    case SUM_DOUBLE:
        *type = MPI_DOUBLE;
        *sz = sizeof(double);
        break;
    default:
        return ERROR;
    }

    switch(op) {
    case MIN_INT:
    case MIN_DOUBLE:
        *oq = MPI_MIN;
        break;
    case MAX_INT:
    case MAX_DOUBLE:
        *oq = MPI_MAX;
        break;
    case SUM_INT:
    case SUM_DOUBLE:
        *oq = MPI_SUM;
        break;
    default:
        return ERROR;
    }

    return SUCCESS;
}


VirtualClusterMPI::VirtualClusterMPI(const std::string& name)
: VirtualCluster(name)
{
}

int VirtualClusterMPI::init()
{
    auto ret = _mpi_(MPI_Init(NULL, NULL));
    if (unlikely(ret))
        return ret;

    ret = _mpi_(MPI_Comm_dup(MPI_COMM_WORLD,
                          &world_));
    if (unlikely(ret))
        return ret;

    return SUCCESS;
}

int VirtualClusterMPI::finalize()
{
    EXEC_NOFAIL(_mpi_(MPI_Comm_free(&world_)));

    return _mpi_(MPI_Finalize());
}

int VirtualClusterMPI::rank()
{
    int rank;
    auto ret = _mpi_(MPI_Comm_rank(world_, &rank));
    if (unlikely(ret))
        rank = -1;

    return rank;
}

int VirtualClusterMPI::size()
{
    int size;
    auto ret = _mpi_(MPI_Comm_size(world_, &size));
    if (unlikely(ret))
        size = -1;

    return size;
}

int VirtualClusterMPI::send(int dst, MemoryBuffer& buf)
{
    return _mpi_(MPI_Send(buf.pointer<void>(), buf.len(), MPI_BYTE, 
                  dst, 0, world_));
}

int VirtualClusterMPI::recv(int src, MemoryBuffer& buf)
{
    return _mpi_(MPI_Recv(buf.pointer<void>(), buf.len(), MPI_BYTE, 
                  src, 0, world_,
                  MPI_STATUS_IGNORE));
}
#ifdef DEBUG_BARRIER
static int counter = 1;
#endif
int VirtualClusterMPI::barrier()
{
    #ifdef DEBUG_BARRIER
    debug("VirtualClusterMPI::barrier %d",counter++);
    #endif
    return _mpi_(MPI_Barrier(world_));
}

int VirtualClusterMPI::bcast(int root, MemoryBuffer& buf)
{
    return _mpi_(MPI_Bcast(buf.pointer<void>(), buf.len(), MPI_BYTE, 
                   root, world_));
}

int VirtualClusterMPI::gather(int root, MemoryBuffer& dst, MemoryBuffer& src)
{
    return _mpi_(MPI_Gather(src.pointer<void>(), src.len(), MPI_BYTE,
                dst.pointer<void>(), dst.len(), MPI_BYTE, root, world_));
}

int VirtualClusterMPI::reduce(int root, MemoryBuffer& dst, MemoryBuffer& src, int op)
{
    MPI_Datatype type;
    int sz;
    MPI_Op oq;

    EXEC_IFFAIL(translate_op_to_mpi_type_and_mpi_op(op, &type, &sz, &oq), fatal("Could not translate op."); return ERROR);

    return _mpi_(MPI_Reduce(src.pointer<void>(), 
                dst.pointer<void>(), 
                (src.len()/sz), type, oq, root, world_));
}

int VirtualClusterMPI::kpingpong(const int from, const int to, MemoryBuffer& buf, 
                                 const int num_msg, double* const timing){
    static const int PING = 100;
    static const int PONG = 101;

    double tv = 0;

    if(rank() == from) {
        tv = walltime();

        for (auto i = 0; i < num_msg; ++i) {
            MPI_Send(buf.pointer<void>(), buf.len(), MPI_BYTE, to, PING, world_);
            MPI_Recv(buf.pointer<void>(), buf.len(), MPI_BYTE, to, PONG, world_, 
                    MPI_STATUS_IGNORE);
        }

        tv = walltime() - tv;
        tv = tv / (2.0*num_msg);
    } else if(rank() == to) {
        for (auto i = 0; i < num_msg; ++i) {
            MPI_Recv(buf.pointer<void>(), buf.len(), MPI_BYTE, from, PING, world_, 
                    MPI_STATUS_IGNORE);
            MPI_Send(buf.pointer<void>(), buf.len(), MPI_BYTE, from, PONG, world_);
        }
    } else {
        error("Invalid rank.");
        return ERROR;
    }

    if(timing)
        *timing = tv;

    return SUCCESS;
}

int VirtualClusterMPI::kUniDir(const int from,const int to,
                               MemoryBuffer& buf1, MemoryBuffer& buf2,
                               const int num_msg, double* const timing,
                               const bool do_barrier
                              ){
    /* Unidirectional ping test in which in which the origin rank times
       the time needed to send N messages back-to-back to its partner,
       who returns a receipt message upon receiving all messages, till
       the origin rank receives said receipt. The returned time is
       averaged over the number of messages sent from the origin rank to
       its partner.
     */
    double tv=0.0;
    int i;
    void* const  msg_ptr=buf1.pointer<void>();
    void* const  rcp_ptr=buf2.pointer<void>();
    const size_t msg_siz=buf1.len();
    const size_t rcp_siz=buf2.len();
    MPI_Request* const req =(MPI_Request*)malloc(sizeof(MPI_Request)*num_msg);
    MPI_Status * const stat=(MPI_Status *)malloc(sizeof(MPI_Status )*num_msg);

    if(rank()==from){
        if(do_barrier){
            MPI_Barrier(world_);                                   //Ensures test runs on all simultaneously
        }else{
            MPI_Recv(rcp_ptr,rcp_siz,MPI_BYTE,to,0,world_,stat);   //Wait for partner to initialize receive buffers
        }
        tv=walltime();                                         //Start timing
        for(i=0;i<num_msg;i++){                                //Send messages
            MPI_Isend(msg_ptr,msg_siz,MPI_BYTE,to,i,world_,&req[i]);
        }
        MPI_Recv(rcp_ptr,rcp_siz,MPI_BYTE,to,0,world_,stat);   //Receive receipt (no waitall necessary)
        tv=(walltime()-tv)/num_msg;                            //Stop timing
        MPI_Waitall(num_msg,req,stat);                         //Process all sends to ensure all buffers are freed
    }else if(rank()==to){
        for(i=0;i<num_msg;i++){                                //Initialize receive buffers - this improves performance marginally
            MPI_Irecv(msg_ptr,msg_siz,MPI_BYTE,from,i,world_,&req[i]);
        }
        if(do_barrier){
            MPI_Barrier(world_);                                   //Ensures test runs on all simultaneously
        }else{
            MPI_Isend(rcp_ptr,rcp_siz,MPI_BYTE,from,0,world_,req); //Instruct partner to begin transfer
        }
        MPI_Waitall(num_msg,req,stat);                         //Wait for all messages to be received
        MPI_Send(rcp_ptr,rcp_siz,MPI_BYTE,from,0,world_);      //Send receipt
    }else{
        error("Invalid rank.");
        return ERROR;
    }
    free(req);
    free(stat);

    if(timing!=NULL) *timing = tv;
    return SUCCESS;
}
int VirtualClusterMPI::kUniDirMultiBuf(const int from,const int to,
                                       MemoryBufferMulti& buf_multi,MemoryBuffer& buf2,
                                       const int num_msg, double* const timing,
                                       const bool do_barrier
                                      ){
    /* Unidirectional ping test in which in which the origin rank times
       the time needed to send N messages back-to-back to its partner,
       who returns a receipt message upon receiving all messages, till
       the origin rank receives said receipt. The returned time is
       averaged over the number of messages sent from the origin rank to
       its partner.

       The difference to the previous function is that as many buffers
       are used as there are messages to send.
     */
    double tv=0.0;
    int i;
    void* const* const array_pointer=buf_multi.array_pointer<void*>();
    const size_t       msg_size     =buf_multi.size();
    void* const        receipt      =buf2.pointer<void>();
    const int          receipt_size =buf2.len();
    MPI_Request* const req =(MPI_Request*)malloc(sizeof(MPI_Request)*num_msg);
    MPI_Status * const stat=(MPI_Status *)malloc(sizeof(MPI_Status )*num_msg);

    if(rank()==from){
        if(do_barrier){
            MPI_Barrier(world_);                                      //Ensures test runs on all simultaneously
        }else{
            MPI_Recv(receipt,receipt_size,MPI_BYTE,to,0,world_,stat); //Wait for partner to initialize receive buffers
        }
        tv=walltime();                                              //Start timing
        for(i=0;i<num_msg;i++){                                     //Send messages
            MPI_Isend(array_pointer[i],msg_size,MPI_BYTE,to,i,world_,&req[i]);
        }
        MPI_Recv(receipt,receipt_size,MPI_BYTE,to,0,world_,stat);   //Receive receipt (no waitall necessary)
        tv=(walltime()-tv)/num_msg;                                 //Stop timing
        MPI_Waitall(num_msg,req,stat);                              //Process all sends to ensure all buffers are freed
    }else if(rank()==to){
        for(i=0;i<num_msg;i++){                                     //Initialize receive buffers - this improves performance marginally
            MPI_Irecv(array_pointer[i],msg_size,MPI_BYTE,from,i,world_,&req[i]);
        }
        if(do_barrier){
            MPI_Barrier(world_);                                        //Ensures test runs on all simultaneously
        }else{
            MPI_Isend(receipt,receipt_size,MPI_BYTE,from,0,world_,req); //Instruct partner to begin transfer
        }
        MPI_Waitall(num_msg,req,stat);                              //Wait for all messages to be received
        MPI_Send(receipt,receipt_size,MPI_BYTE,from,0,world_);      //Send receipt
    }else{
        error("Invalid rank.");
        return ERROR;
    }
    free(req);
    free(stat);

    if(timing!=NULL) *timing = tv;
    return SUCCESS;
}
int VirtualClusterMPI::kUniDirLimitedMultiBuf(const int from,const int to,
                                              MemoryBufferMulti& buf_multi,MemoryBuffer& buf2,
                                              const int num_msg, double* const timing,
                                              const bool do_barrier
                                             ){
    /* Unidirectional ping test in which in which the origin rank times
       the time needed to send N messages back-to-back to its partner,
       who returns a receipt message upon receiving all messages, till
       the origin rank receives said receipt. The returned time is
       averaged over the number of messages sent from the origin rank to
       its partner.

       The difference to the previous function is that multiple buffers
       are used. The number of buffers used is less than the number of
       messages to be sent.
     */
          double             tv                  =0.0                                              ;
          int                i                                                                     ;
          std::size_t        j                   =0                                                ;
          void*        const* const array_pointer=buf_multi.array_pointer<void*>()                 ;
    const std::size_t               msg_size     =buf_multi.size()                                 ;
    const std::size_t               num_buf      =buf_multi.number()                               ;
          void*        const        receipt      =buf2.pointer<void>()                             ;
    const int                       receipt_size =buf2.len()                                       ;
          MPI_Request* const        req          =(MPI_Request*)malloc(sizeof(MPI_Request)*num_msg);
          MPI_Status * const        stat         =(MPI_Status *)malloc(sizeof(MPI_Status )*num_msg);

    if(rank()==from){
        if(do_barrier){
            MPI_Barrier(world_);                                      //Ensures test runs on all simultaeneously
        }else{
            MPI_Recv(receipt,receipt_size,MPI_BYTE,to,0,world_,stat); //Wait for partner to initialize receive buffers
        }
        tv=walltime();                                              //Start timing
        for(i=0;i<num_msg;i++){                                     //Send messages
            MPI_Isend(array_pointer[j],msg_size,MPI_BYTE,to,i,world_,&req[i]);
            if(unlikely((++j)==num_buf)) j=0;
        }
        MPI_Recv(receipt,receipt_size,MPI_BYTE,to,0,world_,stat);   //Receive receipt (no waitall necessary)
        tv=(walltime()-tv)/num_msg;                                 //Stop timing
        MPI_Waitall(num_msg,req,stat);                              //Process all sends to ensure all buffers are freed
    }else if(rank()==to){
        for(i=0;i<num_msg;i++){                                     //Initialize receive buffers - this improves performance marginally
            MPI_Irecv(array_pointer[j],msg_size,MPI_BYTE,from,i,world_,&req[i]);
            if(unlikely((++j)==num_buf)) j=0;
        }
        if(do_barrier){
            MPI_Barrier(world_);                                        //Ensures test runs on all simultaeneously
        }else{
            MPI_Isend(receipt,receipt_size,MPI_BYTE,from,0,world_,req); //Instruct partner to begin transfer
        }
        MPI_Waitall(num_msg,req,stat);                              //Wait for all messages to be received
        MPI_Send(receipt,receipt_size,MPI_BYTE,from,0,world_);      //Send receipt
    }else{
        error("Invalid rank.");
        return ERROR;
    }
    free(req);
    free(stat);

    if(timing!=NULL) *timing = tv;
    return SUCCESS;
}

int VirtualClusterMPI::kbipingpong(const int from, const int to,
                                   MemoryBuffer& buf1, MemoryBuffer& buf2,
                                   const int num_msg, double* const timing){
    static const int PING = 100;
    static const int PONG = 101;

    MPI_Request req[2];

    double tv = 0;

    if(rank() == from) {
        tv = walltime();

        for (auto i = 0; i < num_msg; ++i) {
            MPI_Irecv(buf2.pointer<void>(), buf2.len(), MPI_BYTE, to, PONG, world_, &req[0]);
            MPI_Isend(buf1.pointer<void>(), buf1.len(), MPI_BYTE, to, PING, world_, &req[1]);

            MPI_Waitall(2, req, MPI_STATUSES_IGNORE);
        }

        tv = walltime() - tv;
        tv = tv / (2.0*num_msg);
    } else if(rank() == to) {
        for (auto i = 0; i < num_msg; ++i) {
            MPI_Irecv(buf2.pointer<void>(), buf2.len(), MPI_BYTE, from, PING, world_, &req[0]);
            MPI_Isend(buf1.pointer<void>(), buf1.len(), MPI_BYTE, from, PONG, world_, &req[1]);

            MPI_Waitall(2, req, MPI_STATUSES_IGNORE);
        }
    } else {
        error("Invalid rank.");
        return ERROR;
    }

    if(timing)
        *timing = tv;

    return SUCCESS;
}

int VirtualClusterMPI::kalltoall(MemoryBuffer& buffer, //Send & Receive Buffer
                                 const int len_msg,    //Message Length (input)
                                 const int num_msg,    //Number of messages (number of iterations)
                                 double* const timing  //Timing (output)
                                )
{
    /* Initialize Timing */
    double tv=walltime();

    /* Perform All-To-ALL Communication */
    /* MPI_IN_PLACE is used to conserve memory, which allows for
     * the testing of systems with larger messages. Otherwise the
     * maximum message size would have to halved. The downside
     * is that this test therefore does take longer.
     */
    for (auto i = 0; i < num_msg; ++i) {
    MPI_Alltoall(MPI_IN_PLACE          ,len_msg,MPI_BYTE,
             buffer.pointer<void>(),len_msg,MPI_BYTE,world_);
    }
    
    /* Finalize Timing */
    tv=(walltime()-tv)/num_msg; //This is the average time to complete the MPI_Alltoall over the number of messages for the current rank
    if(timing) *timing = tv;

    /* Return */
    return SUCCESS;
}

int VirtualClusterMPI::write_parallel(const linktest_args* args, const std::vector<LinktestStats>& stats)
{
#ifdef USE_SION
    char *buffer;
    long long sz;
    sion_int64 chunksize, sum;
    sion_int32 fsblksize;
    int nfiles;
    char filename[256];
    char *newfname;
    MPI_Comm comm;
    FILE *fp;
    int sid, grank;
    size_t bwrote;

    EXEC_IFFAIL(linktest_output_sion_collect_local_data(this, args, stats, &buffer, &sz), error("linktest_output_sion_collect_local_data failed."); return ERROR);

    std::unique_ptr<StopwatchI> rootWatch = Stopwatchfactory::getRootWatch(rank());
    MPI_Barrier(world_);

    rootWatch->start();
    std::snprintf(filename, sizeof(filename), "%s", args->output.c_str());

    nfiles = 1;
    fsblksize = -1;
    grank = rank();
    chunksize = sz;
    newfname = NULL;
    sid = sion_paropen_mpi(filename, "bw,ansi,collective",
                   &nfiles,
                   world_,
                   &comm,
                   &chunksize,
                   &fsblksize, &grank, &fp, &newfname);

    MPI_Barrier(world_);
    rootWatch->stop();

    printTimingIfRoot(rank(), "[sionopen]", rootWatch->getDuration());

    rootWatch->start();
    /* write buffer */
    bwrote = sion_coll_fwrite_mpi(buffer, 1, chunksize, sid);
    if (static_cast<sion_int64>(bwrote) != chunksize) {
        fatal("problems writing data to sionfile on rank %d, buffer[%lu] ...",
              rank(), chunksize);
        return ERROR;
    }
    rootWatch->stop();

    MPI_Reduce(&chunksize, &sum, 1, SION_MPI_INT64, MPI_SUM, 0, world_);
    printTimingIfRoot(rank(), "[sionwrite]", rootWatch->getDuration());
    if (0 == rank()) {
        std::printf("OUTPUT: writing data to file %-30s (%sB)\n",
            filename, 
            UnitPrefix::IEC_prefix(sum, 9).get());

    }

    rootWatch->start();
    sion_parclose_mpi(sid);
    rootWatch->stop();

    printTimingIfRoot(rank(), "[sionclose]", rootWatch->getDuration());    

    return SUCCESS;
#else
    error("This version of the benchmark was compiled "
          "without SIONlib support.");

    return ERROR;
#endif
}
