/****************************************************************************
**  LinkTest                                                               **
*****************************************************************************
**  Copyright (c) 2008-2022                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/
#include "vcluster_portals.h"
#include "portals4_macros.h"
#include "memory.h"
#include "error.h"
#include "timing.h"
#include "stopwatch.h"
#include <assert.h>

std::vector<ptl_process_t> VirtualClusterPortals::getPhysicalFromRank() {
    ptl_process_t physId;
    CHECK_RETURNVAL( PtlGetPhysId(mni_handle, &physId) );

    debug("PMI Rank=%d, Hostname=%10s, Portals NID=%d PID=%d",
        rank(),
        hostname().c_str(),
        physId.phys.nid,
        physId.phys.pid);
    
    std::vector<ptl_process_t> physicalFromRank(size());
    gather(0, physicalFromRank.data(), &physId, 1);
    bcast(0, physicalFromRank.data(), size());

    if(physicalFromRank.at(rank()).phys.nid != physId.phys.nid) fatal("Failed to broadcast physicalFromRank");
    if(physicalFromRank.at(rank()).phys.pid != physId.phys.pid) fatal("Failed to broadcast physicalFromRank");

    return physicalFromRank;
}

int VirtualClusterPortals::init()
{
    set_helper_pointer(VirtualCluster::factory("mpi"));
    EXEC_NOFAIL(helper_->init());
    
    if(PTL_MAJOR_VERSION != 4 || PTL_MINOR_VERSION != 0) {
        warn("Portals versions other than 4.0 may not be suppported");
    }
    CHECK_RETURNVAL( PtlInit() );
    CHECK_RETURNVAL( PtlNIInit(
        PTL_IFACE_DEFAULT, // Manual 3.3.5: "Check README"
        PTL_NI_LOGICAL | PTL_NI_MATCHING, // Logical => using ranks, Matching => using send/recv semantics
        PTL_PID_ANY,
        nullptr,// &mni_limits_desired
        &mni_limits_actual,
        &mni_handle));

    auto physicalFromRank = getPhysicalFromRank();
    CHECK_RETURNVAL( PtlSetMap(mni_handle, physicalFromRank.size(), physicalFromRank.data()) );
    const auto DEFAULT_OPTIONS = 0;
    CHECK_RETURNVAL( PtlEQAlloc(mni_handle, 1000, &pt_eq_handle) );
    CHECK_RETURNVAL( PtlPTAlloc(mni_handle, DEFAULT_OPTIONS, pt_eq_handle, PTL_PT_ANY, &pt_index) );

    return SUCCESS;
}

int VirtualClusterPortals::finalize()
{
    debug("VirtualClusterPortals::finalize()");
    CHECK_RETURNVAL( PtlPTFree(mni_handle, pt_index) );
    CHECK_RETURNVAL( PtlNIFini(mni_handle) );
    PtlFini();
    EXEC_NOFAIL(helper_->finalize());
    return SUCCESS;
}

void  VirtualClusterPortals::prepareSendStructs(const MemoryBuffer& buf) {
    debug("VirtualClusterPortals::prepareSendStructs(%p)", buf.p());
    md.start  = buf.p();
    md.length = buf.len();
    md.options   = PTL_MD_EVENT_CT_ACK;
    md.eq_handle = PTL_EQ_NONE;   // i.e. don't queue send events
    CHECK_RETURNVAL( PtlCTAlloc(mni_handle, &md.ct_handle) ); 
    CHECK_RETURNVAL( PtlMDBind(mni_handle, &md, &md_handle) ); // Bind memory descriptor
}

void VirtualClusterPortals::prepareRecvStructs(const MemoryBuffer& buf) {
    debug("VirtualClusterPortals::prepareRecvStructs(%p)", buf.p());
    me.start  = buf.p();
    me.length = buf.len();
    me.uid    = PTL_UID_ANY;
    me.match_id.rank = PTL_RANK_ANY;
    me.match_bits    = MATCH_BITS;
    me.ignore_bits   = IGNORE_BITS;
    me.options = (PTL_ME_OP_PUT | PTL_ME_EVENT_CT_COMM | PTL_ME_EVENT_COMM_DISABLE ); // React to puts, count communication events, do not generate full communication events
    CHECK_RETURNVAL( PtlCTAlloc(mni_handle, &me.ct_handle));
    CHECK_RETURNVAL( PtlMEAppend(mni_handle, pt_index, &me, PTL_PRIORITY_LIST, nullptr, &me_handle));
    ptl_event_t event;
    CHECK_RETURNVAL( PtlEQWait(pt_eq_handle, &event) ); // TODO allow PTL_EQ_DROPPED 
    if (! (event.type == PTL_EVENT_LINK && event.ni_fail_type == PTL_NI_OK)) {
        // TODO Check for overflow/dropped events
        error("PtlMEAppend failed");
    }
}

ptl_size_t VirtualClusterPortals::getSendCounter() {
    debug("VirtualClusterPortals::getSendCounter()");
    CHECK_RETURNVAL( PtlCTGet(md.ct_handle, &send_ct) );
    debug("Send (MD): success %d - failure %d", send_ct.success, send_ct.failure);
    if(send_ct.failure > 0) {
        error("Failed operation on MD");
    }
    return send_ct.success;
}

ptl_size_t VirtualClusterPortals::getRecvCounter() {
    debug("VirtualClusterPortals::getRecvCounter()");
    CHECK_RETURNVAL( PtlCTGet(me.ct_handle, &recv_ct) );
    debug("Recv (ME): success %d - failure %d", recv_ct.success, recv_ct.failure);
    if(recv_ct.failure > 0) {
        error("Failed operation on ME");
    }
    return recv_ct.success;
}

void VirtualClusterPortals::recvMessages(const unsigned long num_msg, const unsigned long counter_start) {
    debug("VirtualClusterPortals::recvMessages(%lu, %lu)", num_msg, counter_start);
    CHECK_RETURNVAL( PtlCTWait(me.ct_handle, counter_start + num_msg, &recv_ct) );
}

void VirtualClusterPortals::sendMessages(const int to, MemoryBuffer& buf, const unsigned long num_msg, const unsigned long counter_start) {
    debug("VirtualClusterPortals::sendMessages(%d, %p, %lu, %lu)", to, num_msg, buf.p() , counter_start);
    const ptl_size_t localOffset = 0;
    const ptl_size_t remoteOffset = 0;
    const ptl_hdr_data_t header_data = 0;
    ptl_process_t target;
    target.rank = to;
    
    for(unsigned long n = 1; n <= num_msg; n++) {
        CHECK_RETURNVAL( PtlPut(md_handle, localOffset, buf.len(), PTL_CT_ACK_REQ, target, pt_index, MATCH_BITS, remoteOffset, nullptr, header_data) );
    }
    CHECK_RETURNVAL( PtlCTWait(md.ct_handle, counter_start + num_msg, &send_ct) );
}

void VirtualClusterPortals::releaseRecvStructs() {
    debug("releaseRecvStructs()");
    CHECK_RETURNVAL( PtlMEUnlink(me_handle) );
    CHECK_RETURNVAL( PtlCTFree(me.ct_handle) );
};

void VirtualClusterPortals::releaseSendStructs() {
    debug("releaseSendStructs()");
    CHECK_RETURNVAL( PtlMDRelease(md_handle) );
    CHECK_RETURNVAL( PtlCTFree(md.ct_handle) );
};

int VirtualClusterPortals::kpingpong(const int from, const int to, MemoryBuffer& buf, const int num_msg, double* const timing)
{
    bool isSender = rank() == from;
    bool isReceiver = rank() == to;
    auto watch = Stopwatchfactory::getRankWatch(rank(), from);
    prepareSendStructs(buf);
    prepareRecvStructs(buf);
    auto sendCounterBeforeKernel = getSendCounter();
    auto recvCounterBeforeKernel = getRecvCounter();
    
    barrier();

    if(isSender) {
        watch->start();
        sendMessages(to, buf, num_msg, sendCounterBeforeKernel);
        recvMessages(num_msg, recvCounterBeforeKernel);
        watch->stop();
    }
    if(isReceiver) {
        watch->start();
        recvMessages(num_msg, recvCounterBeforeKernel);
        sendMessages(from, buf, num_msg, sendCounterBeforeKernel);
        watch->stop();
    }

    barrier();

    getSendCounter();
    getRecvCounter();
    releaseSendStructs();
    releaseRecvStructs();
    if (timing != nullptr)
    {
        *timing = watch->getDuration().count() / (2*num_msg);
    }

    return SUCCESS;
}

int VirtualClusterPortals::kUniDir(
    const int from, const int to,
    MemoryBuffer &buf1, MemoryBuffer &buf2,
    const int num_msg, double *const timing,
    const bool /*doBarrier*/)
{
    bool isSender = rank() == from;
    bool isReceiver = rank() == to;
    auto watch = Stopwatchfactory::getRankWatch(rank(), from);
    prepareSendStructs(buf1);
    prepareRecvStructs(buf2);
    auto sendCounterBeforeKernel = getSendCounter();
    auto recvCounterBeforeKernel = getRecvCounter();

    barrier();

    if(isSender) {
        watch->start();
        sendMessages(to, buf1, num_msg, sendCounterBeforeKernel);
        recvMessages(1, recvCounterBeforeKernel);
        watch->stop();
    }
    if(isReceiver) {
        watch->start();
        recvMessages(num_msg, recvCounterBeforeKernel);
        sendMessages(from, buf1, 1, sendCounterBeforeKernel);
        watch->stop();
    }

    barrier();

    getSendCounter();
    getRecvCounter();
    releaseSendStructs();
    releaseRecvStructs();
    if (timing != nullptr)
    {
        *timing = watch->getDuration().count() / num_msg;
    }

    return SUCCESS;

}

int VirtualClusterPortals::kbipingpong(
    const int from, const int to,
    MemoryBuffer& buf1, MemoryBuffer& buf2,
    const int num_msg, double* const timing)
{
    int partner;
    if(rank() == from) {
        partner = to;
    }
    if(rank() == to) {
        partner = from;
    }
    auto watch = Stopwatchfactory::getRankWatch(rank(), from);
    prepareSendStructs(buf1);
    prepareRecvStructs(buf2);
    auto sendCounterBeforeKernel = getSendCounter();
    auto recvCounterBeforeKernel = getRecvCounter();

    barrier();

    watch->start();
    sendMessages(partner, buf1, num_msg, sendCounterBeforeKernel);
    recvMessages(num_msg, recvCounterBeforeKernel);
    watch->stop();

    barrier();

    getSendCounter();
    getRecvCounter();
    releaseSendStructs();
    releaseRecvStructs();
    if (timing != nullptr)
    {
        *timing = watch->getDuration().count() / (2.0 * num_msg);
    }

    return SUCCESS;

}


VirtualClusterPortals::VirtualClusterPortals()
: VirtualClusterWithHelper(VirtualClusterPortals::NAME)
{
}
