/****************************************************************************
**  LinkTest                                                               **
*****************************************************************************
**  Copyright (c) 2008-2022                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/
#include "vcluster_psm2.h"
#include "config.h"
#include "compiler.h"
#include "timing.h"
#include "vcluster_helper.h"
#include "environ.h"
#include "error.h"
#include "output_sion.h"
#include "pmi.h"
#include <cstdlib>
#include <cstring>
#include <unistd.h>
#include <malloc.h>
#include <cinttypes>
#include <psm2.h>
#include <psm2_mq.h>

#if 1 == HAVE_MINIPMI
    #ifdef __cplusplus
    extern "C" {
    #endif
    #include <minipmi.h>
    #ifdef __cplusplus
    }
    #endif
#endif

#undef  PSM2_ENVIRON_PREFIX
#define PSM2_ENVIRON_PREFIX LINKTEST_ENVIRON_PREFIX "PSM2_"


namespace linktest
{
namespace psm2
{

const int kDefaultSL = 0;

// psm2 specific environment variables.
static struct environ
{
    int             hfi;
    int             port;
    int             sl;
    const char      *helper;
    int             debug;
} env = {
    .hfi    = -1,
    .port   = -1,
    .sl     = kDefaultSL,
    .helper = "tcp",
    .debug  = 0
};

struct context
{
    psm2_ep_t       ep;
    psm2_epid_t     epid;

    psm2_mq_t       mq;
};

/* AoS instead of SoA so that we can directly pass them
 * them to psm2_ep_connect() without requiring temporary
 * memory.
 */
struct connections
{
    psm2_epid_t             *epid;
    psm2_epaddr_t           *epaddr;
};

class VirtualClusterImpl
{

public:
    VirtualClusterImpl(VirtualCluster* owner);

public:
    int init();
    int finalize();

public:
    context                 ctx;
    connections             conns;

private:
    minipmi_context*        pmi_;

private:
    VirtualCluster*         owner_;

public:
    int fill_ep_open_opts(struct psm2_ep_open_opts *opts);
    int create_context();
    int delete_context();
    int exchange_epids(int rank, int size, psm2_epid_t here, psm2_epid_t *epid);
    int connect();
    int disconnect();

};

static int read_psm2_environ()
{
    read_environ_str(PSM2_ENVIRON_PREFIX "HELPER", &env.helper);

    read_environ_int(PSM2_ENVIRON_PREFIX "HFI", &env.hfi);
    read_environ_int(PSM2_ENVIRON_PREFIX "PORT", &env.port);
    read_environ_int(PSM2_ENVIRON_PREFIX "SL", &env.sl);
    read_environ_int(PSM2_ENVIRON_PREFIX "DEBUG", &env.debug);

    return SUCCESS;
}

static int initialize_psm2_lib()
{
    psm2_error_t ret;
    int vermaj, vermin;

    vermaj = PSM2_VERNO_MAJOR;
    vermin = PSM2_VERNO_MINOR;

    EXEC_IFFAIL(psm2_init(&vermaj, &vermin), error("psm2_init failed with return value %d (%s).", ret, psm2_error_get_string(ret)); return ERROR);

    EXEC_IFFAIL(psm2_error_register_handler(NULL, PSM2_ERRHANDLER_NO_HANDLER), error("psm2_error_register_handler failed with return value %d (%s).", ret, psm2_error_get_string(ret)); return ERROR);

    return SUCCESS;
}

static int finalize_psm2_lib()
{
    psm2_error_t ret;

    EXEC_IFFAIL(psm2_finalize(), error("psm2_finalize failed with return value %d (%s).", ret, psm2_error_get_string(ret)); return ERROR);

    return SUCCESS;
}

VirtualClusterImpl::VirtualClusterImpl(VirtualCluster* owner)
: owner_(owner)
{
}

int VirtualClusterImpl::fill_ep_open_opts(struct psm2_ep_open_opts *opts)
{
    std::memset(opts, 0, sizeof(struct psm2_ep_open_opts));
    psm2_ep_open_opts_get_defaults(opts);   /* Ignore errors */

    opts->unit  = env.hfi;
    opts->outsl = env.sl;
    if (env.port > 0) {
        opts->port = env.port;
    }
    opts->affinity = PSM2_EP_OPEN_AFFINITY_SKIP;

    return SUCCESS;
}

int VirtualClusterImpl::create_context()
{
    psm2_error_t ret;
    struct psm2_ep_open_opts opts;
    psm2_uuid_t uuid;

    fill_ep_open_opts(&opts);

    psm2_uuid_generate(uuid);
    EXEC_IFFAIL(owner_->bcast(0, uuid, sizeof(psm2_uuid_t)), error("Broadcast of uuid failed"); return ERROR);

    EXEC_IFFAIL(psm2_ep_open(uuid, &opts, &ctx.ep, &ctx.epid), error("psm2_ep_open failed with return value %d (%s).", ret, psm2_error_get_string(ret)); return ERROR);

    ret = psm2_mq_init(ctx.ep, PSM2_MQ_ORDERMASK_NONE, NULL, 0, &ctx.mq);
    if (unlikely(ret)) {
        error("psm2_mq_init failed with return value %d (%s).", ret, psm2_error_get_string(ret));
        psm2_ep_close(ctx.ep, PSM2_EP_CLOSE_FORCE, 0);  // Ignore return value
        return ERROR;
    }

    return SUCCESS;
}

int VirtualClusterImpl::delete_context()
{
    psm2_error_t ret;

    EXEC_IFFAIL(psm2_mq_finalize(ctx.mq), error("psm2_mq_finalize failed with return value %d (%s).", ret, psm2_error_get_string(ret)); return ERROR);

    EXEC_IFFAIL(psm2_ep_close(ctx.ep, PSM2_EP_CLOSE_GRACEFUL, 0), error("psm2_ep_close failed with return value %d (%s).", ret, psm2_error_get_string(ret)); return ERROR);

    return SUCCESS;
}

int VirtualClusterImpl::exchange_epids(int rank, int size, psm2_epid_t here, psm2_epid_t *epid)
{
#if 1 == HAVE_MINIPMI
    char key[64], val[64];
    int i, len;
    int ret;

    ret = std::snprintf(key, sizeof(key), PSM2_ENVIRON_PREFIX "EPID" "_%08d", rank);
    if (unlikely((ret < 0) || (ret >= static_cast<int>(sizeof(key))))) {
        error("std::snprintf failed: string truncated.");
        return ERROR;
    }

    ret = std::snprintf(val, sizeof(val), "%" PRIu64, here);
    if (unlikely((ret < 0) || (ret >= static_cast<int>(sizeof(key))))) {
        error("std::snprintf failed: string truncated.");
        return ERROR;
    }

    EXEC_IFFAIL(minipmi_kvs_put(pmi_, key, val), error("minipmi_kvs_put() failed."); return ERROR);

    EXEC_IFFAIL(minipmi_kvs_sync(pmi_), error("minipmi_kvs_sync() failed."); return ERROR);

    for (i = 0; i < size; ++i) {
        ret = std::snprintf(key, sizeof(key), PSM2_ENVIRON_PREFIX "EPID" "_%08d", i);
        if (unlikely((ret < 0) || (ret >= static_cast<int>(sizeof(key))))) {
            error("std::snprintf failed: string truncated.");
            return ERROR;
        }

        EXEC_IFFAIL(minipmi_kvs_get(pmi_, key, val, sizeof(val), &len), error("minipmi_kvs_get() failed."); return ERROR);

        ret = sscanf(val, "%" PRIu64, &epid[i]);
        if (unlikely(1 != ret)) {
            error("sccanf() failed.");
            return ERROR;
        }
    }

    return SUCCESS;
#else
    error("exchange_epids() is currently implemented on top of minipmi.");
    return ERROR;
#endif
}

int VirtualClusterImpl::connect()
{
    psm2_error_t ret;
    int err;
    int rank, size;
    psm2_error_t *errors;

    rank = owner_->rank();
    size = owner_->size();

    conns.epid = (psm2_epid_t *)malloc(size*sizeof(psm2_epid_t));
    if (unlikely(!conns.epid)) {
        error("malloc failed.");
        return ERROR;
    }

    err = exchange_epids(rank, size, ctx.epid, conns.epid);
    if (unlikely(err)) {
        error("exchange_epids failed.");
        return ERROR;       /* Leaks memory */
    }

    conns.epaddr = (psm2_epaddr_t *)malloc(size*sizeof(psm2_epaddr_t));
    if (unlikely(!conns.epaddr)) {
        error("malloc failed.");
        return ERROR;       /* Leaks memory */
    }

    errors = (psm2_error_t *)malloc(size*sizeof(psm2_error_t));
    if (unlikely(!errors)) {
        error("malloc failed.");
        return ERROR;       /* Leaks memory */
    }

    EXEC_IFFAIL(psm2_ep_connect(ctx.ep, size, conns.epid, NULL, errors, conns.epaddr, 0), error("psm2_ep_connect failed with return value %d (%s).", ret, psm2_error_get_string(ret)); return ERROR);

    free(errors);

    return SUCCESS;
}

int VirtualClusterImpl::disconnect()
{
    free(conns.epid);
    free(conns.epaddr);

    conns.epid   = NULL;
    conns.epaddr = NULL;

    return SUCCESS;
}

int VirtualClusterImpl::init()
{
    auto ret = linktest_minipmi_context_borrow(&pmi_);
#if 1 == HAVE_MINIPMI
    if (unlikely(ret)) {
        error("linktest_minipmi_context_borrow() failed.");
        return ERROR;
    }
#endif

    EXEC_NOFAIL(linktest::psm2::initialize_psm2_lib());

    EXEC_NOFAIL(create_context());

    EXEC_IFFAIL(connect(), delete_context(); return ERROR);

    return SUCCESS;
}

int VirtualClusterImpl::finalize()
{
    auto ret1 = disconnect();
    auto ret2 = delete_context();
    auto ret3 = linktest::psm2::finalize_psm2_lib();

    auto ret4 = linktest_minipmi_context_return(pmi_);
#if 1 == HAVE_MINIPMI
    if (unlikely(ret4)) {
        error("linktest_minipmi_context_return() failed.");
    }
#endif

    return (ret1 || ret2 || ret3 || ret4);
}

}
}

VirtualClusterPSM2::VirtualClusterPSM2(const std::string& name)
: VirtualClusterWithHelper(name), 
  impl_(new linktest::psm2::VirtualClusterImpl(this))
{
}

int VirtualClusterPSM2::init()
{
    EXEC_NOFAIL(linktest::psm2::read_psm2_environ());

    set_helper_pointer(VirtualCluster::factory(linktest::psm2::env.helper));
    EXEC_NOFAIL(helper_->init());

    return impl_->init();
}

int VirtualClusterPSM2::finalize()
{
    helper_->finalize();    // Ignore error code
    delete helper_;

    return impl_->finalize();
}

int VirtualClusterPSM2::kpingpong_send(int other, MemoryBuffer& buf, int tag0)
{
    psm2_error_t ret;

    psm2_mq_tag_t tag;
#if 0
    psm2_mq_req_t req;
#endif

    std::memset(&tag, 0, sizeof(tag));
    tag.tag0 = tag0;

#if 0
    ret = psm2_mq_isend2(impl_->ctx.mq, conns.epaddr[other], 0, &tag,
                 buf.pointer<void>(), buf.len(), NULL, &req);
    if (unlikely(ret)) {
        error("psm2_mq_isend2 failed with return value %d (%s).", ret, psm2_error_get_string(ret));
        return ERROR;
    }

    EXEC_IFFAIL(psm2_mq_wait2(&req, NULL), error("psm2_mq_wait2 failed with return value %d (%s).", ret, psm2_error_get_string(ret)); return ERROR);
#else
    ret = psm2_mq_send2(impl_->ctx.mq, impl_->conns.epaddr[other], 0, &tag,
                buf.pointer<void>(), buf.len());
    if (unlikely(ret)) {
        error("psm2_mq_send2 failed with return value %d (%s).", ret, psm2_error_get_string(ret));
        return ERROR;
    }
#endif

    return SUCCESS;
}

int VirtualClusterPSM2::kpingpong_recv(int other, MemoryBuffer& buf, int tag0)
{
    psm2_error_t ret;

    psm2_mq_tag_t tag;
    psm2_mq_tag_t tagsel;
    psm2_mq_req_t req;

    std::memset(&tag   , 0, sizeof(tag));
    std::memset(&tagsel, 0, sizeof(tagsel));
    tag.tag0    = tag0;
    tagsel.tag0 = ~0;

    ret = psm2_mq_irecv2(impl_->ctx.mq, impl_->conns.epaddr[other], &tag, &tagsel, 0,
                 buf.pointer<void>(), buf.len(), NULL, &req);
    if (unlikely(ret)) {
        error("psm2_mq_irecv2 failed with return value %d (%s).", ret, psm2_error_get_string(ret));
        return ERROR;
    }

    EXEC_IFFAIL(psm2_mq_wait2(&req, NULL), error("psm2_mq_wait2 failed with return value %d (%s).", ret, psm2_error_get_string(ret)); return ERROR);

    return SUCCESS;
}

int VirtualClusterPSM2::kpingpong(const int from, const int to, MemoryBuffer& buf,
                                  const int num_msg, double* const timing){
    static const int PING = 100;
    static const int PONG = 101;

    double tv = 0;

    if(rank() == from) {
        tv = walltime();

        for (auto i = 0; i < num_msg; ++i) {
            EXEC_NOFAIL(kpingpong_send(to, buf, PING));

            EXEC_NOFAIL(kpingpong_recv(to, buf, PONG));
        }

        tv = walltime() - tv;
        tv = tv / (2.0*num_msg);
    } else if(rank() == to) {
        for (auto i = 0; i < num_msg; ++i) {
            EXEC_NOFAIL(kpingpong_recv(from, buf, PING));

            EXEC_NOFAIL(kpingpong_send(from, buf, PONG));
        }
    } else {
        error("Invalid rank.");
        return ERROR;
    }

    if(timing)
        *timing = tv;

    return SUCCESS;
}

int VirtualClusterPSM2::kUniDir(const int from, const int to,
                                MemoryBuffer& buf1, MemoryBuffer& buf2,
                                const int num_msg, double* const timing,
                                const bool doBarrier
                               ){
/* Unidirectional ping test in which in which the origin rank times
   the time needed to send N messages back-to-back to its partner,
   who returns a receipt message upon receiving all messages, till
   the origin rank receives said receipt. The returned time is
   averaged over the number of messages sent from the origin rank to
   its partner.
 */
    double tv=0.0;
    if(rank()==from){
        tv=walltime();
        for(int i=0;i<num_msg;i++){
            auto ret=kpingpong_send(to,buf1,i);
            if(unlikely(ret))return ERROR;
        }
        auto ret=kpingpong_recv(to,buf2,0);
        if(unlikely(ret))return ERROR;
        tv=(walltime()-tv)/num_msg;
    }else if(rank()==to){
        for(int i=0;i<num_msg;i++){
            auto ret=kpingpong_recv(from,buf1,i);
            if(unlikely(ret))return ERROR;
        }
        auto ret=kpingpong_send(from,buf2,0);
        if(unlikely(ret)){return ERROR;}
    }else{
        error("Invalid rank.");
        return ERROR;
    }
    if(timing!=NULL) *timing = tv;
    return SUCCESS;
}
int VirtualClusterPSM2::kUniDirMultiBuf(const int from,const int to,
                                        MemoryBufferMulti& buf_multi, MemoryBuffer& buf2,
                                        const int num_msg, double* const timing,
                                        const bool doBarrier
                                       ){
    fatal("not yet implemented.");
    return ERROR;
}
int VirtualClusterPSM2::kUniDirLimitedMultiBuf(const int from,const int to,
                                               MemoryBufferMulti& buf_multi, MemoryBuffer& buf2,
                                               const int num_msg, double* const timing,
                                               const bool doBarrier
                                              ){
    fatal("not yet implemented.");
    return ERROR;
}
