/****************************************************************************
**  LinkTest                                                               **
*****************************************************************************
**  Copyright (c) 2008-2022                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#include "vcluster_sion_generic_adapter.h"
#include "sion_const.h"
#include "sion_datatypes.h"
#include "error.h"
#include <stdexcept>

#include <unistd.h>

namespace VirtualClusterSionGenericAdapter {

int create_lcg_cb(void **local_vcluster, void *vcluster, 
                  int grank, int gsize, 
                  int lrank, int lsize,
                  int filenumber, int numfiles
                 ) {
    if(numfiles > 1) {
        fatal("Comm splitting not implemented, single file required");
    }

    *local_vcluster = vcluster;

    return SUCCESS;
}

int free_lcg_cb(void *local_commgroup) {
    local_commgroup = nullptr;
    // local comm points to global comm. Don't destroy global comm, just remove pointer

    return SUCCESS;
}

int barrier_cb(void * vcluster){
    return reinterpret_cast<VirtualCluster *>(vcluster)->barrier();
}

int bcastr_cb(void * data, void * vcluster, int datatype, int nelem, int root){ 
    auto vcluster_ptr = reinterpret_cast<VirtualCluster *>(vcluster);
    switch(datatype) {
    case _SION_INT64:
        return vcluster_ptr->bcast(root, reinterpret_cast<sion_int64 *>(data), nelem);
    case _SION_INT32:
        return vcluster_ptr->bcast(root, reinterpret_cast<sion_int32 *>(data), nelem);
    case _SION_CHAR:
        return vcluster_ptr->bcast(root, reinterpret_cast<char *>(data), nelem);
    default:
        throw std::invalid_argument("Unknown datatype id");
    }
}

int gatherr_cb(void * indata, void * outdata, void * vcluster, int datatype, int nelem, int root){ 
    auto vcluster_ptr = reinterpret_cast<VirtualCluster *>(vcluster);
    switch(datatype) {
    case _SION_INT64:
        return vcluster_ptr->gather(root, reinterpret_cast<sion_int64 *>(outdata), reinterpret_cast<sion_int64 *>(indata), nelem);
    case _SION_INT32:
        return vcluster_ptr->gather(root, reinterpret_cast<sion_int32 *>(outdata), reinterpret_cast<sion_int32 *>(indata), nelem);
    case _SION_CHAR:
        return vcluster_ptr->gather(root, reinterpret_cast<char *>(outdata), reinterpret_cast<char *>(indata), nelem);
    default:
        throw std::invalid_argument("Unknown datatype id");
    } 
}

int gathervr_cb(void * indata, void * outdata, void * vcluster, int datatype, int * counter, int nelem, int root){ 
    throw("Not implemented"); 
}

int scatterr_cb(void * indata, void * outdata, void * vcluster, int datatype, int nelem, int root){
    auto vcluster_ptr = reinterpret_cast<VirtualCluster *>(vcluster);
    switch(datatype) {
    case _SION_INT64:
        return vcluster_ptr->scatter(root, reinterpret_cast<sion_int64 *>(outdata), reinterpret_cast<sion_int64 *>(indata), nelem);
    case _SION_INT32:
        return vcluster_ptr->scatter(root, reinterpret_cast<sion_int32 *>(outdata), reinterpret_cast<sion_int32 *>(indata), nelem);
    case _SION_CHAR:
        return vcluster_ptr->scatter(root, reinterpret_cast<char *>(outdata), reinterpret_cast<char *>(indata), nelem);
    default:
        throw std::invalid_argument("Unknown datatype id");
    } 
}

int scattervr_cb(void * indata, void * outdata, void * vcluster, int datatype, int * counter, int nelem, int root){ 
    throw("Not implemented"); 
}

int global_rank(VirtualCluster& cl) {
    return cl.rank();
}

int global_size(VirtualCluster& cl) {
    return cl.size();
}

// Only one vcluster exists, we do not allow splitting
int local_rank(VirtualCluster& cl) {
    return global_rank(cl);
}

int local_size(VirtualCluster& cl) {
    return global_size(cl);
}

}