/****************************************************************************
**  LinkTest                                                               **
*****************************************************************************
**  Copyright (c) 2008-2022                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/
#include "vcluster_ucp.h"
#include "config.h"
#include "compiler.h"
#include "timing.h"
#include "vcluster_helper.h"
#include "environ.h"
#include "error.h"
#include "output_sion.h"
#include "pmi.h"
#include <cstdlib>
#include <cstring>
#include <unistd.h>
#include <malloc.h>
#include <cinttypes>
#include <cassert>
#include <utility>
#include <vector>

#if 1 == HAVE_MINIPMI
    #ifdef __cplusplus
    extern "C" {
    #endif
    #include <minipmi.h>
    #ifdef __cplusplus
    }
    #endif
#endif

#undef  UCP_ENVIRON_PREFIX
#define UCP_ENVIRON_PREFIX LINKTEST_ENVIRON_PREFIX "UCP_"


//  UCP specific environment variables.
static struct environ
{
    const char      *helper;
    int             debug;
} env = {
    .helper = "tcp",
    .debug  = 0
};

namespace ucp
{

// Wait for the request with the handle `req` to finish.
static int wait_for_request_completion(ucp_worker_h worker, void* req);

class WorkerAddress
{

public:
    WorkerAddress();
    WorkerAddress(const WorkerAddress&) = delete;
    WorkerAddress(WorkerAddress&& other);

    ~WorkerAddress();

public:
    enum {
        kOwnerUndefined = 0,
        kOwnerUCP = 1,
        kOwnerSelf = 2
    };

public:
    /* Initiale the worker address from the `worker`. The owner will
     * be UCP.
     */
    int init(ucp_worker_h worker);
    /* Matching finalization call to `init()`. We need to be able to
     * release the address before destroying the worker.
     */
    int finalize(ucp_worker_h worker);

    // Resize the address to fit `len` bytes. The owner will be the code.
    int resize(std::uint64_t len);

public:
    ucp_address_t*  addr;
    std::uint64_t        len;
    int             owner;

};


class Connection
{

public:
    Connection();
    Connection(const Connection&) = delete;
    Connection(Connection&& other);

public:
    int     create_ucp_ep(ucp_worker_h worker);
    int     close_ucp_ep(ucp_worker_h worker);

public:
    WorkerAddress   remote_addr_;                  
    ucp_ep_h        ep_;

};

class VirtualClusterImpl
{

public:
    VirtualClusterImpl(VirtualCluster* obj);

    VirtualClusterImpl(const VirtualCluster&) = delete;
    VirtualClusterImpl(VirtualCluster&&) = delete;

public:
    int     initialize_ucp_lib();
    int     finalize_ucp_lib();

public:
    int     connect();
    int     disconnect();

private:
    int     init_ucp_context();
    int     init_ucp_worker();

private:
    int     connect_ep(int rank, int other, int server);
    int     exchange_address(int rank, int other, int server);

public:
    // Poll on the request until it is done.
    int     wait_for_request_completion(void* req);

private:
    /* Reference to the actual object. Due to possible dependency
     * circles we have to be careful with calling `obj_` functions
     * from the implementation. 
     * The `rank()` and `size()` functions may be used outside of
     * the constructor. 
     */
    VirtualCluster*                 obj_;

public:
    minipmi_context*                pmi;

public:
    ucp_context_h                   ctx_;
    ucp_worker_h                    worker_;

public:
    std::uint64_t                        ucx_request_size_;

public:
    static const int kReqSize = 8*1024*1024;

    /* Buffer space that can be used to store request handles. UCX 
     * requires very large request handles, so that we only keep a 
     * handful of the for the entire program. That is enough to implement
     * the pingpong logic
     */
    char                            recv_req[kReqSize];
    char                            send_req[kReqSize];

public:
    WorkerAddress                   worker_addr_;

public:
    std::vector<Connection>         conns_;

};

}

static int read_ucp_environ()
{
    read_environ_str(UCP_ENVIRON_PREFIX "HELPER", &env.helper);
    read_environ_int(UCP_ENVIRON_PREFIX "DEBUG", &env.debug);

    return SUCCESS;
}

static int ucp::wait_for_request_completion(ucp_worker_h worker, void* req)
{
    if (req) {
        ucs_status_t status = UCS_INPROGRESS;
        do {
            if(ucp_worker_progress(worker)){
                status=ucp_request_check_status(req);
#if defined(DEBUG_UCP)
                std::printf("wait_for_request_completion: status=%d\n",status);
#endif
            }
        } while (status==UCS_INPROGRESS);
#if defined(DEBUG_UCP)
        std::printf("wait_for_request_completion: DONE status=%d\n",status);
#endif
        if (UCS_STATUS_IS_ERR(status)) {
            error("ucp_request_check_status() failed with error = %d.", status);
            return -1;
        }
    }
    return SUCCESS;
}


ucp::WorkerAddress::WorkerAddress()
: addr(nullptr), len(0), owner(WorkerAddress::kOwnerUndefined)
{
}

ucp::WorkerAddress::~WorkerAddress()
{
    if ((WorkerAddress::kOwnerSelf == owner) and addr) {
        delete[] reinterpret_cast<char*>(addr);
    }
}

ucp::WorkerAddress::WorkerAddress(ucp::WorkerAddress&& other)
: addr(std::exchange(other.addr, nullptr)), 
  len(std::exchange(other.len, 0)),
  owner(std::exchange(other.owner, WorkerAddress::kOwnerUndefined))
{
}

int ucp::WorkerAddress::init(ucp_worker_h worker)
{
    auto err = ucp_worker_get_address(worker, &addr, &len);
    if (UCS_OK != err) {
        error("ucp_worker_create() returned %lld", (long long )err);
        return -1;
    }

    owner = WorkerAddress::kOwnerUCP;

    return SUCCESS;
}

int ucp::WorkerAddress::finalize(ucp_worker_h worker)
{
    if ((WorkerAddress::kOwnerUCP == owner) and likely(len >= 0))
        ucp_worker_release_address(worker, addr);

    addr = nullptr;
    len = 0;
    owner = WorkerAddress::kOwnerUndefined;

    return SUCCESS;
}

int ucp::WorkerAddress::resize(std::uint64_t n)
{
    len = n;
    addr = reinterpret_cast<ucp_address_t*>(new char[n]);
    owner = WorkerAddress::kOwnerSelf;

    return SUCCESS;
}

ucp::Connection::Connection()
: ep_(nullptr)
{
}

ucp::Connection::Connection(Connection&& other)
: remote_addr_(std::move(other.remote_addr_)),
  ep_(std::exchange(other.ep_, nullptr))
{
}

int ucp::Connection::create_ucp_ep(ucp_worker_h worker)
{
    ucp_ep_params_t params;

    std::memset(&params, 0, sizeof(params));
    params.field_mask = UCP_EP_PARAM_FIELD_REMOTE_ADDRESS;
    params.address = remote_addr_.addr;

    auto err = ucp_ep_create(worker, &params, &ep_);
    if (UCS_OK != err) {
        error("ucs_ep_create() returned %lld", (long long )err);
        return -1;
    }

    return SUCCESS;
}

int ucp::Connection::close_ucp_ep(ucp_worker_h worker)
{
	if (nullptr == ep_) return SUCCESS;

	/* Request closure of connection */
	ucs_status_ptr_t status_or_request=ucp_ep_close_nb(ep_,UCP_EP_CLOSE_MODE_FLUSH);
#if defined(DEBUG_UCP)
	printf("ucp close connection status=%zu\n",(std::size_t)status_or_request);std::fflush(stdout);
#endif
	if(status_or_request==(ucs_status_ptr_t*)UCS_OK){ // Connection closed
		/* Start Fix for bug: "UCX WARN unexpected tag-receive descriptor was not matched" on JUSUF */
		// It is not quite clear why this is necessary, it seem as if this updates some internal state
		// such that the related error message is not printed when the program finalizes.
#if defined(DEBUG_UCP)
	printf("ucp close connection status=OK\n");std::fflush(stdout);
#endif
		ucp_worker_progress(worker);
		/* End   Fix for bug: "UCX WARN unexpected tag-receive descriptor was not matched" on JUSUF */
		return SUCCESS;
	}
	if(UCS_PTR_IS_ERR(status_or_request)) { //Request to close connection failed
		error("ucp_ep_close_nb() returned an error");
		return -1;
	} //Otherwise request to close connection has been started and will close.
	auto err = ucp::wait_for_request_completion(worker, status_or_request);
	if (err) return err;
	ucp_request_free(status_or_request);
	return SUCCESS;
}

ucp::VirtualClusterImpl::VirtualClusterImpl(VirtualCluster* obj)
: obj_(obj)
{
    std::memset(&ctx_, 0, sizeof(ctx_));
    std::memset(&worker_, 0, sizeof(worker_));
}

int ucp::VirtualClusterImpl::init_ucp_context()
{
    ucp_params_t params;

    std::memset(&params, 0, sizeof(params));
    params.field_mask = UCP_PARAM_FIELD_FEATURES;
    params.features   = UCP_FEATURE_TAG;

    auto err = ucp_init(&params, nullptr, &ctx_);
    if (UCS_OK != err) {
        error("ucp_init() returned %lld", (long long )err);
        return -1;
    }

    ucx_request_size_ = VirtualClusterImpl::kReqSize;

    return SUCCESS;
}

int ucp::VirtualClusterImpl::init_ucp_worker()
{
    ucp_worker_params_t params;

    std::memset(&params, 0, sizeof(params));
    params.field_mask = UCP_WORKER_PARAM_FIELD_THREAD_MODE;
    params.thread_mode = UCS_THREAD_MODE_SINGLE;

    auto err = ucp_worker_create(ctx_, &params, &worker_);
    if (UCS_OK != err) {
        error("ucp_worker_create() returned %lld", (long long )err);
        return -1;
    }

    return worker_addr_.init(worker_);
}

int ucp::VirtualClusterImpl::initialize_ucp_lib()
{
    auto err = init_ucp_context();
    if (err)
        return -1;

    err = init_ucp_worker();
    if (err)
        return -1;

    return SUCCESS;
}

int ucp::VirtualClusterImpl::finalize_ucp_lib()
{
    worker_addr_.finalize(worker_);
    ucp_worker_destroy(worker_);
    ucp_cleanup(ctx_);

    return SUCCESS;
}

int ucp::VirtualClusterImpl::connect_ep(int rank, int other, int server)
{
    auto& conn = conns_[other];

    EXEC_NOFAIL(exchange_address(rank, other, server));

    EXEC_NOFAIL(conn.create_ucp_ep(worker_));

    return SUCCESS;
}

template<typename T>
static int send_recv(VirtualCluster* obj, 
             int rank, int other, int server, 
             T* sendbuf, std::uint64_t sendlen,
             T* recvbuf, std::uint64_t recvlen)
{
    if (rank == server) {
        EXEC_NOFAIL(obj->send(other, sendbuf, sendlen));

        EXEC_NOFAIL(obj->recv(other, recvbuf, recvlen));
    } else {
        EXEC_NOFAIL(obj->recv(other, recvbuf, recvlen));

        EXEC_NOFAIL(obj->send(other, sendbuf, sendlen));
    }

    return SUCCESS;
}

int ucp::VirtualClusterImpl::exchange_address(int rank, int other, int server)
{
    auto& conn = conns_[other];

    std::uint64_t k = worker_addr_.len;
    std::uint64_t n;

    assert(k > 0);

    EXEC_NOFAIL(send_recv(obj_, rank, other, server, &k, 1, &n, 1));

    if (k != n) {
        warn("UCP worker address lengths do not match. "
             "Expected %lld, got %lld", k, n);
    }

    conn.remote_addr_.resize(n);

    EXEC_NOFAIL(send_recv(obj_, rank, other, server, reinterpret_cast<char*>(worker_addr_.addr), worker_addr_.len, reinterpret_cast<char*>(conn.remote_addr_.addr), conn.remote_addr_.len));

#if defined(DEBUG_UCP)
    std::printf("rank %d: local: 0x%x 0x%x 0x%x 0x%x\n", rank,
        reinterpret_cast<char*>(worker_addr_.addr)[0],
        reinterpret_cast<char*>(worker_addr_.addr)[1],
        reinterpret_cast<char*>(worker_addr_.addr)[2],
        reinterpret_cast<char*>(worker_addr_.addr)[3]);
    std::printf("rank %d: remote: 0x%x 0x%x 0x%x 0x%x\n", rank,
        reinterpret_cast<char*>(conn.remote_addr_.addr)[0],
        reinterpret_cast<char*>(conn.remote_addr_.addr)[1],
        reinterpret_cast<char*>(conn.remote_addr_.addr)[2],
        reinterpret_cast<char*>(conn.remote_addr_.addr)[3]);
    std::fflush(stdout);
#endif

    return SUCCESS;
}

int ucp::VirtualClusterImpl::connect()
{
    const auto size = obj_->size();
    const auto rank = obj_->rank();

    // conns_.reserve(size());
    // for (auto i = 0; i < size(); ++i)
    //         conns_.push_back(ibverbs::Connection(ctx_, i));
    conns_.resize(obj_->size());

    auto err = connect_all_to_all_via_ring(size, rank, 
        [this, rank] (int other) -> int {
            return connect_ep(rank, other, other);
        }, 
        [this, rank] (int other) -> int {
            return connect_ep(rank, other, rank);
        });

    return err;
}

int ucp::VirtualClusterImpl::disconnect()
{
    for(auto& conn: conns_)
        conn.close_ucp_ep(worker_);    // Ignore error

    return SUCCESS;
}

int ucp::VirtualClusterImpl::wait_for_request_completion(void* req)
{
    return ucp::wait_for_request_completion(worker_, req);
}

VirtualClusterUCP::VirtualClusterUCP(const std::string& name)
: VirtualClusterWithHelper(name), impl_(new ucp::VirtualClusterImpl(this))
{
}

int VirtualClusterUCP::init()
{
    auto ret = linktest_minipmi_context_borrow(&impl_->pmi);
#if 1 == HAVE_MINIPMI
    if (ret) {
        error("linktest_minipmi_context_borrow() failed.");
        return ERROR;
    }
#endif

    EXEC_NOFAIL(read_ucp_environ());

    set_helper_pointer(VirtualCluster::factory(env.helper));
    EXEC_NOFAIL(helper_->init());

    ret = impl_->initialize_ucp_lib();
    if (ret)
        goto out;

    ret = impl_->connect();
    if (ret)
        goto out;

    return SUCCESS;

out:
    // ignore return values at this point
    helper_->finalize();
    delete helper_;

    return ERROR;
}

int VirtualClusterUCP::finalize()
{
    // Proceed in case of errors but tell the caller.
    
    auto ret1 = impl_->disconnect();
    auto ret2 = impl_->finalize_ucp_lib();

    auto ret3 = helper_->finalize();
    delete helper_;

    auto ret4 = linktest_minipmi_context_return(impl_->pmi);
#if 1 == HAVE_MINIPMI
    if (ret4) {
        error("linktest_minipmi_context_return() failed.");
    }
#endif

    impl_.reset();

    return ret1 | ret2 | ret3 | ret4;
}

ucs_status_t VirtualClusterUCP::kpingpong_isend(int other,         //Destination Rank
                                                MemoryBuffer &buf, //Message-buffer pointer
                                                int tag0           //Tag
                                               ){
#if defined(DEBUG_UCP)
	printf("rank=%d ucp isend from=%d to=%d\n",rank(),rank(),other);std::fflush(stdout);
#endif
	const ucs_status_t status=ucp_tag_send_nbr(impl_->conns_[other].ep_,                //Destination endpoint handle
	                                           buf.pointer<void>(),                     //Message-buffer pointer
	                                           buf.len(),                               //Message-buffer length
	                                           ucp_dt_make_contig(1),                   //Message-buffer datatype
	                                           (ucp_tag_t) tag0,                        //Message tag
	                                           impl_->send_req+impl_->ucx_request_size_ //Request handle
	                                          );
#if defined(DEBUG_UCP)
	printf("isend status=%d\n",status);std::fflush(stdout);
#endif
	if(UCS_STATUS_IS_ERR(status)){ //Failed to request send
		error("ucp_tag_send_nbr() failed with error = %d.", status);
	}
	return status;
}

ucs_status_t VirtualClusterUCP::kpingpong_irecv(int other,         //Destination Rank
                                                MemoryBuffer &buf, //Message-buffer pointer
                                                int tag0           //Tag
                                               ){
#if defined(DEBUG_UCP)
	printf("rank=%d ucp irecv from=%d to=%d\n",rank(),other,rank());std::fflush(stdout);
#endif
	const ucs_status_t status=ucp_tag_recv_nbr(impl_->worker_,                          //Receiving Worker
	                                           buf.pointer<void>(),                     //Message-buffer pointer
	                                           buf.len(),                               //Message-buffer length
	                                           ucp_dt_make_contig(1),                   //Message-buffer datatype
	                                           (ucp_tag_t) tag0,                        //Message tag
	                                           ~((ucp_tag_t)0),                         //Message-tag bit mask
	                                           impl_->recv_req+impl_->ucx_request_size_ //Request handle
	                                          );
	if (UCS_STATUS_IS_ERR(status)){ //Failed to request recv
		error("ucp_tag_recv_nbr() failed with error = %d.", status);
	}
	return status;
}

int VirtualClusterUCP::kpingpong_send(int other, MemoryBuffer& buf, int tag0)
{
#if defined(DEBUG_UCP)
	printf("rank=%d ucp send from=%d to=%d\n",rank(),rank(),other);std::fflush(stdout);
#endif
	const ucs_status_t status = kpingpong_isend(other, buf, tag0);
	if(status==UCS_OK) return SUCCESS;
	if(UCS_STATUS_IS_ERR(status)) return -1;
	else return impl_->wait_for_request_completion(impl_->send_req+impl_->ucx_request_size_);
}

int VirtualClusterUCP::kpingpong_recv(int other, MemoryBuffer& buf, int tag0)
{
#if defined(DEBUG_UCP)
	printf("rank=%d ucp recv from=%d to=%d\n",rank(),other,rank());std::fflush(stdout);
#endif
	const ucs_status_t status = kpingpong_irecv(other, buf, tag0);
	if(UCS_STATUS_IS_ERR(status)) return -1;
	return impl_->wait_for_request_completion(impl_->recv_req + impl_->ucx_request_size_);
}

int VirtualClusterUCP::kpingong_waitall(ucs_status_t status_recv,ucs_status_t status_send)
{
	int err0, err1;
	if(status_recv==UCS_INPROGRESS) err0 = impl_->wait_for_request_completion(impl_->recv_req + impl_->ucx_request_size_);
	else err0=0;
	if(status_send==UCS_INPROGRESS) err1 = impl_->wait_for_request_completion(impl_->send_req + impl_->ucx_request_size_);
	else err1=0;
	return err0 || err1;
}

int VirtualClusterUCP::kpingpong(const int from, const int to, MemoryBuffer& buf,
                                 const int num_msg, double* const timing)
{
    static const int PING = 100;
    static const int PONG = 101;

    double tv = 0;

    if(rank() == from) {
        tv = walltime();

        for (auto i = 0; i < num_msg; ++i) {
            EXEC_NOFAIL(kpingpong_send(to, buf, PING));
            EXEC_NOFAIL(kpingpong_recv(to, buf, PONG));
        }

        tv = walltime() - tv;
        tv = tv / (2.0*num_msg);
    } else if(rank() == to) {
        for (auto i = 0; i < num_msg; ++i) {
            EXEC_NOFAIL(kpingpong_recv(from, buf, PING));
            EXEC_NOFAIL(kpingpong_send(from, buf, PONG));
        }

    } else {
        error("Invalid rank.");
        return ERROR;
    }

    if(timing)
        *timing = tv;

    return SUCCESS;
}

int VirtualClusterUCP::kUniDir(const int from, const int to,
                                MemoryBuffer& buf1, MemoryBuffer& buf2,
                                const int num_msg, double* const timing,
                                const bool doBarrier
                               ){
    fatal("not yet implemented.");
    return ERROR;
}
int VirtualClusterUCP::kUniDirMultiBuf(const int from,const int to,
                                       MemoryBufferMulti& buf_multi, MemoryBuffer& buf2,
                                       const int num_msg, double* const timing,
                                       const bool doBarrier
                                      ){
    fatal("not yet implemented.");
    return ERROR;
}
int VirtualClusterUCP::kUniDirLimitedMultiBuf(const int from,const int to,
                                              MemoryBufferMulti& buf_multi, MemoryBuffer& buf2,
                                              const int num_msg, double* const timing,
                                              const bool doBarrier
                                             ){
    fatal("not yet implemented.");
    return ERROR;
}

int VirtualClusterUCP::kbipingpong(const int from, const int to,
                                   MemoryBuffer& buf1, MemoryBuffer& buf2,
                                   const int num_msg, double* const timing){
    static const int PING = 100;
    static const int PONG = 101;

    double tv = 0;

    if(rank() == from) {
        tv = walltime();

        for (auto i = 0; i < num_msg; ++i) {

            ucs_status_t status_recv = kpingpong_irecv(to, buf2, PONG);
            if (UCS_STATUS_IS_ERR(status_recv)) {
                fatal("kpingpong_irecv() failed.");
                return ERROR;
            }

            ucs_status_t status_send = kpingpong_isend(to, buf1, PING);
            if (UCS_STATUS_IS_ERR(status_send)) {
                fatal("kpingpong_isend() failed.");
                return ERROR;
            }

            auto ret0 = kpingong_waitall(status_recv, status_send);
            if (ret0) {
                fatal("kpingong_waitall() failed.");
                return ERROR;
            }
        }

        tv = walltime() - tv;
        tv = tv / (2.0*num_msg);
    } else if(rank() == to) {
        for (auto i = 0; i < num_msg; ++i) {
            ucs_status_t status_recv = kpingpong_irecv(to, buf2, PING);
            if (UCS_STATUS_IS_ERR(status_recv)) {
                fatal("kpingpong_irecv() failed.");
                return ERROR;
            }

            ucs_status_t status_send = kpingpong_isend(to, buf1, PONG);
            if (UCS_STATUS_IS_ERR(status_send)) {
                fatal("kpingpong_isend() failed.");
                return ERROR;
            }

            auto ret0 = kpingong_waitall(status_recv,status_send);
            if (ret0) {
                fatal("kpingong_waitall() failed.");
                return ERROR;
            }
        }
    }

    if(timing)
        *timing = tv;

    return SUCCESS;
}
