/****************************************************************************
**  LinkTest                                                               **
*****************************************************************************
**  Copyright (c) 2008-2022                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#include <linktest_DoNotIncludeInMain.h>
#include <linktest.h>
#include <linktest_2_0_0.h>
#include <linktest_helper.h>

/*********************************/
/* Constant File-Scope Variables */
/*********************************/
#define LINKTEST_MAPPING_MAX_LOCATION_LEN_2_0_0     256
#define LINKTEST_MAPPING_MAX_DIM_2_0_0               15
#define LINKTEST_ID_STR_2_0_0                   "LKTST"

/**************/
/* Structures */
/**************/
struct LinkTestInfo_2_0_0{ //Description for LinkTest 2.0.0
	int32_t iInfo[  10]; /* Auxiliary Integer Information
	                      * istats[ 0]: The number of messages
	                      *             timings were averaged
	                      *             over.
	                      * istats[ 1]: The message size sent
	                      *             between tasks.
	                      * istats[ 2]: If non-zero all-to-all
	                      *             tests were performed at
	                      *             the beginning and end of
	                      *             each iteration (?).
	                      * istats[ 3]: CollectPNum
	                      * istats[ 4]: If non-zero main testing
	                      *             was performed in serial.
	                      * istats[ 5]: If non-zero step order
	                      *             was randomized before
	                      *             tests were performed.
	                      * istats[ 6]: Unused
	                      * istats[ 7]: Unused
	                      * istats[ 8]: Unused
	                      * istats[ 9]: Major Version
	                      */
	double  dInfo[  10]; /* Global Summary Information
	                      * dstats[ 0]: Minimum average point-to-
	                      *             point time between two
	                      *             tasks.
	                      * dstats[ 1]: Average point-to-point
	                      *             across all tasks.
	                      * dstats[ 2]: Maximum average point-to-
	                      *             point time between two
	                      *             tasks.
	                      * dstats[ 3]: Minimum all-to-all time
	                      *             across all tasks.
	                      * dstats[ 4]: Average all-to-all time
	                      *             across all tasks.
	                      * dstats[ 5]: Maximum all-to-all time
	                      *             across all tasks.
	                      * dstats[ 6]: Unused
	                      * dstats[ 7]: Unused
	                      * dstats[ 8]: Unused
	                      * dstats[ 9]: Unused
	                      */
};

/*****************************/
/* Read LinkTest Information */
/*****************************/
int readInfo_2_0_0(struct SionFile* const fh, const int swap, const Version version, struct LinkTestInfo_2_0_0* const linkTestInfo){
	size_t n;

	fseek(fh->fp, LINKTEST_MAPPING_MAX_LOCATION_LEN_2_0_0+LINKTEST_MAPPING_MAX_DIM_2_0_0*sizeof(int32_t), SEEK_CUR);

	/* Read 32-bit Integer Information */
	n=fread(linkTestInfo->iInfo,sizeof(int32_t),10,fh->fp);
	if(UNLIKELY(n != 10)) {
		PyErr_SetString(PyExc_RuntimeError, "readInfo_2_0_0: Failed to read 32-bit integer LinkTest information.");
		return -1;
	}
	sion_swap(linkTestInfo->iInfo, linkTestInfo->iInfo, sizeof(int32_t), 10, swap);
	#ifdef DEBUG_OUTPUT
		printf("num_msg             =%d\n",linkTestInfo->iInfo[ 0]);
		printf("len_msg             =%d\n",linkTestInfo->iInfo[ 1]);
		printf("do_alltoall         =%d\n",linkTestInfo->iInfo[ 2]);
		printf("collectpnum         =%d\n",linkTestInfo->iInfo[ 3]);
		printf("do_serial           =%d\n",linkTestInfo->iInfo[ 4]);
		printf("do_mix              =%d\n",linkTestInfo->iInfo[ 5]);
		printf("major version repeat=%d\n",linkTestInfo->iInfo[ 9]);
		fflush(stdout);
	#endif

	/* Check If Major Version Is Stored In Last LinkTest Information Integer */
	if(UNLIKELY(linkTestInfo->iInfo[9]!=version[0])){
		PyErr_SetString(PyExc_RuntimeError,"readInfo_2_0_0: Last LinkTest Integer Information Is Not Major Version.");
		return -1;
	}

	/* Read Double-Precision Floating-Point Information */
	n=fread(linkTestInfo->dInfo,sizeof(double),10,fh->fp);
	if(UNLIKELY(n != 10)) {
		PyErr_SetString(PyExc_RuntimeError,"readInfo_2_0_0: Failed to read double LinkTest information.");
		return -1;
	}
	sion_swap(linkTestInfo->dInfo, linkTestInfo->dInfo, sizeof(double), 10, swap);
	#ifdef DEBUG_OUTPUT
		printf("min_time            =%e\n",linkTestInfo->dInfo[0]);
		printf("max_time            =%e\n",linkTestInfo->dInfo[1]);
		printf("avg_time            =%e\n",linkTestInfo->dInfo[2]);
		if(linkTestInfo->iInfo[ 2]!=0){
			printf("min_time_a2a        =%e\n",linkTestInfo->dInfo[3]);
			printf("max_time_a2a        =%e\n",linkTestInfo->dInfo[4]);
			printf("avg_time_a2a        =%e\n",linkTestInfo->dInfo[5]);
		}
		fflush(stdout);
	#endif

	/* Correct Average Times Assuming 1 Iteration */
	linkTestInfo->dInfo[2]/=(double)(fh->size-1);
	linkTestInfo->dInfo[5]/=2.0;

	/* Check if Multiple Iterations Occurred [NOT PERFECT]*/
	if (linkTestInfo->dInfo[2] > linkTestInfo->dInfo[1]){
		PyErr_SetString(PyExc_RuntimeWarning,"readInfo_2_0_0: Average LinkTest time greater than maximum time. Multiple Iterations suspected.");
		return -1;
	}

	return 0;
}

/*******************************************/
/* Import LinkTest Information Into Python */
/*******************************************/
int importInfo_2_0_0(const Version version, struct LinkTestInfo_2_0_0* const linkTestInfo, PyObject* const* const pyLinkTestInfo){
	int err;

	/* Set LinkTest Version */
	/* TODO: There is a potential bug here. %d is used for a digit type,
	 * however this may be insufficiently long for 32-bit version numbers.
	 * This can cause errors. Python, however, does not support the standard
	 * C way of using PRId32.
	 */
	err=PyDict_SetItem(*pyLinkTestInfo,
	                   PyUnicode_FromString("Version"),
	                   PyUnicode_FromFormat("%d.%d.%d",version[0],version[1],version[2])
	                  );
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_0_0: Failed to add version to LinkTestInfo dictionary.");
		return -1;
	}

	/* Set Integer Information */
	// NOTE: Casting to a long int from an int32_t to a long int should
	// be safe since a long int on standard compliant compilers have a
	// minimum size of 32 bytes.
	err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Number of messages"            ),PyLong_FromLong((long int)linkTestInfo->iInfo[0]));
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_0_0: Failed to add numberOfMessages to LinkTestInfo dictionary.");
		return -1;
	}
	err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Message size"                  ),PyLong_FromLong((long int)linkTestInfo->iInfo[1]));
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_0_0: Failed to add lengthOfMessages to LinkTestInfo dictionary.");
		return -1;
	}
	err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Test MPI all-to-all?"          ),PyLong_FromLong((long int)linkTestInfo->iInfo[2]));
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_0_0: Failed to add doAlltoall to LinkTestInfo dictionary.");
		return -1;
	}
	err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Test serially?"                ),PyLong_FromLong((long int)linkTestInfo->iInfo[4]));
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_0_0: Failed to add doSerial to LinkTestInfo dictionary.");
		return -1;
	}
	err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Randomize step order?"         ),PyLong_FromLong((long int)linkTestInfo->iInfo[5]));
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_0_0: Failed to add doMix to LinkTestInfo dictionary.");
		return -1;
	}

	return 0;
}

/*******************************/
/* Read & Import LinkTest Data */
/*******************************/
int readImportData_2_0_0(struct SionFile* const fh, const int swap, struct LinkTestInfo_2_0_0* const linkTestInfo, PyObject* const* const pyLinkTestInfo, PyObject* const* const pyLinkTestData){
	char hostname[LINKTEST_MAPPING_MAX_LOCATION_LEN_2_0_0];
	int err;
	int32_t rank;
	size_t n;
	npy_intp len[2];
	void *ptr;

	//TODO: Handle Errors

	/* Set Python Array Sizes */
	len[0] = fh->size;
	len[1] = fh->size;

	/* Initialize Hostnames */
	PyObject* const pyHosts = PyList_New((Py_ssize_t)fh->size);

	/* Initialize Timings, Access Pattern & All-To-All Timings */
	PyObject* const pyTimings = NewNumericPyArray(NPY_DOUBLE,2,len);
	PyObject* const pyAccesspattern = NewNumericPyArray(NPY_INT,2,len);
	PyObject* const pyA2ATimings   =(linkTestInfo->iInfo[2]!=0)?NewNumericPyArray(NPY_DOUBLE, 2, len):NULL; //Silences maybe-used-uninitialized warning

	/* Initialize Serial Retest Timings */
	PyObject* const pySerialTimingsNew=(len[0]>0)?NewNumericPyArray(NPY_DOUBLE, 1, len):NULL;
	PyObject* const pySerialTimingsOld=(len[0]>0)?NewNumericPyArray(NPY_DOUBLE, 1, len):NULL;
	PyObject* const pySerialFrom      =(len[0]>0)?NewNumericPyArray(NPY_INT   , 1, len):NULL;
	PyObject* const pySerialTo        =(len[0]>0)?NewNumericPyArray(NPY_INT   , 1, len):NULL;

	for (rank = 0; rank < fh->size; ++rank) { //Loop over the ranks, each one was saved to its own page
		/* Seek To Start Of Data */
		sion_seek(fh->sid, rank, 0, 0); //Seek to start of rank
		fseek(fh->fp, strlen(LINKTEST_ID_STR_2_0_0) + 3*sizeof(int32_t), SEEK_CUR); //Seek past ID string and version

		/* Read Host Name (Multiple Tasks May Have The Same Host Name */
		n=fread(hostname, sizeof(char), LINKTEST_MAPPING_MAX_LOCATION_LEN_2_0_0, fh->fp);
				if (UNLIKELY((int32_t)n != LINKTEST_MAPPING_MAX_LOCATION_LEN_2_0_0)) {
			PyErr_SetString(PyExc_RuntimeError, "readImportData_2_0_0: Failed to read/import Hostname.");
			return -1;
		}
		sion_swap(&hostname, &hostname, sizeof(char), 1, swap);
		#ifdef DEBUG_OUTPUT
			printf("Hostname[%d]          =%s\n",rank,hostname);fflush(stdout);
		#endif
		PyList_SET_ITEM(pyHosts,(Py_ssize_t)rank,PyBytes_FromStringAndSize(hostname,(Py_ssize_t)LINKTEST_MAPPING_MAX_LOCATION_LEN_2_0_0));

		/* Seek Past Header Area & LinkTest Information*/
		fseek(fh->fp, LINKTEST_MAPPING_MAX_DIM_2_0_0*sizeof(int32_t), SEEK_CUR);
		if (0 == rank) {
			fseek(fh->fp, 10*sizeof(int32_t) + 10*sizeof(double), SEEK_CUR);
		}

		/* Read LinkTest Timings */
		ptr=PyArray_GETPTR2((PyArrayObject *)pyTimings      , rank, 0);
		n=fread(ptr, sizeof(double), fh->size, fh->fp);
		if (UNLIKELY((int32_t)n != fh->size)) {
			PyErr_SetString(PyExc_RuntimeError, "readImportData_2_0_0: Failed to read/import LinkTest timings.");
			return -1;
		}
		sion_swap(ptr, ptr, sizeof(double), fh->size, swap);
		for (n=0;(int32_t)n<fh->size;n++){
			if(((double*)ptr)[n]==0.0) ((double*)ptr)[n]=nan("");
		}
		#ifdef DEBUG_OUTPUT
			printf("timings[%d]          =%e",rank,((double*)ptr)[0]);
			for(n=1;(int32_t)n<fh->size;n++) printf(",%e",((double*)ptr)[n]);
			printf("\n");
			fflush(stdout);
		#endif
		/* Read LinkTest Access Pattern */
		ptr=PyArray_GETPTR2((PyArrayObject *)pyAccesspattern, rank, 0);
		n=fread(ptr, sizeof(int   ), fh->size, fh->fp);
		if (UNLIKELY((int32_t)n != fh->size)) {
			PyErr_SetString(PyExc_RuntimeError, "readImportData_2_0_0: Failed to read/import LinkTest access pattern.");
			return -1;
		}
		sion_swap(ptr, ptr, sizeof(int   ), fh->size, swap);
		#ifdef DEBUG_OUTPUT
			printf("accesspattern[%d]    =%d",rank,((int32_t*)ptr)[0]);
			for(n=1;(int32_t)n<fh->size;n++) printf(",%d",((int32_t*)ptr)[n]);
			printf("\n");
			fflush(stdout);
		#endif
		/* Read LinkTest All-To-All Timings */
		if (linkTestInfo->iInfo[2]){
			ptr=PyArray_GETPTR2((PyArrayObject *)pyA2ATimings, rank,0);
			n=fread(ptr, sizeof(double), fh->size, fh->fp);
			if (UNLIKELY((int32_t)n != fh->size)) {
				PyErr_SetString(PyExc_RuntimeError, "readImportData_2_0_0: Failed to read/import all-to-all timings.");
				return -1;
			}
			sion_swap(ptr, ptr, sizeof(double), fh->size, swap);
			#ifdef DEBUG_OUTPUT
				printf("1st a2a time [%d]    =%f\n",rank,((double*)ptr)[0]);fflush(stdout);
			#endif
		}
		if (0 == rank) {
			ptr=PyArray_GETPTR1((PyArrayObject *)pySerialTimingsNew, 0);
			n=fread(ptr, sizeof(double), fh->size, fh->fp);
			if (UNLIKELY((int32_t)n != fh->size)) {
				PyErr_SetString(PyExc_RuntimeError, "readImportData_2_0_0: Failed to read/import retested serial timings.");
				return -1;
			}
			sion_swap(ptr, ptr, sizeof(double), fh->size, swap);
			for (n=0;(int32_t)n<fh->size;n++){
				if(((double*)ptr)[n]==0.0) ((double*)ptr)[n]=nan("");
			}
			#ifdef DEBUG_OUTPUT
				printf("serialTimingsNew[0] =%f",((double*)ptr)[0]);
				for(n=1;(int32_t)n<fh->size;n++) printf(",%f",((double*)ptr)[n]);
				printf("\n");
				fflush(stdout);
			#endif
			ptr=PyArray_GETPTR1((PyArrayObject *)pySerialTimingsOld, 0);
			n=fread(ptr, sizeof(double), fh->size, fh->fp);
			if (UNLIKELY((int32_t)n != fh->size)) {
				PyErr_SetString(PyExc_RuntimeError, "readImportData_2_0_0: Failed to read/import retest-needed timings.");
				return -1;
			}
			sion_swap(ptr, ptr, sizeof(double), fh->size, swap);
			for (n=0;(int32_t)n<fh->size;n++){
				if(((double*)ptr)[n]==0.0) ((double*)ptr)[n]=nan("");
			}
			#ifdef DEBUG_OUTPUT
				printf("serialTimingsOld[0] =%f",((double*)ptr)[0]);
				for(n=1;(int32_t)n<fh->size;n++) printf(",%f",((double*)ptr)[n]);
				printf("\n");
				fflush(stdout);
			#endif
			ptr=PyArray_GETPTR1((PyArrayObject *)pySerialFrom      , 0);
			n=fread(ptr, sizeof(int   ), fh->size, fh->fp);
			if (UNLIKELY((int32_t)n != fh->size)) {
				PyErr_SetString(PyExc_RuntimeError, "readImportData_2_0_0: Failed to read/import serial retest origin.");
				return -1;
			}
			sion_swap(ptr, ptr, sizeof(int   ), fh->size, swap);
			#ifdef DEBUG_OUTPUT
				printf("serialTimingsFrom[0]=%d",((int*)ptr)[0]);
				for(n=1;(int32_t)n<fh->size;n++) printf(",%d",((int*)ptr)[n]);
				printf("\n");
				fflush(stdout);
			#endif
			ptr=PyArray_GETPTR1((PyArrayObject *)pySerialTo        , 0);
			n=fread(ptr, sizeof(int   ), fh->size, fh->fp);
			if (UNLIKELY((int32_t)n != fh->size)) {
				PyErr_SetString(PyExc_RuntimeError, "readImportData_2_0_0: Failed to read/import serial retest destination.");
				return -1;
			}
			sion_swap(ptr, ptr, sizeof(int   ), fh->size, swap);
			#ifdef DEBUG_OUTPUT
				printf("serialTimingsTo[0]  =%d",((int*)ptr)[0]);
				for(n=1;(int32_t)n<fh->size;n++) printf(",%d",((int*)ptr)[n]);
				printf("\n");
				fflush(stdout);
			#endif
		}
	}

	sion_close(fh->sid);

	/* Set Hosts */
	err=PyDict_SetItem(*pyLinkTestInfo, PyUnicode_FromString("Hosts"                          ), pyHosts);
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"readImportData_2_0_0: Failed to add hosts to LinkTestData dictionary.");
		return -1;
	}

	/* Set Min/Max/Avg Times */
	err=PyDict_SetItem(*pyLinkTestData,PyUnicode_FromString("Minimum time"                  ),PyFloat_FromDouble(linkTestInfo->dInfo[0]));
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_0_0: Failed to add minTimeAll to LinkTestData dictionary.");
		return -1;
	}
	err=PyDict_SetItem(*pyLinkTestData,PyUnicode_FromString("Maximum time"                  ),PyFloat_FromDouble(linkTestInfo->dInfo[1]));
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_0_0: Failed to add maxTimeAll to LinkTestData dictionary.");
		return -1;
	}
	err=PyDict_SetItem(*pyLinkTestData,PyUnicode_FromString("Average time"                  ),PyFloat_FromDouble(linkTestInfo->dInfo[2]));
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_0_0: Failed to add avgTimeAll to LinkTestData dictionary.");
		return -1;
	}
	if(linkTestInfo->iInfo[2]!=0){ //Only execute if all-to-all testing was performed
		err=PyDict_SetItem(*pyLinkTestData,PyUnicode_FromString("Minimum all-to-all time"),PyFloat_FromDouble(linkTestInfo->dInfo[3]));
		if(err){
			PyErr_SetString(PyExc_RuntimeError,"importInfo_2_0_0: Failed to add a2aMinTimeAll to LinkTestData dictionary.");
			return -1;
		}
		err=PyDict_SetItem(*pyLinkTestData,PyUnicode_FromString("Maximum all-to-all time"),PyFloat_FromDouble(linkTestInfo->dInfo[4]));
		if(err){
			PyErr_SetString(PyExc_RuntimeError,"importInfo_2_0_0: Failed to add a2aMaxTimeAll to LinkTestData dictionary.");
			return -1;
		}
		err=PyDict_SetItem(*pyLinkTestData,PyUnicode_FromString("Average all-to-all time"),PyFloat_FromDouble(linkTestInfo->dInfo[5]));
		if(err){
			PyErr_SetString(PyExc_RuntimeError,"importInfo_2_0_0: Failed to add a2aAvgTimeAll to LinkTestData dictionary.");
			return -1;
		}
	}

	/* Set Full Timing Data */
	err=PyDict_SetItem(*pyLinkTestData, PyUnicode_FromString("Timings"                        ), pyTimings);
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"readImportData_2_0_0: Failed to add timings to LinkTestData dictionary.");
		return -1;
	}
	err=PyDict_SetItem(*pyLinkTestData, PyUnicode_FromString("Access Pattern"                 ), pyAccesspattern);
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"readImportData_2_0_0: Failed to add accessPatterns to LinkTestData dictionary.");
		return -1;
	}
	err=PyDict_SetItem(*pyLinkTestData, PyUnicode_FromString("New serially retested timings"  ), pySerialTimingsNew);
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"readImportData_2_0_0: Failed to add serialTimingsNew to LinkTestData dictionary.");
		return -1;
	}
	err=PyDict_SetItem(*pyLinkTestData, PyUnicode_FromString("Old serially retested timings"  ), pySerialTimingsOld);
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"readImportData_2_0_0: Failed to add serialTimingsOld to LinkTestData dictionary.");
		return -1;
	}
	err=PyDict_SetItem(*pyLinkTestData, PyUnicode_FromString("Serial-retest origin hosts"     ), pySerialFrom);
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"readImportData_2_0_0: Failed to add serialFrom to LinkTestData dictionary.");
		return -1;
	}
	err=PyDict_SetItem(*pyLinkTestData, PyUnicode_FromString("Serial-retest destination hosts"), pySerialTo);
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"readImportData_2_0_0: Failed to add serialTo to LinkTestData dictionary.");
		return -1;
	}
	if (linkTestInfo->iInfo[2]){
		err=PyDict_SetItem(*pyLinkTestData, PyUnicode_FromString("All-to-all timings"), pyA2ATimings);
		if(err){
			PyErr_SetString(PyExc_RuntimeError,"readImportData_2_0_0: Failed to add a2aTimings to LinkTestData dictionary.");
			return -1;
		}
	}

	return 0;
}

/**********************/
/* Read & Import Data */
/**********************/
int readImport_2_0_0(struct SionFile* const fh, const int swap, const Version version, PyObject** const pyLinkTestInfo, PyObject** const pyLinkTestData){
	struct LinkTestInfo_2_0_0 linkTestInfo;
	int err;
	/* Read LinkTest Information */
	err = readInfo_2_0_0(fh, swap, version, &linkTestInfo);
	if(err){
//		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_0_0: Failed to read the LinkTest information.");
		return err;
	}
	/* Import LinkTest Information Into Python */
	*pyLinkTestInfo = PyDict_New(); //Create LinkTest information dictionary
	err = importInfo_2_0_0(version, &linkTestInfo, pyLinkTestInfo);
	if(err){
//		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_0_0: Failed to populate the LinkTestInfo dictionary.");
		//TODO: Safely Deallocate Python Dictionary
		return err;
	}
	/* Read & Import LinkTest Data Into Python */
	*pyLinkTestData = PyDict_New(); //Create LinkTest data dictionary
	err = readImportData_2_0_0(fh, swap, &linkTestInfo, pyLinkTestInfo, pyLinkTestData);
	if(err){
//		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_0_0: Failed to read LinkTest data or to populate the LinkTestData dictionary.");
		//TODO: Safely Deallocate Python Dictionary
		return err;
	}
	PyErr_WarnEx(NULL,"LinkTest bisection runs prior to version 2.1.10 do not conform to a bisection test and average times may be wrong.\nCannot determine if this is a bisection run.",(Py_ssize_t )1);
	return(0);
}