/****************************************************************************
**  LinkTest                                                               **
*****************************************************************************
**  Copyright (c) 2008-2022                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#include <linktest_DoNotIncludeInMain.h>
#include <linktest.h>
#include <linktest_helper.h>

/*********************************/
/* Constant File-Scope Variables */
/*********************************/
#define LINKTEST_OUTPUT_DATETIME_LENGTH_2_1_16          32
#define GIT_HASH_SIZE_2_1_16                            41
#define LINKTEST_ID_STR_2_1_16                     "LKTST"
#define END_BLOCK_2_1_16                       "END_BLOCK"

/**************/
/* Structures */
/**************/
struct LinkTestInfo_2_1_16{ //Description for LinkTest 2.1.16
	char     hash[GIT_HASH_SIZE_2_1_16];  // LinkTest Git Hash
	char    *mode;                        // Virtual-Cluster Implementation
	int8_t   options[13];                 /* LinkTest Options
	                                       * Unused in Version [1.X.X & 2.0.0]
	                                       * options[ 0]: If non-zero all-to-all
	                                       *              tests were performed at
	                                       *              the beginning and end of
	                                       *              each iteration.
	                                       * options[ 1]: If non-zero bidirectional
	                                       *              tests were performed.
	                                       * options[ 2]: If non-zero unidirectional
	                                       *              tests were performed.
	                                       * options[ 3]: If non-zero bisection
	                                       *              tests were performed.
	                                       * options[ 4]: If non-zero steps were
	                                       *              mixed before tests were
	                                       *              performed.
	                                       * options[ 5]: If non-zero main testing
	                                       *              was performed in serial.
	                                       * options[ 6]: If non-zero output was
	                                       *              written to a SION file.
	                                       *              Always non-zero since
	                                       *              these options were
	                                       *              written to a SION file.
	                                       * options[ 7]: If non-zero the SION file
	                                       *              was generated using
	                                       *              parallelized output.
	                                       * options[ 8]: If non-zero messages were
	                                       *              sent from device memory,
	                                       *              likely GPU memory.
	                                       * options[ 9]: If non-zero multiple
	                                       *              buffers were used.
	                                       * options[10]: If non-zero the buffers
	                                       *              were filled with pseudo-
	                                       *              random numbers.
	                                       * options[11]: If non-zero buffers were
	                                       *              checked after each kernel
	                                       *              run.
	                                       * options[12]: Memory allocator type:
	                                       *                0: Invalid - ERROR
	                                       *                1: Default - ERROR
	                                       *                2: Memory-aligned malloc
	                                       *                3: Pinned-Memory Map
	                                       *                4: POSIX Memory-aligned malloc
	                                       *                5: CUDA Malloc
	                                       */
	uint64_t  iInfo[   9];                /* Auxiliary Integer Information
	                                       * istats[  0]: The number of messages
	                                       *              timings were averaged
	                                       *              over.
	                                       * istats[  1]: The message size sent
	                                       *              between tasks.
	                                       * istats[  2]: The number of warm-up
	                                       *              messages sent before
	                                       *              beginning timing.
	                                       * istats[  3]: CollectPNum
	                                       * istats[  4]: Number of serial retests
	                                       * istats[  5]: Number of memory buffers
	                                       *              if multiple memory
	                                       *              buffers were used.
	                                       * istats[  6]: Mersenne Twister seed
	                                       *              for buffer seeding. 
	                                       * istats[  7]: If non-zero indicates the
	                                       *              number of different task
	                                       *              permutations used for
	                                       *              averaging communication times.
	                                       * istats[  8]: MT_19937 seed for task
	                                       *              randomization.
	                                       */
};

/*****************************/
/* Read LinkTest Information */
/*****************************/
int readInfo_2_1_16(struct SionFile* const fh, const int swap, const Version version, struct LinkTestInfo_2_1_16* const linkTestInfo){
	size_t n;
	int32_t strSize;

	/* Git Hash */
	n=fread(linkTestInfo->hash, sizeof(char), GIT_HASH_SIZE_2_1_16, fh->fp);
	if (UNLIKELY(n != GIT_HASH_SIZE_2_1_16)) {
		PyErr_SetString(PyExc_RuntimeError, "readInfo_2_1_16: Failed to read git hash.");
		return -1;
	}
	sion_swap(linkTestInfo->hash, linkTestInfo->hash, sizeof(char), GIT_HASH_SIZE_2_1_16, swap);
	if (UNLIKELY(linkTestInfo->hash[0] == '\0')){
		PyErr_SetString(PyExc_RuntimeError, "readInfo_2_1_16: NULL git hashes are not supported.");
		return -1;
	}
	linkTestInfo->hash[GIT_HASH_SIZE_2_1_16-1]='\0'; //Make sure string is null terminated
	#ifdef DEBUG_OUTPUT
		printf("git hash                   =%s\n",linkTestInfo->hash);fflush(stdout);
	#endif

	/* Read Virtual Cluster Implementation */
	n=fread(&strSize, sizeof(int32_t), 1, fh->fp);
	if (UNLIKELY(n != 1)) {
		PyErr_SetString(PyExc_RuntimeError, "readInfo_2_1_16: Failed to read string sizes for git hash, datetime and mode.");
		return -1;
	}
	sion_swap(&strSize, &strSize, sizeof(int32_t), 1, swap);
	linkTestInfo->mode=malloc(strSize*sizeof(char)); //TODO: Dangerous if sizeof(char) mismatches between writing and reading system.
	n=fread(linkTestInfo->mode, sizeof(char), strSize, fh->fp);
	if (UNLIKELY((int32_t)n != strSize)) {
		PyErr_SetString(PyExc_RuntimeError, "readInfo_2_1_16: Failed to read virtual-cluster implementation.");
		return -1;
	}
	sion_swap(linkTestInfo->mode, linkTestInfo->mode, sizeof(char), strSize, swap);
	if (UNLIKELY(linkTestInfo->mode[0] == '\0')){
		PyErr_SetString(PyExc_RuntimeError, "readInfo_2_1_16: Virtual-cluster implementation NULL is not supported.");
		return -1;
	}
	linkTestInfo->mode[strSize-1]='\0'; //Add null terminator to string
	#ifdef DEBUG_OUTPUT
		printf("mode                       =%s\n",linkTestInfo->mode);fflush(stdout);
	#endif

	/* Read 32-bit Integer Options */
	n=fread(linkTestInfo->options, sizeof(int8_t), 13, fh->fp);
	if (UNLIKELY(n != 13)) {
		PyErr_SetString(PyExc_RuntimeError, "readInfo_2_1_16: Failed to read 32-bit integer LinkTest options.");
		return -1;
	}
	sion_swap(linkTestInfo->options, linkTestInfo->options, sizeof(int8_t), 13, swap);
	#ifdef DEBUG_OUTPUT
		printf("do_alltoall                =%" PRId8 "\n",linkTestInfo->options[ 0]);
		printf("do_bidir                   =%" PRId8 "\n",linkTestInfo->options[ 1]);
		printf("do_unidir                  =%" PRId8 "\n",linkTestInfo->options[ 2]);
		printf("do_bisection               =%" PRId8 "\n",linkTestInfo->options[ 3]);
		printf("do_mix                     =%" PRId8 "\n",linkTestInfo->options[ 4]);
		printf("do_serial                  =%" PRId8 "\n",linkTestInfo->options[ 5]);
		printf("do_nosion                  =%" PRId8 "\n",linkTestInfo->options[ 6]);
		printf("do_sion_par                =%" PRId8 "\n",linkTestInfo->options[ 7]);
		printf("do_use_gpus                =%" PRId8 "\n",linkTestInfo->options[ 8]);
		printf("use_multi_buf              =%" PRId8 "\n",linkTestInfo->options[ 9]);
		printf("randomize_buffers          =%" PRId8 "\n",linkTestInfo->options[10]);
		printf("check_buffers              =%" PRId8 "\n",linkTestInfo->options[11]);
		printf("mem alloc type             =%" PRId8 "\n",linkTestInfo->options[12]);
		fflush(stdout);
	#endif

	/* Read 32-bit Integer Information */
	n=fread(linkTestInfo->iInfo, sizeof(uint64_t), 9, fh->fp);
	if (UNLIKELY(n != 9)) {
		PyErr_SetString(PyExc_RuntimeError, "readInfo_2_1_16: Failed to read 32-bit integer LinkTest information.");
		return -1;
	}
	sion_swap(linkTestInfo->iInfo, linkTestInfo->iInfo, sizeof(uint64_t), 9, swap);
	#ifdef DEBUG_OUTPUT
		printf("num_msg                    =%" PRIu64 "\n",linkTestInfo->iInfo[ 0]);
		printf("len_msg                    =%" PRIu64 "\n",linkTestInfo->iInfo[ 1]);
		printf("num_warmup_msg             =%" PRIu64 "\n",linkTestInfo->iInfo[ 2]);
		printf("collectpnum                =%" PRIu64 "\n",linkTestInfo->iInfo[ 3]);
		printf("max_stest                  =%" PRIu64 "\n",linkTestInfo->iInfo[ 4]);
		printf("num_multi_buf              =%" PRIu64 "\n",linkTestInfo->iInfo[ 5]);
		printf("buf_mt_seed                =%" PRIu64 "\n",linkTestInfo->iInfo[ 6]);
		printf("num_randomize_tasks        =%" PRIu64 "\n",linkTestInfo->iInfo[ 7]);
		printf("seed_randomize_tasks       =%" PRIu64 "\n",linkTestInfo->iInfo[ 8]);
		fflush(stdout);
	#endif

	/* Check If 0/1 Multiple Buffer Were Specified */
	if(linkTestInfo->options[9] != 0){
		if(linkTestInfo->iInfo[5] == 0){
			PyErr_WarnEx(NULL,"Detected that multiple buffers were used, however, an unknown (0) number of buffers were used.\n ReadSION: The number of buffers used was likely equal to the number of messages. Interpret results with caution.",(Py_ssize_t )1);
		}else if(linkTestInfo->iInfo[5] == 1){
			PyErr_WarnEx(NULL,"Detected that multiple buffers were used, however, only one buffer was used.\n ReadSION: In this scenario multiple buffers should not have been used. Interpret results with caution.",(Py_ssize_t )1);
		}
	}

	/* Check If Mersenne Twister Seed Is Zero */
	if(linkTestInfo->options[10] != 0 && linkTestInfo->iInfo[6] == 0){
		PyErr_WarnEx(NULL,"Buffer content was to be randomized, however, the stored zero Mersenne Twister seed would have filled the buffers with zero. Interpret results with caution.",(Py_ssize_t )1);
	}

	return 0;
}

/*******************************************/
/* Import LinkTest Information Into Python */
/*******************************************/
int importInfo_2_1_16(const Version version, struct LinkTestInfo_2_1_16* const linkTestInfo, PyObject* const* const pyLinkTestInfo){
	int err;

	// Allocator Type
	enum AllocatorType{
		AllocatorInvalid           =0, //Invalid                     Allocator - ERROR
		AllocatorDefault           =1, //Default                     Allocator - ERROR
		AllocatorAlignedMalloc     =2, //Memory-aligned malloc       Allocator
		AllocatorPinnedMmap        =3, //Pinned--memory-map          Allocator
		AllocatorPOSIXAlignedMalloc=4, //POSIX Memory-aligned malloc Allocator
		AllocatorCUDA              =5  //CUDA maloc                  Allocator
	};

	/* Set LinkTest Version */
	/* TODO: There is a potential bug here. %d is used for a digit type,
	 * however this may be insufficiently long for 32-bit version numbers.
	 * This can cause errors. Python, however, does not support the standard
	 * C way of using PRId32.
	 */
	err=PyDict_SetItem(*pyLinkTestInfo,
	                   PyUnicode_FromString("Version"),
	                   PyUnicode_FromFormat("%d.%d.%d",version[0],version[1],version[2])
	                  );
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add Version to LinkTestInfo dictionary.");
		return -1;
	}

	/* Set LinkTest Git Hash */
	err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Git hash"                      ),PyUnicode_FromString(linkTestInfo->hash                ));
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add hash to LinkTestInfo dictionary.");
		return -1;
	}

	/* Set Virtual-Cluster Implementation */
	err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Mode"),PyUnicode_FromString(linkTestInfo->mode));
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add mode to LinkTestInfo dictionary.");
		return -1;
	}
	free(linkTestInfo->mode);

	/* Set Options */
	// NOTE: Casting to a long int from an int32_t to a long int should
	// be safe since a long int on standard compliant compilers have a
	// minimum size of 32 bytes.
	err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Test MPI all-to-all?"                ),PyLong_FromLong((long int)linkTestInfo->options[0]));
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add doAlltoall to LinkTestInfo dictionary.");
		return -1;
	}
	err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Do bidirectional tests?"             ),PyLong_FromLong((long int)linkTestInfo->options[1]));
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add doBidir to LinkTestInfo dictionary.");
		return -1;
	}
	err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Do unidirectional tests?"            ),PyLong_FromLong((long int)linkTestInfo->options[2]));
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add doUnidir to LinkTestInfo dictionary.");
		return -1;
	}
	err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Do a bisection test?"                ),PyLong_FromLong((long int)linkTestInfo->options[3]));
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add doBisec to LinkTestInfo dictionary.");
		return -1;
	}
	err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Randomize step order?"               ),PyLong_FromLong((long int)linkTestInfo->options[4]));
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add doMix to LinkTestInfo dictionary.");
		return -1;
	}
	err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Test serially?"                      ),PyLong_FromLong((long int)linkTestInfo->options[5]));
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add doSerial to LinkTestInfo dictionary.");
		return -1;
	}
	err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Do not store results in a SION file?"),PyLong_FromLong((long int)linkTestInfo->options[6]));
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add doWrite to LinkTestInfo dictionary.");
		return -1;
	}
	err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Write SION output in parallel?"      ),PyLong_FromLong((long int)linkTestInfo->options[7]));
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add doParallelIO to LinkTestInfo dictionary.");
		return -1;
	}
	err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Store messages in GPU memory?"       ),PyLong_FromLong((long int)linkTestInfo->options[8]));
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add useGPUs to LinkTestInfo dictionary.");
		return -1;
	}
	err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Use multiple memory buffers?"        ),PyLong_FromLong((long int)linkTestInfo->options[9]));
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add use_multi_buf to LinkTestInfo dictionary.");
		return -1;
	}
	err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Randomize memory-buffer content?"    ),PyLong_FromLong((long int)linkTestInfo->options[10]));
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add randomize_buffers to LinkTestInfo dictionary.");
		return -1;
	}
	err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Check memory-buffer content?"        ),PyLong_FromLong((long int)linkTestInfo->options[11]));
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add check_buffers to LinkTestInfo dictionary.");
		return -1;
	}
	switch(linkTestInfo->options[12]){
		case AllocatorInvalid:
			err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Allocator type"),PyUnicode_FromString("Invalid - ERROR"            ));
			break;
		case AllocatorDefault:
			err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Allocator type"),PyUnicode_FromString("Default - ERROR"            ));
			break;
		case AllocatorAlignedMalloc:
			err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Allocator type"),PyUnicode_FromString("Memory-aligned malloc"      ));
			break;
		case AllocatorPinnedMmap:
			err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Allocator type"),PyUnicode_FromString("Pinned memory-map"         ));
			break;
		case AllocatorPOSIXAlignedMalloc:
			err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Allocator type"),PyUnicode_FromString("POSIX memory-aligned malloc"));
			break;
		case AllocatorCUDA:
			err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Allocator type"),PyUnicode_FromString("CUDA malloc"                ));
			break;
		default:
			err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Allocator type"),PyUnicode_FromString("UNKNOWN"                    ));
			break;
	}
		if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add memory-buffer allocator to LinkTestInfo dictionary.");
		return -1;
	}

	/* Set Integer Information */
	err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Number of messages"               ),PyLong_FromLong((long int)linkTestInfo->iInfo[0]));
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add numberOfMessages to LinkTestInfo dictionary.");
		return -1;
	}
	err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Message size"                     ),PyLong_FromLong((long int)linkTestInfo->iInfo[1]));
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add lengthOfMessages to LinkTestInfo dictionary.");
		return -1;
	}
	err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Number of warm-up messages"       ),PyLong_FromLong((long int)linkTestInfo->iInfo[2]));
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add numberOfWarmUpMessages to LinkTestInfo dictionary.");
		return -1;
	}
	err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("collectPNum"                      ),PyLong_FromLong((long int)linkTestInfo->iInfo[3]));
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add collectPNum to LinkTestInfo dictionary.");
		return -1;
	}
	err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Number of serial retests"         ),PyLong_FromLong((long int)linkTestInfo->iInfo[4]));
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add numberOfSerialRetests to LinkTestInfo dictionary.");
		return -1;
	}
	if(linkTestInfo->options[9]){
		err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Number of multiple buffers"  ),PyLong_FromLong((long int)linkTestInfo->iInfo[5]));
		if(err){
			PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add num_multi_buf to LinkTestInfo dictionary.");
			return -1;
		}
	}
	if(linkTestInfo->options[10]){
		err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Buffer Mersenne Twister seed"),PyLong_FromLong((long int)linkTestInfo->iInfo[6]));
		if(err){
			PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add buf_mt_seed to LinkTestInfo dictionary.");
			return -1;
		}
	}
	err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Number of tests with randomized rank order"),PyLong_FromLong((long int)linkTestInfo->iInfo[7]));
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add num_randomize_tasks to LinkTestInfo dictionary.");
		return -1;
	}
	if(linkTestInfo->iInfo[7]){
		err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Seed for rank-order randomization"         ),PyLong_FromLong((long int)linkTestInfo->iInfo[8]));
		if(err){
			PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add seed_randomize_tasks to LinkTestInfo dictionary.");
			return -1;
		}
	}

	return 0;
}

/*******************************/
/* Read & Import LinkTest Data */
/*******************************/
int readImportData_2_1_16(struct SionFile* const fh, const int swap, struct LinkTestInfo_2_1_16* const linkTestInfo, PyObject* const* const pyLinkTestInfo, PyObject** pyLinkTestData){
	const size_t end_block_len=sizeof(END_BLOCK_2_1_16)/sizeof(char);
	char end_block[end_block_len];
	int err;
	int32_t rank, hostname_len;
	size_t n;
	size_t i;
	char* hostname;
	npy_intp len[2];
	void *ptr;

	const size_t num_iterations     =(linkTestInfo->iInfo[7]>1)?(size_t)linkTestInfo->iInfo[7]:1;
	const size_t numMinMaxAvgTimings=(linkTestInfo->options[0]!=0)?6:3;
	      double minMaxAvgTimings[num_iterations*numMinMaxAvgTimings];

	//TODO: Handle Errors

	/* Set Python Array Sizes */
	len[0] = fh->size;
	len[1] = fh->size;

	/* Initialize Hostnames & Core IDs */
	PyObject* const pyHosts = PyList_New((Py_ssize_t)fh->size);
	PyObject* const pyCore  = NewNumericPyArray(NPY_INT, 1, len);

	/* Initialize Start & End Time Strings */
	PyObject** const pyStartTime=(PyObject**)malloc(num_iterations*sizeof(PyObject));
	for(i=0;i<num_iterations;i++) pyStartTime[i]=PyDict_New();
	PyObject** const pyEndTime  =(PyObject**)malloc(num_iterations*sizeof(PyObject));
	for(i=0;i<num_iterations;i++) pyEndTime[i]  =PyDict_New();

	/* Initialize Timings, Access Pattern & All-To-All Timings */
	PyObject** const pyTimings      =(PyObject**)malloc(num_iterations*sizeof(PyObject));
	for(i=0;i<num_iterations;i++) pyTimings[i]      =NewNumericPyArray(NPY_DOUBLE,2,len); //A python double-precision floating-point array of the LinkTest point-to-point timings
	PyObject** const pyAccesspattern=(PyObject**)malloc(num_iterations*sizeof(PyObject));
	for(i=0;i<num_iterations;i++) pyAccesspattern[i]=NewNumericPyArray(NPY_UINT64,2,len); //A python integer array indicating the access order
	PyObject** const pyA2ATimings   =(linkTestInfo->options[0]!=0)?(PyObject**)malloc(num_iterations*sizeof(PyObject)):NULL; //Silences maybe-used-uninitialized warning
	if(linkTestInfo->options[0]!=0) for(i=0;i<num_iterations;i++) pyA2ATimings[i]   =NewNumericPyArray(NPY_DOUBLE, 1, len); //A python double-precision floating-point array of the linkTest all-to-all imings

	/* Initialize Serial Retest Timings */
	len[0] = linkTestInfo->iInfo[4];
	PyObject** const pySerialTimingsNew=(len[0]>0)?(PyObject**)malloc(num_iterations*sizeof(PyObject)):NULL;
	if(len[0]>0) for(i=0;i<num_iterations;i++) pySerialTimingsNew[i] = NewNumericPyArray(NPY_DOUBLE, 1, len);
	PyObject** const pySerialTimingsOld=(len[0]>0)?(PyObject**)malloc(num_iterations*sizeof(PyObject)):NULL;
	if(len[0]>0) for(i=0;i<num_iterations;i++) pySerialTimingsOld[i] = NewNumericPyArray(NPY_DOUBLE, 1, len);
	PyObject** const pySerialFrom      =(len[0]>0)?(PyObject**)malloc(num_iterations*sizeof(PyObject)):NULL;
	if(len[0]>0) for(i=0;i<num_iterations;i++) pySerialFrom[i]       = NewNumericPyArray(NPY_UINT64, 1, len);
	PyObject** const pySerialTo        =(len[0]>0)?(PyObject**)malloc(num_iterations*sizeof(PyObject)):NULL;
	if(len[0]>0) for(i=0;i<num_iterations;i++) pySerialTo[i]         = NewNumericPyArray(NPY_UINT64, 1, len);

	for (rank = 0; rank < fh->size; ++rank) { //Loop over the ranks, each one was saved to its own page
		/* Seek To Start Of Data */
		if(LIKELY(rank!=0)){
			sion_seek(fh->sid, rank, 0, 0); //Seek to start of rank
			fseek(fh->fp, strlen(LINKTEST_ID_STR_2_1_16), SEEK_CUR); //Seek past ID string
		}

		/* Read Host Name (Multiple Tasks May Have The Same Host Name */
		n=fread(&hostname_len, sizeof(uint32_t), 1, fh->fp);
		if (UNLIKELY(n != 1)) {
			PyErr_SetString(PyExc_RuntimeError, "readImportData_2_1_16: Failed to read/import hostname length.");
			return -1;
		}
		sion_swap(&hostname_len, &hostname_len, sizeof(uint32_t), 1, swap);
		hostname=(char*)malloc(hostname_len*sizeof(char));
		n=fread(hostname, sizeof(char), hostname_len, fh->fp);
		if (UNLIKELY((int32_t)n != hostname_len)) {
			PyErr_SetString(PyExc_RuntimeError, "readImportData_2_1_16: Failed to read/import Hostname.");
			return -1;
		}
		sion_swap(&hostname, &hostname, sizeof(char), 1, swap);
		#ifdef DEBUG_OUTPUT
			printf("Hostname[%d]                =%s\n",rank,hostname);fflush(stdout);
		#endif
		PyList_SET_ITEM(pyHosts,(Py_ssize_t)rank,PyBytes_FromStringAndSize(hostname,(Py_ssize_t)hostname_len));
		/* Read Pinned-Core ID */
		ptr=PyArray_GETPTR1((PyArrayObject *)pyCore, rank);
		n=fread(ptr, sizeof(int32_t), 1, fh->fp);
		if (UNLIKELY(n != 1)) {
			PyErr_SetString(PyExc_RuntimeError, "readImportData_2_1_16: Failed to read/import Pinned Core ID.");
			return -1;
		}
		sion_swap(ptr, ptr, sizeof(int32_t), 1, swap);
		#ifdef DEBUG_OUTPUT
			printf("cpuID[%d]                   =%d\n",rank,*(int32_t*)ptr);fflush(stdout);
		#endif

		for(i=0;i<num_iterations;i++){
			if(rank==0){
				/* Read Start Time */
				char datetime_string[LINKTEST_OUTPUT_DATETIME_LENGTH_2_1_16+1]; //TODO: Dangerous if sizeof(char) mismatches between writing and reading system.
				n=fread(datetime_string, sizeof(char), LINKTEST_OUTPUT_DATETIME_LENGTH_2_1_16, fh->fp);
				if (UNLIKELY(n != LINKTEST_OUTPUT_DATETIME_LENGTH_2_1_16)) {
					PyErr_SetString(PyExc_RuntimeError, "readInfo_2_1_16: Failed to read startDatetime.");
					return -1;
				}
				sion_swap(datetime_string, datetime_string, sizeof(char), LINKTEST_OUTPUT_DATETIME_LENGTH_2_1_16, swap);
				if (UNLIKELY(datetime_string[0] == '\0')){
					PyErr_SetString(PyExc_RuntimeError, "readInfo_2_1_16: startDatetime NULL is not supported.");
					return -1;
				}
				datetime_string[LINKTEST_OUTPUT_DATETIME_LENGTH_2_1_16]='\0'; //Ensure string is null terminated
				#ifdef DEBUG_OUTPUT
					printf("startDatetime[0,%zu]          =%s",i,datetime_string);
					fflush(stdout);
				#endif
				struct Datetime startDatetime;
				n=sscanf(datetime_string,
						"%2"SCNu8"/%2"SCNu8"/%"SCNu64" %2"SCNu8":%2"SCNu8":%2"SCNu8" UTC+0",
						&startDatetime.day,
						&startDatetime.month,
						&startDatetime.year,
						&startDatetime.hour,
						&startDatetime.minute,
						&startDatetime.second
						);
				if (UNLIKELY(n!=6)){
					PyErr_SetString(PyExc_RuntimeError, "readInfo_2_1_16: startDatetime is not a correctly formatted string.");
					return -1;
				}
				#ifdef DEBUG_OUTPUT
					printf("startDatetime.year[0,%zu]    =%" PRIu64 "\n",i,startDatetime.year  );
					printf("startDatetime.month[0,%zu]   =%" PRIu8  "\n",i,startDatetime.month );
					printf("startDatetime.day[0,%zu]     =%" PRIu8  "\n",i,startDatetime.day   );
					printf("startDatetime.hour[0,%zu]    =%" PRIu8  "\n",i,startDatetime.hour  );
					printf("startDatetime.minute[0,%zu]  =%" PRIu8  "\n",i,startDatetime.minute);
					printf("startDatetime.second[0,%zu]  =%" PRIu8  "\n",i,startDatetime.second);
					printf("startDatetime.timezone[0,%zu]=UTC+0\n"     ,i                      );
					fflush(stdout);
				#endif
				/* Import Start Time */
				err=PyDict_SetItem(pyStartTime[i],PyUnicode_FromString("Year"    ),PyLong_FromLong((long int)startDatetime.year  ));
				if(err){//TODO: Add iteration index to error message
					PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add DateTime_year to pyStartTime dictionary.");
					return -1;
				}
				err=PyDict_SetItem(pyStartTime[i],PyUnicode_FromString("Month"   ),PyLong_FromLong((long int)startDatetime.month ));
				if(err){
					PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add DateTime_month to pyStartTime dictionary.");
					return -1;
				}
				err=PyDict_SetItem(pyStartTime[i],PyUnicode_FromString("Day"     ),PyLong_FromLong((long int)startDatetime.day   ));
				if(err){
					PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add DateTime_day to pyStartTime dictionary.");
					return -1;
				}
				err=PyDict_SetItem(pyStartTime[i],PyUnicode_FromString("Hour"    ),PyLong_FromLong((long int)startDatetime.hour  ));
				if(err){
					PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add DateTime_hour to pyStartTime dictionary.");
					return -1;
				}
				err=PyDict_SetItem(pyStartTime[i],PyUnicode_FromString("Minute"  ),PyLong_FromLong((long int)startDatetime.minute));
				if(err){
					PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add DateTime_minute to pyStartTime dictionary.");
					return -1;
				}
				err=PyDict_SetItem(pyStartTime[i],PyUnicode_FromString("Second"  ),PyLong_FromLong((long int)startDatetime.second));
				if(err){
					PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add DateTime_second to pyStartTime dictionary.");
					return -1;
				}
				err=PyDict_SetItem(pyStartTime[i],PyUnicode_FromString("Timezone"),PyUnicode_FromFormat("UTC+0"));
				if(err){
					PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add DateTime_timezone to pyStartTime dictionary.");
					return -1;
				}

				/* Read Minimum, Maximum & Average LinkTest Timings */
				n=fread(&minMaxAvgTimings[i*numMinMaxAvgTimings], sizeof(double), numMinMaxAvgTimings, fh->fp);
				if (UNLIKELY(n != numMinMaxAvgTimings)) {
					PyErr_SetString(PyExc_RuntimeError, "readInfo_2_1_16: Failed to read minimum, maximum & average LinkTest timings.");
					return -1;
				}
				sion_swap(minMaxAvgTimings, minMaxAvgTimings, sizeof(double), numMinMaxAvgTimings, swap);
				#ifdef DEBUG_OUTPUT
					printf("min_time[0,%zu]              =%e\n",i,minMaxAvgTimings[i*numMinMaxAvgTimings+0]);
					printf("max_time[0,%zu]              =%e\n",i,minMaxAvgTimings[i*numMinMaxAvgTimings+1]);
					printf("avg_time[0,%zu]              =%e\n",i,minMaxAvgTimings[i*numMinMaxAvgTimings+2]);
					if(numMinMaxAvgTimings==6){
						printf("min_time_a2a[0,%zu]          =%e\n",i,minMaxAvgTimings[i*numMinMaxAvgTimings+3]);
						printf("max_time_a2a[0,%zu]          =%e\n",i,minMaxAvgTimings[i*numMinMaxAvgTimings+4]);
						printf("avg_time_a2a[0,%zu]          =%e\n",i,minMaxAvgTimings[i*numMinMaxAvgTimings+5]);
					}
					fflush(stdout);
				#endif
			}
			/* Read LinkTest Timings */
			ptr=PyArray_GETPTR2((PyArrayObject *)pyTimings[i]      , rank, 0);
			n=fread(ptr, sizeof(double), fh->size, fh->fp);
			if (UNLIKELY((int32_t)n != fh->size)) {
				PyErr_SetString(PyExc_RuntimeError, "readImportData_2_1_16: Failed to read/import LinkTest timings.");
				return -1;
			}
			sion_swap(ptr, ptr, sizeof(double), fh->size, swap);
			#ifdef DEBUG_OUTPUT
				printf("timings[%d,%zu]               =%e",rank,i,((double*)ptr)[0]);
				for(n=1;(int32_t)n<fh->size;n++) printf(",%e",((double*)ptr)[n]);
				printf("\n");
				fflush(stdout);
			#endif
			/* Read LinkTest Access Pattern */
			ptr=PyArray_GETPTR2((PyArrayObject *)pyAccesspattern[i], rank, 0);
			n=fread(ptr, sizeof(int64_t), fh->size, fh->fp);
			if (UNLIKELY((int64_t)n != fh->size)) {
				PyErr_SetString(PyExc_RuntimeError, "readImportData_2_1_16: Failed to read/import LinkTest access pattern.");
				return -1;
			}
			sion_swap(ptr, ptr, sizeof(uint64_t), fh->size, swap);
			#ifdef DEBUG_OUTPUT
				printf("accesspattern[%d,%zu]         =%" PRIu64,rank,i,((uint64_t*)ptr)[0]);
				for(n=1;(int32_t)n<fh->size;n++) printf(",%" PRIu64,((uint64_t*)ptr)[n]);
				printf("\n");
				fflush(stdout);
			#endif
			/* Read LinkTest All-To-All Timings */
			if (linkTestInfo->options[0]){
				ptr=PyArray_GETPTR1((PyArrayObject *)pyA2ATimings[i], rank);
				n=fread(ptr, sizeof(double), 1, fh->fp);
				if (UNLIKELY(n != 1)) {
					PyErr_SetString(PyExc_RuntimeError, "readImportData_2_1_16: Failed to read/import all-to-all timing.");
					return -1;
				}
				sion_swap(ptr, ptr, sizeof(double), 1, swap);
				#ifdef DEBUG_OUTPUT
					printf("all-to-all time[%d,%zu]       =%f\n",rank,i,((double*)ptr)[0]);fflush(stdout);
				#endif
			}
			/* Read Serial Retest Data */
			if (rank==0){
				if (linkTestInfo->iInfo[4]>0){
					/* Read New Serially Retested Timings */
					ptr=PyArray_GETPTR1((PyArrayObject *)pySerialTimingsNew[i], 0);
					n=fread(ptr, sizeof(double  ), linkTestInfo->iInfo[4], fh->fp);
					if (UNLIKELY((uint64_t)n != linkTestInfo->iInfo[4])) {
						PyErr_SetString(PyExc_RuntimeError, "readImportData_2_1_16: Failed to read/import retested serial timings.");
						return -1;
					}
					sion_swap(ptr, ptr, sizeof(double  ), linkTestInfo->iInfo[4], swap);
					#ifdef DEBUG_OUTPUT
						printf("serialTimingsNew[0,%zu]       =%f",i,((double*)ptr)[0]);
						for(n=1;(uint64_t)n<linkTestInfo->iInfo[4];n++) printf(",%f",((double*)ptr)[n]);
						printf("\n");
						fflush(stdout);
					#endif
					/* Read Old Serially Retested Timings */
					ptr=PyArray_GETPTR1((PyArrayObject *)pySerialTimingsOld[i], 0);
					n=fread(ptr, sizeof(double  ), linkTestInfo->iInfo[4], fh->fp);
					if (UNLIKELY((uint64_t)n != linkTestInfo->iInfo[4])) {
						PyErr_SetString(PyExc_RuntimeError, "readImportData_2_1_16: Failed to read/import retest-needed timings.");
						return -1;
					}
					sion_swap(ptr, ptr, sizeof(double  ), linkTestInfo->iInfo[4], swap);
					#ifdef DEBUG_OUTPUT
						printf("serialTimingsOld[0,%zu]       =%f",i,((double*)ptr)[0]);
						for(n=1;(uint64_t)n<linkTestInfo->iInfo[4];n++) printf(",%f",((double*)ptr)[n]);
						printf("\n");
						fflush(stdout);
					#endif
					/* Read Origin Partner For Serial Retests */
					ptr=PyArray_GETPTR1((PyArrayObject *)pySerialFrom[i]      , 0);
					n=fread(ptr, sizeof(uint64_t), linkTestInfo->iInfo[4], fh->fp);
					if (UNLIKELY((uint64_t)n != linkTestInfo->iInfo[4])) {
						PyErr_SetString(PyExc_RuntimeError, "readImportData_2_1_16: Failed to read/import serial retest origin.");
						return -1;
					}
					#ifdef DEBUG_OUTPUT
						printf("serialTimingsFrom[0,%zu]      =%d",i,((uint64_t*)ptr)[0]);
						for(n=1;(uint64_t)n<linkTestInfo->iInfo[4];n++) printf(",%"PRIu64,((uint64_t*)ptr)[n]);
						printf("\n");
						fflush(stdout);
					#endif
					sion_swap(ptr, ptr, sizeof(uint64_t), linkTestInfo->iInfo[4], swap);
					/* Read Destination Partner For Serial Retests */
					ptr=PyArray_GETPTR1((PyArrayObject *)pySerialTo[i]        , 0);
					n=fread(ptr, sizeof(uint64_t), linkTestInfo->iInfo[4], fh->fp);
					if (UNLIKELY((uint64_t)n != linkTestInfo->iInfo[4])) {
						PyErr_SetString(PyExc_RuntimeError, "readImportData_2_1_16: Failed to read/import serial retest destination.");
						return -1;
					}
					sion_swap(ptr, ptr, sizeof(uint64_t), linkTestInfo->iInfo[4], swap);
					#ifdef DEBUG_OUTPUT
						printf("serialTimingsTo[0,%zu]        =%d",i,((uint64_t*)ptr)[0]);
						for(n=1;(uint64_t)n<linkTestInfo->iInfo[4];n++) printf(",%"PRIu64,((uint64_t*)ptr)[n]);
						printf("\n");
						fflush(stdout);
					#endif
				}
				/* Read End Time */
				char datetime_string[LINKTEST_OUTPUT_DATETIME_LENGTH_2_1_16+1]; //TODO: Dangerous if sizeof(char) mismatches between writing and reading system.
				n=fread(datetime_string, sizeof(char), LINKTEST_OUTPUT_DATETIME_LENGTH_2_1_16, fh->fp);
				if (UNLIKELY(n != LINKTEST_OUTPUT_DATETIME_LENGTH_2_1_16)) {
					PyErr_SetString(PyExc_RuntimeError, "readInfo_2_1_16: Failed to read endDatetime.");
					return -1;
				}
				sion_swap(datetime_string, datetime_string, sizeof(char), LINKTEST_OUTPUT_DATETIME_LENGTH_2_1_16, swap);
				if (UNLIKELY(datetime_string[0] == '\0')){
					PyErr_SetString(PyExc_RuntimeError, "readInfo_2_1_16: endDatetime NULL is not supported.");
					return -1;
				}
				datetime_string[LINKTEST_OUTPUT_DATETIME_LENGTH_2_1_16]='\0'; //Ensure string is null terminated
				#ifdef DEBUG_OUTPUT
					printf("endDatetime[0,%zu]            =%s",i,datetime_string);
					fflush(stdout);
				#endif
				struct Datetime endDatetime;
				n=sscanf(datetime_string,
						"%2"SCNu8"/%2"SCNu8"/%"SCNu64" %2"SCNu8":%2"SCNu8":%2"SCNu8" UTC+0",
						&endDatetime.day,
						&endDatetime.month,
						&endDatetime.year,
						&endDatetime.hour,
						&endDatetime.minute,
						&endDatetime.second
						);
				if (UNLIKELY(n!=6)){
					PyErr_SetString(PyExc_RuntimeError, "readInfo_2_1_16: endDatetime is not a correctly formatted string.");
					return -1;
				}
				#ifdef DEBUG_OUTPUT
					printf("endDatetime.year[0,%zu]      =%" PRIu64 "\n",i,endDatetime.year  );
					printf("endDatetime.month[0,%zu]     =%" PRIu8  "\n",i,endDatetime.month );
					printf("endDatetime.day[0,%zu]       =%" PRIu8  "\n",i,endDatetime.day   );
					printf("endDatetime.hour[0,%zu]      =%" PRIu8  "\n",i,endDatetime.hour  );
					printf("endDatetime.minute[0,%zu]    =%" PRIu8  "\n",i,endDatetime.minute);
					printf("endDatetime.second[0,%zu]    =%" PRIu8  "\n",i,endDatetime.second);
					printf("endDatetime.timezone[0,%zu]  =UTC+0\n"     ,i                    );
					fflush(stdout);
				#endif
				/* Import End Time */
				err=PyDict_SetItem(pyEndTime[i],PyUnicode_FromString("Year"    ),PyLong_FromLong((long int)endDatetime.year  ));
				if(err){//TODO: Add iteration index to error message
					PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add DateTime_year to pyEndTime dictionary.");
					return -1;
				}
				err=PyDict_SetItem(pyEndTime[i],PyUnicode_FromString("Month"   ),PyLong_FromLong((long int)endDatetime.month ));
				if(err){
					PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add DateTime_month to pyEndTime dictionary.");
					return -1;
				}
				err=PyDict_SetItem(pyEndTime[i],PyUnicode_FromString("Day"     ),PyLong_FromLong((long int)endDatetime.day   ));
				if(err){
					PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add DateTime_day to pyEndTime dictionary.");
					return -1;
				}
				err=PyDict_SetItem(pyEndTime[i],PyUnicode_FromString("Hour"    ),PyLong_FromLong((long int)endDatetime.hour  ));
				if(err){
					PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add DateTime_hour to pyEndTime dictionary.");
					return -1;
				}
				err=PyDict_SetItem(pyEndTime[i],PyUnicode_FromString("Minute"  ),PyLong_FromLong((long int)endDatetime.minute));
				if(err){
					PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add DateTime_minute to pyEndTime dictionary.");
					return -1;
				}
				err=PyDict_SetItem(pyEndTime[i],PyUnicode_FromString("Second"  ),PyLong_FromLong((long int)endDatetime.second));
				if(err){
					PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add DateTime_second to pyEndTime dictionary.");
					return -1;
				}
				err=PyDict_SetItem(pyEndTime[i],PyUnicode_FromString("Timezone"),PyUnicode_FromFormat("UTC+0"));
				if(err){
					PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add DateTime_timezone to pyEndTime dictionary.");
					return -1;
				}
			}
		}

		/* Check Alignment by Testing for END_BLOCK */
		n=fread(end_block,sizeof(char),end_block_len,fh->fp);
		if (UNLIKELY(n != end_block_len)){
			PyErr_SetString(PyExc_RuntimeError, "readImportData_2_1_16: Failed to read END_BLOCK string.");
			return -1;
		}
		if(UNLIKELY(end_block[end_block_len-1]!='\0')){
			PyErr_SetString(PyExc_RuntimeError, "readImportData_2_1_16: Alignment Error Detected. \"END_BLOCK\" is not null terminated.");
			return -1;
		}
		if(UNLIKELY(strcmp(end_block,END_BLOCK_2_1_16))){
			PyErr_SetString(PyExc_RuntimeError, "readImportData_2_1_16: Alignment Error Detected. \"END_BLOCK\" not found at the correct location.");
			return -1;
		}
	}

	/* Close SION File */
	sion_close(fh->sid);

	/* Add Hostname To LinkTestInfo Dictionary */
	err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Hosts"),pyHosts);
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add Hosts to LinkTestInfo dictionary.");
		return -1;
	}

	/* Add Core ID To LinkTestInfo Dictionary */
	err=PyDict_SetItem(*pyLinkTestInfo,PyUnicode_FromString("Core ID"),pyCore);
	if(err){
		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to add Core ID to LinkTestInfo dictionary.");
		return -1;
	}

	/* Create Data Dictionaries */
	PyObject** const pyLinkTestDataDicts=(PyObject**)malloc(num_iterations*sizeof(PyObject));
	for(i=0;i<num_iterations;i++) pyLinkTestDataDicts[i] = PyDict_New();

	/* Fill Dictionaries */
	for(i=0;i<num_iterations;i++){
		err=PyDict_SetItem(pyLinkTestDataDicts[i], PyUnicode_FromString("Start Time"    ), pyStartTime[i]    );
		if(err){
			PyErr_SetString(PyExc_RuntimeError,"readImportData_2_1_16: Failed to add start time to LinkTestData dictionary.");
			return -1;
		}
		/* Set Double-Precision Floating-Point Information */
		err=PyDict_SetItem(pyLinkTestDataDicts[i],PyUnicode_FromString("Minimum time"                  ),PyFloat_FromDouble(minMaxAvgTimings[i*numMinMaxAvgTimings+0]));
		if(err){
			PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_2: Failed to add minTimeAll to LinkTestData dictionary.");
			return -1;
		}
		err=PyDict_SetItem(pyLinkTestDataDicts[i],PyUnicode_FromString("Maximum time"                  ),PyFloat_FromDouble(minMaxAvgTimings[i*numMinMaxAvgTimings+1]));
		if(err){
			PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_2: Failed to add maxTimeAll to LinkTestData dictionary.");
			return -1;
		}
		err=PyDict_SetItem(pyLinkTestDataDicts[i],PyUnicode_FromString("Average time"                  ),PyFloat_FromDouble(minMaxAvgTimings[i*numMinMaxAvgTimings+2]));
		if(err){
			PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_2: Failed to add avgTimeAll to LinkTestData dictionary.");
			return -1;
		}
		if(numMinMaxAvgTimings==6){ //Only execute if all-to-all testing was performed
			err=PyDict_SetItem(pyLinkTestDataDicts[i],PyUnicode_FromString("Minimum all-to-all time"),PyFloat_FromDouble(minMaxAvgTimings[i*numMinMaxAvgTimings+3]));
			if(err){
				PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_2: Failed to add a2aMinTimeAll to LinkTestData dictionary.");
				return -1;
			}
			err=PyDict_SetItem(pyLinkTestDataDicts[i],PyUnicode_FromString("Maximum all-to-all time"),PyFloat_FromDouble(minMaxAvgTimings[i*numMinMaxAvgTimings+4]));
			if(err){
				PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_2: Failed to add a2aMaxTimeAll to LinkTestData dictionary.");
				return -1;
			}
			err=PyDict_SetItem(pyLinkTestDataDicts[i],PyUnicode_FromString("Average all-to-all time"),PyFloat_FromDouble(minMaxAvgTimings[i*numMinMaxAvgTimings+5]));
			if(err){
				PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_2: Failed to add a2aAvgTimeAll to LinkTestData dictionary.");
				return -1;
			}
		}

		err=PyDict_SetItem(pyLinkTestDataDicts[i], PyUnicode_FromString("Timings"       ), pyTimings[i]      );
		if(err){
			PyErr_SetString(PyExc_RuntimeError,"readImportData_2_1_16: Failed to add timings to LinkTestData dictionary.");
			return -1;
		}
		err=PyDict_SetItem(pyLinkTestDataDicts[i], PyUnicode_FromString("Access Pattern"), pyAccesspattern[i]);
		if(err){
			PyErr_SetString(PyExc_RuntimeError,"readImportData_2_1_16: Failed to add accessPatterns to LinkTestData dictionary.");
			return -1;
		}
		if(linkTestInfo->iInfo[4]>0){
			err=PyDict_SetItem(pyLinkTestDataDicts[i], PyUnicode_FromString("New serially retested timings"  ), pySerialTimingsNew[i]);
			if(err){
				PyErr_SetString(PyExc_RuntimeError,"readImportData_2_1_16: Failed to add serialTimingsNew to LinkTestData dictionary.");
				return -1;
			}
			err=PyDict_SetItem(pyLinkTestDataDicts[i], PyUnicode_FromString("Old serially retested timings"  ), pySerialTimingsOld[i]);
			if(err){
				PyErr_SetString(PyExc_RuntimeError,"readImportData_2_1_16: Failed to add serialTimingsOld to LinkTestData dictionary.");
				return -1;
			}
			err=PyDict_SetItem(pyLinkTestDataDicts[i], PyUnicode_FromString("Serial-retest origin hosts"     ), pySerialFrom[i]);
			if(err){
				PyErr_SetString(PyExc_RuntimeError,"readImportData_2_1_16: Failed to add serialFrom to LinkTestData dictionary.");
				return -1;
			}
			err=PyDict_SetItem(pyLinkTestDataDicts[i], PyUnicode_FromString("Serial-retest destination hosts"), pySerialTo[i]);
			if(err){
				PyErr_SetString(PyExc_RuntimeError,"readImportData_2_1_16: Failed to add serialTo to LinkTestData dictionary.");
				return -1;
			}
		}
		if(linkTestInfo->options[0]){
			err=PyDict_SetItem(pyLinkTestDataDicts[i], PyUnicode_FromString("All-to-all timings"), pyA2ATimings[i]);
			if(err){
				PyErr_SetString(PyExc_RuntimeError,"readImportData_2_1_16: Failed to add a2aTimings to LinkTestData dictionary.");
				return -1;
			}
		}
		err=PyDict_SetItem(pyLinkTestDataDicts[i], PyUnicode_FromString("End Time"      ), pyEndTime[i]      );
		if(err){
			PyErr_SetString(PyExc_RuntimeError,"readImportData_2_1_16: Failed to add end time to LinkTestData dictionary.");
			return -1;
		}
	}

	/* Create Python List If Necessary */
	if(num_iterations>1){
		*pyLinkTestData=PyList_New(num_iterations);
		for(i=0;i<num_iterations;i++)PyList_SET_ITEM(*pyLinkTestData,i,pyLinkTestDataDicts[i]);
	}else{
		*pyLinkTestData=pyLinkTestDataDicts[0];
	}

	/* Free Memory */
	free(pyTimings      );
	free(pyAccesspattern);
	if(pyA2ATimings!=NULL) free(pyA2ATimings);
	if(linkTestInfo->iInfo[4]>0){
		free(pySerialTimingsNew);
		free(pySerialTimingsOld);
		free(pySerialFrom      );
		free(pySerialTo        );
	}

	return 0;
}

/**********************/
/* Read & Import Data */
/**********************/
int readImport_2_1_16(struct SionFile* const fh, const int swap, const Version version, PyObject** const pyLinkTestInfo, PyObject** const pyLinkTestData){
	struct LinkTestInfo_2_1_16 linkTestInfo;
	int err;
	/* Read LinkTest Information */
	err = readInfo_2_1_16(fh, swap, version, &linkTestInfo);
	if(err){
//		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to read LinkTest information.");
		return err;
	}
	/* Import LinkTest Information Into Python */
	*pyLinkTestInfo = PyDict_New(); //Create LinkTest information dictionary
	err = importInfo_2_1_16(version, &linkTestInfo, pyLinkTestInfo);
	if(err){
//		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to populate the LinkTestInfo dictionary.");
		//TODO: Safely Deallocate Python Dictionary
		return err;
	}
	/* Read & Import LinkTest Data Into Python */
	err = readImportData_2_1_16(fh, swap, &linkTestInfo, pyLinkTestInfo, pyLinkTestData);
	if(err){
//		PyErr_SetString(PyExc_RuntimeError,"importInfo_2_1_16: Failed to read LinkTest data or to populate the LinkTestData dictionary.");
		//TODO: Safely Deallocate Python Dictionary
		return err;
	}

	return(0);
}
