/****************************************************************************
**  LinkTest                                                               **
*****************************************************************************
**  Copyright (c) 2008-2022                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#include <linktest_DoNotIncludeInMain.h>
#include <linktest.h>

/**************************/
/* SION-Related Functions */
/**************************/
int openSionFile(char* const path, struct SionFile* const fh, int* const swap){
	int nfiles;
	int32_t fsblksz;
	long long int* chunksz     = NULL;
	int*           globalranks = NULL;

	/* Open SION File */
	fh->sid = sion_open(path, "br,ansi", &fh->size, &nfiles, &chunksz, &fsblksz, &globalranks, &fh->fp);
	if (UNLIKELY(-1 == fh->sid)) {
		PyErr_SetString(PyExc_RuntimeError, "sion_open() failed");
		return -1;
	}

	/* Compare System Endianess to File Endianess */
	*swap=(sion_get_file_endianness(fh->sid)!=sion_get_endianness());

	free(chunksz);
	free(globalranks);

	return 0;
}

/****************************/
/* Python-Related Functions */
/****************************/
PyObject *NewNumericPyArray(const int32_t type,const int32_t dim,const npy_intp* const len){
	return PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrNewFromType(type), dim, len, NULL, NULL, 0, NULL);
}


/****************************/
/* Error-Handling Functions */
/****************************/
inline int LinkTestIdentifierReadFailure(){
	PyErr_SetString(PyExc_RuntimeError, "Failed to read identifier string from file");
	return -1;
}
inline int LinkTestIdentifierMismatchError(const char* const STR, const char* const str){
	PyErr_SetObject(PyExc_RuntimeError,PyUnicode_FromFormat("Identifier string mismatch: Expected \"%s\" but got \"%s\"", STR, str));
	return -1;
}

/******************************************************/
/* LinkTest Version-Independent SION-Header Functions */
/******************************************************/
int checkLinkTestIdentifier(struct SionFile* const fh, const int swap){
	/* checkLinkTestIdentifier(struct SionFile *fh,  struct LinkTestStats *stats)
	 * Checks the LinkTest identifying string at the beginning of a file.
	 * This string should either be "LKTST" (prior to version 2.1.17) or
	 * "LinkTest" (post version 2.1.17).
	 *
	 * INPUTS:
	 *        struct SionFile     *fh:           SION file structure.
	 *        const int           swap:          If non-zero indicates endianess needs to be swapped
	 *
	 * Output: Integer boolean, if non-zero a read error was encountered
	 */
	const char* const LINKTEST_ID_STR_OLD="LKTST";    //Prior to version 2.1.17
	const char* const LINKTEST_ID_STR    ="LinkTest"; //Post version 2.1.17
	char str[strlen(LINKTEST_ID_STR)+1];
	size_t n, m;

	m = strlen(LINKTEST_ID_STR_OLD);
	n = fread(str, 1, m, fh->fp);
	if (UNLIKELY(n != m)) return LinkTestIdentifierReadFailure();
	sion_swap(str, str, sizeof(char), m, swap);
	str[n] = '\0';
	if (strcmp(str, LINKTEST_ID_STR_OLD)) {
		if (UNLIKELY(strncmp(str, LINKTEST_ID_STR, n))) return LinkTestIdentifierMismatchError(LINKTEST_ID_STR, str);
		const size_t o=strlen(LINKTEST_ID_STR);
		n = fread(&str[m], 1, o-m, fh->fp);
		if (UNLIKELY(n != o-m)) return LinkTestIdentifierReadFailure();
		str[o] = '\0';
		if (strcmp(str, LINKTEST_ID_STR)) return LinkTestIdentifierMismatchError(LINKTEST_ID_STR, str);
	}

	return 0;
}
int getVersion(struct SionFile* const fh, Version* version, const int swap){
	/* getVersion(struct SionFile *fh,  struct LinkTestStats *stats) reads
	 * the major & minor LinkTest version as well as the patch level.
	 *
	 * INPUTS:
	 *        struct SionFile     *fh:           SION file structure.
	 *        struct LinkTestInfo *linktestInfo: LinkTest information structure
	 *        const int           swap:          If non-zero indicates endianess needs to be swapped
	 *
	 * Output: Integer boolean, if non-zero a read error was encountered
	 */
	const size_t n=fread(*version,sizeof(int32_t), 3, fh->fp);
	if(UNLIKELY(n != 3)){
		PyErr_SetString(PyExc_RuntimeError, "getVersion: Failed to read 32-bit integer-triplet LinkTest Version.");
		return -1;
	}
	sion_swap(*version, *version, sizeof(int32_t), 3, swap);
	#ifdef DEBUG_OUTPUT
		printf("version                    =%d.%d.%d\n",(*version)[0],(*version)[1],(*version)[2]);fflush(stdout);
	#endif

	return 0;
}