#/****************************************************************************
#**  fzjlinktest                                                            **
#*****************************************************************************
#**  Copyright (c) 2008-2022                                                **
#**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
#**                                                                         **
#**  See the file COPYRIGHT in the package base directory for details       **
#****************************************************************************/

# Assumptions
# 1) A point is 1/72 of an inch (International Yard Standard)
# 2) UTF-8 is supported

# BUGS:
# 1) Colourbar extends too far in the vertical direction. This is related to bug 1 because the axis limits had to be extended [MatPlotLib bug]
#    Also related to https://github.com/matplotlib/matplotlib/pull/17182
# 2) x tick labels are rotated around corner, they should be rotated around left center for correct locations w.r.t pixels [MatPlotLib bug]
# 3) Colourspace should be CYMK for printing, but is RGBA. CYMK is not supported by MatPlotLib. [MatPlotLib issue]
# 4) Estimation of rendered text bounding box and associated size for dynamic text rescaling is wrong depending on renderer [MatPlotLib bug]

# For the Matplotlib bug that fontsize cannot be smaller than 1 see
# https://github.com/matplotlib/matplotlib/issues/9963
# Probably not going to be resolved

from os.path import isfile,basename
from sys import exit
from argparse import ArgumentParser
from importlib.metadata import metadata,version
from operator import add
from numpy import amax,amin,append,arange,argmax,argsort,argwhere,array,ceil,concatenate,empty,floor,linspace,log2,log10,nan,nansum,nanmin,nanmax,unique,zeros
from time import time
from datetime import datetime
from matplotlib.cm import get_cmap
from matplotlib.colors import Normalize
from matplotlib.figure import Figure
from matplotlib.image import AxesImage
from matplotlib.patches import Rectangle
from matplotlib.transforms import Bbox,ScaledTranslation,TransformedBbox
from matplotlib.backends.backend_pdf import PdfPages
from linktest import readSION

__metadata__=metadata('linktest')
__version__ =__metadata__['version']
__git_hash__=__metadata__.values()[19][12:] #Workaround to the 19th entry (git hash) and remove the string preamble
__mpl_ver__ =[int(i) for i in version('matplotlib').split('.')]

# Global Parameters
cm_per_in               =2.54                  #International Yard Standard
figure_size_cm          =[21  ,29.7          ] #A4 Paper

# Fix for incorrect rounding of tick labels for plots
# Taken from https://github.com/matplotlib/matplotlib/issues/20964
class NotRoundedAxesImage(AxesImage):
    """AxesImage subclass which does not round image extents to fit frame. This rounding leads to improper alignment between data and image."""
    def __init__(self,axis):
        super().__init__(axis)
    
    def make_image(self,renderer,magnification=1.0,unsampled=False):
        x1,x2,y1,y2=self.get_extent()
        bbox=Bbox(array([[x1,y1],[x2,y2]]))
        return self._make_image(self._A,bbox,TransformedBbox(bbox,self.get_transform()),((self.get_clip_box() or self.axes.bbox) if self.get_clip_on() else self.figure.bbox),magnification,unsampled=unsampled,round_to_pixel_border=False)

def fixedPixelImage(axis,data,cmap,vmin,vmax):
    image=NotRoundedAxesImage(axis)                                               # Create image object
    image.set(cmap=cmap,data=data,norm=Normalize(vmin=vmin,vmax=vmax,clip=False)) # Add data to image object
    axis.add_image(image)                                                         # Add image object to axis object
    axis.set_ylim(axis.get_ylim()[::-1])                                          # Invert y-axis
    return image

def fixedPixelImage_shiftLabels(figure,axis):
    label_offset=ScaledTranslation( 0.000,0.025,figure.dpi_scale_trans)
    for label in axis.xaxis.get_majorticklabels():
        label.set_transform(label.get_transform()+label_offset)
    label_offset=ScaledTranslation(-0.025,0.000,figure.dpi_scale_trans)
    for label in axis.yaxis.get_majorticklabels():
        label.set_transform(label.get_transform()+label_offset)

def find_renderer(fig):
    """
    Get renderer from figure, even when backend does not have get_renderer method
    Source: https://stackoverflow.com/a/22689498/3142385
    """

    if hasattr(fig.canvas, "get_renderer"):
        #Some backends, such as TkAgg, have the get_renderer method, which 
        #makes this easy.
        renderer = fig.canvas.get_renderer()
    else:
        #Other backends do not have the get_renderer method, so we have a work 
        #around to find the renderer.  Print the figure to a temporary file 
        #object, and then grab the renderer that was used.
        #(I stole this trick from the matplotlib backend_bases.py 
        #print_figure() method.)
        import io
        fig.canvas.print_figure(io.BytesIO())
        renderer = fig._cachedRenderer
    return(renderer)

def max_bbox_extend_1pt(strings,fast):
    '''Determines maximum bounding box extent using a 1pt
       fontsize which each string individually would fit into.

    BUG: Size Estimation Not Exact (Renderer Dependent)

    Inputs:
        string (string): Input strings to maximum textbox extent
                         Can be a list or array as well.
        fast (boolean):  Scalar that if evaluates to true causes
                         a faster approximate algorithm to be used.

    Output:
        2-element list (double): Maximum textbox extent
    '''
    cm_per_in=2.54                                                                                  #International Yard Standard
    figure=Figure()                                                                                 #Automatically closed by garbage collection at function end
    figure.set_size_inches(1/cm_per_in,1/cm_per_in)                                                 #1-by-1 cm figure size
    renderer=find_renderer(figure)                                                                  #Determine Renderer

    if fast:
        # Get maximum text height
        text=figure.text(x=0,y=0,s="".join(set("".join(strings))),fontsize=1)                       #Finds unique chars and turns into byte string
        max_text_height_cm=text.get_window_extent(renderer=renderer).bounds[3]*cm_per_in/figure.dpi #Incorrectly applies kerning, etc., but is a lot faster

        # Get maximum text width
        text.set_text(strings[argmax(len(strings))])                           #Set text string to longest string
                                                                                                    #This may not be the actually longest printed string,
                                                                                                    #but it should be good enough.
        max_text_width_cm=text.get_window_extent(renderer=renderer).bounds[2]*cm_per_in/figure.dpi  #Does not ignore kerning
    else:
        # Plot all strings to figure out maximum extent
        text=figure.text(x=0,y=0,s="",fontsize=1)
        def get_cm_extent(string):
            text.set_text(string)
            return([size*cm_per_in/figure.dpi for size in text.get_window_extent(renderer=renderer).bounds[2:4]])
        [max_text_width_cm,max_text_height_cm]=amax([get_cm_extent(string) for string in strings],axis=0)
    return([max_text_width_cm,max_text_height_cm])

def auto_fit_fontsize(width_cm_1p,height_cm_1pt,width_cm,height_cm):
    '''Estimates eligible font size based on 1pt text box extend

    Inputs:
        width_cm_1p (float): 1pt textbox width in cm
        height_cm_1pt (float): 1pt textbox height in cm
        width (float): Allowed width in cm
        height (float): Allowed height in cm

    Output:
        fontsize (float): Determined optimal fontsize
    '''

    if not hasattr(auto_fit_fontsize,"warn_fontsize"):
        auto_fit_fontsize.warn_fontsize=1

    #Determine Fontsize
    font_size=floor(amin([width_cm/width_cm_1p,height_cm/height_cm_1pt]))
    if font_size<1.0:
        if auto_fit_fontsize.warn_fontsize==1:
            print("WARNING: Attempted to resize font below font size 1. This is unsupported.\nWARNING: Setting font size to 1.")
            auto_fit_fontsize.warn_fontsize=0
        font_size=1.0
    return(font_size)

def SI_Prefix(num):
    '''Formats a number into a string such that after the number there is a space followed the relevant SI prefix.

    Inputs:
        num: The number to be formatted into a string

    Output:
        str: A string containing the number followed by a white space and the corresponding SI prefix
    '''
    suffixes=["y","z","a","f","p","n","µ","m"," ","k","M","G","T","P","E","Z","Y"]
    if num==0:
        return "0.0000 "
    m=floor(log10(num)/3)
    num/=10**(3*m)
    m=(m+8).astype(int)
    if m<0:m=0
    if num<10: #TODO: Is there not a more elegant solution for this?
        return f"{num:.4f} {suffixes[m]}"
    elif num<100:
        return f"{num:.3f} {suffixes[m]}"
    elif num<1000:
        return f"{num:.2f} {suffixes[m]}"
    else:
        return f"{num:.1f} {suffixes[m]}"

def IEC_Binary_Prefix(num,suffixes=["  ","ki","Mi","Gi","Ti","Pi","Ei","Zi","Yi"]):
    '''Formats a number into a string such that after the number there is a space followed the relevant IEC binary prefix.

    Inputs:
        num: The number to be formatted into a string

    Output:
        str: A string containing the number followed by a white space and the corresponding IEC binary prefix
    '''
    if num==0:
        return "0.0000 "
    m=sum([abs(num/1024.0**x)>=1 for x in range(1,len(suffixes))])
    num/=1024.0**m
    if num<10: #TODO: Is there not a more elegant solution for this?
        return f"{num:.4f} {suffixes[m]}"
    elif num<100:
        return f"{num:.3f} {suffixes[m]}"
    elif num<1000:
        return f"{num:.2f} {suffixes[m]}"
    else:
        return f"{num:.1f} {suffixes[m]}"

def add_coordinates(coordinate1,coordinate2):
    return(list(map(add,coordinate1,coordinate2)))

def normalize_position(position,size):
    return([position[0]/size[0],
            position[1]/size[1],
            position[2]/size[0],
            position[3]/size[1]])

def replace_slowest(timingData):
    replaced=False;
    for index,new_time in enumerate(timingData["New serially retested timings"],start=0):
        if new_time<timingData["Old serially retested timings"][index]:
            timingData["Timings"][timingData["Serial-retest origin hosts"][index]][timingData["Serial-retest destination hosts"][index]]=new_time
            replaced=True;
    return [timingData,replaced]

def get_Randomize_String(metaData):
    if "Randomize step order?" in metaData:
        if "Randomize memory-buffer content?" in metaData:
            if "Number of tests with randomized rank order" in metaData:
                if(metaData["Randomize step order?"]):
                    if(metaData["Randomize memory-buffer content?"]):
                        if(metaData["Number of tests with randomized rank order"]):
                            return "Buffers, ranks & steps"
                        else:
                            return "Buffers & steps"
                    else:
                        if(metaData["Number of tests with randomized rank order"]):
                            return "Ranks & steps"
                        else:
                            return "Steps"
                else:
                    if(metaData["Randomize memory-buffer content?"]):
                        if(metaData["Number of tests with randomized rank order"]):
                            return "Buffers & ranks"
                        else:
                            return "Buffers"
                    else:
                        if(metaData["Number of tests with randomized rank order"]):
                            return "Ranks"
                        else:
                            return "Nothing"
            else:
                if(metaData["Randomize step order?"]):
                    if(metaData["Randomize memory-buffer content?"]):
                        return "Buf.: Y, Rank: ?, Step: Y"
                    else:
                        return "Buf.: N, Rank: ?, Step: Y"
                else:
                    if(metaData["Randomize memory-buffer content?"]):
                        return "Buf.: Y, Rank: ?, Step: N"
                    else:
                        return "Buf.: N, Rank: ?, Step: N"
        else:
            if "Number of tests with randomized rank order" in metaData:
                if(metaData["Randomize step order?"]):
                    if(metaData["Number of tests with randomized rank order"]):
                        return "Buf.: ?, Rank: Y, Step: Y"
                    else:
                        return "Buf.: ?, Rank: N, Step: Y"
                else:
                    if(metaData["Number of tests with randomized rank order"]):
                        return "Buf.: ?, Rank: Y, Step: N"
                    else:
                        return "Buf.: ?, Rank: N, Step: N"
            else:
                if(metaData["Randomize step order?"]):
                    return "Buf.: ?, Rank: ?, Step: Y"
                else:
                    return "Buf.: ?, Rank: ?, Step: N"
    else:
        if "Randomize memory-buffer content?" in metaData:
            if "Number of tests with randomized rank order" in metaData:
                if(metaData["Randomize memory-buffer content?"]):
                    if(metaData["Number of tests with randomized rank order"]):
                        return "Buf.: Y, Rank: Y, Step: ?"
                    else:
                        return "Buf.: Y, Rank: N, Step: ?"
                else:
                    if(metaData["Number of tests with randomized rank order"]):
                        return "Buf.: N, Rank: Y, Step: ?"
                    else:
                        return "Buf.: N, Rank: N, Step: ?"
            else:
                if(metaData["Randomize memory-buffer content?"]):
                    return "Buf.: Y, Rank: ?, Step: ?"
                else:
                    return "Buf.: N, Rank: ?, Step: ?"
        else:
            if "Number of tests with randomized rank order" in metaData:
                if(metaData["Number of tests with randomized rank order"]):
                    return "Buf.: ?, Rank: Y, Step: ?"
                else:
                    return "Buf.: ?, Rank: N, Step: ?"
            else:
                return "Buf.: ?, Rank: ?, Step: ?"

def determine_Cutoffs(data,cutoffs,clip_range):
    minMaxTimings=[nanmin(data),nanmax(data)]
    timings_clipped=False
    if cutoffs[0] is not None and cutoffs[1] is not None and cutoff[1]<cutoffs[0]:
         print("Invalid clip range. Clip-range minimum greater than clip-range maximum. Ignoring specified clip range.")
         cutoffs=[None, None]
    if cutoffs[0] is not None and cutoffs[0]>minMaxTimings[0]:
        minMaxTimings[0]=cutoffs[0]
        timingData["Timings"][timingData["Timings"]<cutoffs[0]]=cutoffs[0]
        timings_clipped=True
    if cutoffs[1] is not None and cutoffs[1]<minMaxTimings[1]:
        minMaxTimings[1]=cutoffs[1]
        timingData["Timings"][timingData["Timings"]>cutoffs[1]]=cutoffs[1]
        timings_clipped=True
    minMaxPlotTimings=list(minMaxTimings)
    if clip_range[0] is not None and clip_range[1] is not None and clip_range[1]<clip_range[0]:
        print("Invalid plot range. Plot-range minimum greater than plot-range maximum. Ignoring specified plot range.")
        clip_range=[None, None]
    if clip_range[0] is not None:
        if clip_range[0]>minMaxTimings[1]:
            print("Data lies outside specified plotting range. Ignoring specified plotting range.")
        else:
            minMaxPlotTimings[0]=clip_range[0]
            timings_clipped=True
    if clip_range[1] is not None:
        if clip_range[1]<minMaxTimings[0]:
            print("Data lies outside specified plotting range. Ignoring specified plotting range.")
        else:
            minMaxPlotTimings[1]=clip_range[1]
            timings_clipped=True
    return cutoffs,minMaxTimings,minMaxPlotTimings,timings_clipped

def aggregateAccordingToHostname(metaData, timingData, aggregate_bandwidth_according_to_hostname):
    # TODO: Store the idx from one call to the next as for
    #       a given SION file they will not change.
    Hosts=array([x.decode('UTF-8')[:-1] for x in metaData["Hosts"]])
    tmp, ind=unique(Hosts,return_index=True)
    ind=argsort(ind)
    uHosts=tmp[ind]
    del(tmp)
    metaData["Hosts"]=array([(x+'!').encode('UTF-8') for x in uHosts]) #Add ! to end of string so that it can be removed by other functions later
    del(ind)
    idx=[argwhere(Hosts==u) for u in uHosts]
    del(Hosts)
    timings=zeros([len(idx),len(idx)])
    if aggregate_bandwidth_according_to_hostname:
        for i in range(len(idx)):
            for j in range(len(idx)):
                if(i==j):
                    timings[i][j]=nan
                    continue;
                for k in range(len(idx[i])):
                    for l in range(len(idx[j])):
                        timings[i][j]+=1/timingData["Timings"][idx[i][k],idx[j][l]]
                timings[i][j]=len(idx[i])/timings[i][j]
    else:
        for i in range(len(idx)):
            for j in range(len(idx)):
                if(i==j):
                    timings[i][j]=nan
                    continue;
                for k in range(len(idx[i])):
                    for l in range(len(idx[j])):
                        timings[i][j]+=timingData["Timings"][idx[i][k],idx[j][l]]
                timings[i][j]/=len(idx[i])*len(idx[j])
    timingData["Timings"]=timings #Bandwidth data still encoded as timing

def include_Page_Number(pageNumberInfo,figure):
    page_number_position_cm=[10.5, 0.7, 0.0, 0.0] #Anchor: Baseline Center
    page_number_font_size  =7
    if isinstance(pageNumberInfo, list):
        if pageNumberInfo[0]!=0:
            text_position_norm=normalize_position(page_number_position_cm,figure_size_cm)
            if len(pageNumberInfo)==2:
                figure.text(x=text_position_norm[0],y=text_position_norm[1],s=f"{pageNumberInfo[0]} of {pageNumberInfo[1]}",fontsize=page_number_font_size,horizontalalignment="center",verticalalignment='baseline')
            else:
                figure.text(x=text_position_norm[0],y=text_position_norm[1],s=f"{pageNumberInfo[0]}",fontsize=page_number_font_size,horizontalalignment="center",verticalalignment='baseline')

def timing___report(pdf,
                    metaData,
                    timingData,
                    title_string,
                    cutoffs,
                    plot_range,
                    domain,
                    downsampling_factor_matrix_ticks,
                    verbose,
                    matrix_colormap,
                    histogram_colormap,
                    replace_slow,
                    correct_image_dimensions,
                    pageNumberInfo,
                    aggregated=False,
                    average_according_to_hostname=False,
                    aggregate_bandwidth_according_to_hostname=False
                   ):

    # Position Data
    title_position_cm       =[ 1.5,28.5, 0.0, 0.0] #Anchor: Upper-Left  Corner
    date_position_cm        =[19.5,28.5, 0.0, 0.0] #Anchor: Upper-Right Corner
    clip_info_position_cm   =[19.5,28.1, 0.0, 0.0] #Anchor: Upper-Right Corner
    matrix_position_cm      =[ 3.0,13.4,14.0,14.0] #From Bottom-Left Corner
    histogram_position_cm   =[ 2.5, 8.4,16.0, 4.6] #From Bottom-Left Corner
    metadata_position_cm    =[ 1.9, 2.1,17.2, 4.8] #From Bottom-Left Corner
    pagenumber_position_cm  =[11.5, 0.7, 0.0, 0.0] #Anchor: Top-Edge Center
    max_matrix_ylabel_length=1.5
    title_font_size         =16
    meta_font_size          =8
    additional_font_size    =7
    hash_font_size          =5

    if verbose:
        time1=time()

    # Replace slowest timings before analysis?
    if replace_slow and metaData["Number of serial retests"]>0 and timingData["Serial-retest origin hosts"][0] != timingData["Serial-retest destination hosts"][0]:
        timingData,timings_replaced=replace_slowest(timingData)
    else:
        timings_replaced=False;

    # Aggregate According to Hostname
    if average_according_to_hostname or aggregate_bandwidth_according_to_hostname:
        mData=metaData
        metaData=metaData.copy() #We modify the metadata so we need to make a copy of it to reconstruct it later. Python can only pass by reference!
        aggregateAccordingToHostname(metaData,timingData,aggregate_bandwidth_according_to_hostname)

    # Precompute Bisection Bandwidth
    # This needs to be precomputed in case timings are clipped
    numHosts=len(metaData["Hosts"])
    if "Do a bisection test?" in metaData and metaData["Do a bisection test?"] != 0:
        bisection_bandwidth=metaData["Message size"]*nansum(1.0/timingData["Timings"])/numHosts

    # Determine Data Range & Apply Cut-Offs
    cutoffs,minMaxTimings,minMaxPlotTimings,timings_clipped=determine_Cutoffs(timingData["Timings"],cutoffs,plot_range)

    if verbose:
        time2=time();print(f"Replacing slowest connections, aggregating according to hostname, precomputing bisection bandwidth and determining cutoffs took {time2-time1:.3f} s.");time1=time2;

    # Determine Hosts
    numDigitsLastHostRank=floor(log10(numHosts)).astype(int)+1
    hosts=[f'{host[:-1].decode("UTF-8")}{"." if numHosts!=0 else ""}{rank:0{numDigitsLastHostRank}d}' for rank,host in enumerate(metaData["Hosts"],start=0)]
    if domain: #Empty strings are false
        hosts=[host.replace(domain,"") for host in hosts]

    # Determine String Y-Label Maximum String Length
    max_matrix_ylabel_height=0.94*float(downsampling_factor_matrix_ticks)*matrix_position_cm[3]/numHosts #0.94 since 6% space minimum around label text (this accounts partially for poor bounding box estimation in renderer
    bbox_extent_1pt_cm=max_bbox_extend_1pt(hosts,True);
    font_scale_factor=auto_fit_fontsize(bbox_extent_1pt_cm[0],bbox_extent_1pt_cm[1],max_matrix_ylabel_length,max_matrix_ylabel_height)
    if font_scale_factor==1.0:
        down_sample_factor=ceil((numHosts-1)*bbox_extent_1pt_cm[1]/matrix_position_cm[3])
        if (numHosts-1)%down_sample_factor!=0:
            down_sample_factor=ceil(1.5*down_sample_factor).astype(int)
        else:
            down_sample_factor=down_sample_factor.astype(int)
    else:
        down_sample_factor=int(downsampling_factor_matrix_ticks)

    # Normalize Position
    title_position_norm    =normalize_position(title_position_cm    ,figure_size_cm)
    date_position_norm     =normalize_position(date_position_cm     ,figure_size_cm)
    clip_info_position_norm=normalize_position(clip_info_position_cm,figure_size_cm)
    matrix_position_norm   =normalize_position(matrix_position_cm   ,figure_size_cm)
    histogram_position_norm=normalize_position(histogram_position_cm,figure_size_cm)
    if verbose:
        time2=time();print(f"Processing meta data and hosts took {time2-time1:.3f} s.");time1=time2;

    # Initialize Figure
    figure_size_in=[size/cm_per_in for size in figure_size_cm]
    figure=Figure()
    figure.set_size_inches(figure_size_in[0],figure_size_in[1])
    figure.patch.set_visible(False)
    renderer=find_renderer(figure)

    # Draw Title
    figure.text(x=title_position_norm[0],y=title_position_norm[1],s=title_string,fontsize=title_font_size,verticalalignment="top")

    # Draw Date
    txt="";
    if "Write timestamp: Hour" in metaData:
        if "Write timestamp: Timezone" in metaData:
            txt=f"{metaData['Write timestamp: Hour']:02d}:{metaData['Write timestamp: Minute']:02d}:{metaData['Write timestamp: Second']:02d} {metaData[ 'Write timestamp: Day']:02d}/{metaData['Write timestamp: Month']:02d}/{metaData['Write timestamp: Year']:0d} {metaData['Write timestamp: Timezone']}"
        else:
            txt=f"{metaData['Write timestamp: Hour']:02d}:{metaData['Write timestamp: Minute']:02d}:{metaData['Write timestamp: Second']:02d} {metaData[ 'Write timestamp: Day']:02d}/{metaData['Write timestamp: Month']:02d}/{metaData['Write timestamp: Year']:0d} UTC+?"
    else:
        if "Start Time" in timingData: #Present from version 2.1.16 onwards
            txt=f"{timingData['Start Time']['Hour']:02d}:{timingData['Start Time']['Minute']:02d}:{timingData['Start Time']['Second']:02d} {timingData['Start Time']['Day']:02d}/{timingData['Start Time']['Month']:02d}/{timingData['Start Time']['Year']:0d} \u2014 {timingData['End Time']['Hour']:02d}:{timingData['End Time']['Minute']:02d}:{timingData['End Time']['Second']:02d} {timingData['End Time']['Day']:02d}/{timingData['End Time']['Month']:02d}/{timingData['End Time']['Year']:0d} {timingData['End Time']['Timezone']}" #\u2014 is the unicode char for the emdash
    if txt!="":
        figure.text(x=date_position_norm[0]     ,y=date_position_norm[1]     ,s=txt,fontsize=meta_font_size,horizontalalignment="right",verticalalignment="top")

    # Draw Clipping/Replacement Info
    if timings_clipped:
        if timings_replaced:
            txt="WARNING: Replaced & Clipped"
        else:
            txt="WARNING: Figures Clipped!!!"
        figure.text(x=clip_info_position_norm[0],y=clip_info_position_norm[1],s=txt,fontsize=meta_font_size,horizontalalignment="right",verticalalignment="top",color="red")
    elif timings_replaced:
        txt="WARNING: Values Replaced!!!"
        figure.text(x=clip_info_position_norm[0],y=clip_info_position_norm[1],s=txt,fontsize=meta_font_size,horizontalalignment="right",verticalalignment="top",color="red")
    if verbose:
        time2=time();print(f"Drawing header took {time2-time1:.3f} s.");time1=time2;

    #####################################
    # Draw Task-Task (Timing) Matrix Plot #
    #####################################
    TTT_axis=figure.add_subplot(221)
    TTT_axis.set_frame_on(False)
    TTT_axis.xaxis.tick_top()
    TTT_axis.set_position(matrix_position_norm)
    if correct_image_dimensions:
        TTT_Image=fixedPixelImage(axis=TTT_axis,data=timingData["Timings"],cmap=get_cmap(matrix_colormap,2048),vmin=minMaxPlotTimings[0],vmax=minMaxPlotTimings[1])
    else:
        TTT_Image=TTT_axis.imshow(timingData["Timings"],interpolation="none",rasterized=True,clip_on=False,cmap=get_cmap(matrix_colormap,2048),vmin=minMaxPlotTimings[0],vmax=minMaxPlotTimings[1])
    ticks=arange(0,numHosts,1)
    TTT_axis.set_xticks(append(ticks[0:-1:down_sample_factor],[ticks[-1]]))
    TTT_axis.set_xticklabels([f"{rank:0{numDigitsLastHostRank}d}" for rank in list(range(len(hosts))[0:-1:down_sample_factor])+[len(hosts)-1]],rotation=90,fontsize=font_scale_factor)
    TTT_axis.set_yticks(append(ticks[0:-1:down_sample_factor],ticks[-1]))
    TTT_axis.set_yticklabels(append(hosts[0:-1:down_sample_factor],hosts[-1]),fontsize=font_scale_factor)
    TTT_axis.tick_params(axis=u"both",which=u"both",length=0,pad=1)
    if correct_image_dimensions:
        fixedPixelImage_shiftLabels(figure=figure,axis=TTT_axis)
    cticks=linspace(minMaxPlotTimings[0],minMaxPlotTimings[1],num=25)
    if aggregate_bandwidth_according_to_hostname:
        clabels=[f"{IEC_Binary_Prefix(metaData['Message size']/tick)}B/s" for tick in cticks]
    else:
        clabels=[f"{SI_Prefix(tick)}s\n{IEC_Binary_Prefix(metaData['Message size']/tick)}B/s" for tick in cticks]
    colorbar_axis=figure.add_axes([TTT_axis.get_position().x1+0.01,TTT_axis.get_position().y0,0.02,TTT_axis.get_position().height])
    colorbar_axis.tick_params(labelsize=6)
    figure.colorbar(TTT_Image,cax=colorbar_axis,ticks=cticks) #BUG: Vertical extent too large
    colorbar_axis.set_yticklabels(clabels)
    colorbar_axis.invert_yaxis()
    if average_according_to_hostname:
        colorbar_axis.set_ylabel("Averaged times for grouping according to hostname")
    if aggregate_bandwidth_according_to_hostname:
        colorbar_axis.set_ylabel("Aggregated bandwidths for grouping according to hostname")
    if verbose:
        time2=time();print(f"Drawing indexed-image took {time2-time1:.3f} s.");time1=time2;

    #######################
    # Draw Histogram Plot #
    #######################
    histogram_axis=figure.add_subplot(212)
    histogram_axis.set_position(histogram_position_norm)
    if __mpl_ver__[0] < 3 or (__mpl_ver__[0]==3 and __mpl_ver__[1] < 3):
        histogram_axis.set_yscale("log",basey=2)
    else:
        histogram_axis.set_yscale("log",base= 2)
    histogram_axis.set_axisbelow(True)
    histogram_axis.spines["top"  ].set_visible(False)
    histogram_axis.spines["left" ].set_visible(False)
    histogram_axis.spines["right"].set_visible(False)
    histogram_axis.tick_params(axis=u"y",which=u"both",length=0,pad=1)
    # Compute histogram bin edges
    n=numHosts+1
    bins=linspace(minMaxPlotTimings[0],minMaxPlotTimings[-1],num=n)
    # Create histogram
    n,_,patches=histogram_axis.hist(concatenate(timingData["Timings"]),bins=bins)
    # Colour Histogram
    binCenter=(bins[1:]+bins[0:-1])/2
    cutoffs=[binCenter.min(),binCenter.max()]
    norms=(binCenter-cutoffs[0])/(cutoffs[1]-cutoffs[0])
    histogram_cmap=get_cmap(histogram_colormap,2048)
    for norm,patch in zip(norms,patches):
        patch.set_facecolor(histogram_cmap(norm))
    # Determine X-Tick Labels
    histogram_x_ticks=linspace(bins[0],bins[-1],num=11)
    if aggregate_bandwidth_according_to_hostname:
        histogram_x_tick_labels=[f'{IEC_Binary_Prefix(metaData["Message size"]/tick)}B/s' for tick in histogram_x_ticks]
    else:
        histogram_x_tick_labels=[f'{SI_Prefix(tick)}s\n{IEC_Binary_Prefix(metaData["Message size"]/tick)}B/s' for tick in histogram_x_ticks]
    histogram_axis.set_xlim(bins[0],bins[-1])
    histogram_axis.set_xticks(histogram_x_ticks)
    histogram_axis.set_xticklabels(histogram_x_tick_labels,fontsize=6)
    histogram_axis.invert_xaxis()
    # Determine Y-Tick Labels
    histogram_y_tick_max=ceil(log2(amax(n)))
    histogram_y_ticks=[2**i for i in range(histogram_y_tick_max.astype(int)+1)]
    histogram_axis.set_yticks(histogram_y_ticks)
    histogram_axis.set_yticklabels(histogram_y_ticks,fontsize=6)
    histogram_axis.grid(axis="y")
    if verbose:
        time2=time();print(f"Drawing histogram took {time2-time1:.3f} s.");time1=time2;

    #####################
    # Draw Metadata Box #
    #####################
    metadata_position_norm=normalize_position(add_coordinates(metadata_position_cm,[-0.45,-0.6,0.90,1.20]),figure_size_cm)
    metadata_box=Rectangle(metadata_position_norm[0:2],metadata_position_norm[2],metadata_position_norm[3],fill=False)
    figure.add_artist(metadata_box)
    if aggregated:
        figure.text(x=metadata_position_norm[0]+metadata_position_norm[2]/2,y=metadata_position_norm[1]+metadata_position_norm[3]/2,s="Aggregated!",color='0.85',fontsize=70,rotation=14,horizontalalignment='center',verticalalignment='center')

    # 1st Row
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 0.00,5.00,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s="Communication API:",fontsize=meta_font_size)
    if "Mode" in metaData:
        txt=metaData["Mode"]
    else:
        txt="???"
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 3.50,5.00,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s=txt,fontsize=meta_font_size)
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 5.90,5.00,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s="Number of Tasks:",fontsize=meta_font_size)
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 9.30,5.00,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s=f"{numHosts}",fontsize=meta_font_size)
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[11.40,5.00,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s="Number of Hosts:",fontsize=meta_font_size)
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[14.10,5.00,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s=f"{len(unique(metaData['Hosts']))}",fontsize=meta_font_size)
    # 2nd Row
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 0.00,4.60,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s="Number of Messages:",fontsize=meta_font_size)
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 3.50,4.60,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s=f"{metaData['Number of messages']}",fontsize=meta_font_size)
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 5.90,4.60,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s="Message size:",fontsize=meta_font_size)
    txt=f"{metaData['Message size']:,} B  ({IEC_Binary_Prefix(metaData['Message size'])}B)"
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 9.30,4.60,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s=txt,fontsize=meta_font_size)
    # 3rd Row
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 0.00,4.20,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s="# Warm-Up Messages:",fontsize=meta_font_size)
    if "Number of warm-up messages" in metaData:
        txt=f"{metaData['Number of warm-up messages']}"
    else:
        txt="???"
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 3.50,4.20,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s=txt,fontsize=meta_font_size)
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 5.90,4.20,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s="# Serial Retests:",fontsize=meta_font_size)
    if "Number of serial retests" in metaData:
        txt=f"{metaData['Number of serial retests']}"
    else:
        txt="???"
        metaData["Number of serial retests"]=len(timingData["Serial-retest origin hosts"]) #Fix for printing of retested connections which requires this value
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 9.30,4.20,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s=txt,fontsize=meta_font_size)
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[11.40,4.20,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s="Domain:",fontsize=meta_font_size)
    if domain:
        text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[14.10,4.20,0,0]),figure_size_cm)
        figure.text(x=text_position_norm[0],y=text_position_norm[1],s=f"{domain}",fontsize=meta_font_size)
    # 3rd Row
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 0.00,3.80,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s="Execution Order:",fontsize=meta_font_size)
    if "Test serially?" in metaData:
        if(metaData["Test serially?"]):
            txt="Serial"
        else:
            txt="Parallel"
    else:
        txt="Unknown"
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 3.50,3.80,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s=txt,fontsize=meta_font_size)
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 5.90,3.80,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s="All-to-All:",fontsize=meta_font_size)
    if "Test MPI all-to-all?" in metaData:
        if(metaData["Test MPI all-to-all?"]): #Change to doAllToAll - fix load file
            txt="Yes"
        else:
            txt="No"
    else:
        txt="???"
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 9.30,3.80,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s=txt,fontsize=meta_font_size)
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[11.40,3.80,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s="Use GPU RAM:",fontsize=meta_font_size)
    if "Store messages in GPU memory?" in metaData:
        if(metaData["Store messages in GPU memory?"]):
            txt="Yes"
        else:
            txt="No"
    else:
        txt="???"
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[14.10,3.80,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s=txt,fontsize=meta_font_size)
    # 4th Row
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 0.00,3.40,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s="Test Configuration:",fontsize=meta_font_size)
    if "Do bidirectional tests?" in metaData:
        if(metaData["Do bidirectional tests?"]):
            txt="Bidirectional"
        else:
            if "Do unidirectional tests?" in metaData:
                if(metaData["Do unidirectional tests?"]):
                    txt="Unidirectional"
                else:
                    txt="Semidirectional"
            else:
                txt="Semidirectional"
    else:
        txt="???"
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 3.50,3.40,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s=txt,fontsize=meta_font_size)
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 5.90,3.40,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s="Bisection:",fontsize=meta_font_size)
    if "Do a bisection test?" in metaData:
        if(metaData["Do a bisection test?"]):
            txt="Yes"
        else:
            txt="No"
    else:
        txt="???"
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 9.30,3.40,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s=txt,fontsize=meta_font_size)
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[11.40,3.40,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s="Randomized:",fontsize=meta_font_size)
    txt=get_Randomize_String(metaData)
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[14.10,3.40,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s=txt,fontsize=meta_font_size)
    # 5th Row
    nslow=5
    ypos=3.00
    if ("Randomize memory-buffer content?" in metaData and metaData["Randomize memory-buffer content?"]!=0) or ("Number of tests with randomized rank order" in metaData and metaData["Number of tests with randomized rank order"]!=0):
        text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 0.00,ypos,0,0]),figure_size_cm)
        figure.text(x=text_position_norm[0],y=text_position_norm[1],s="Buffer PRNG Seed:",fontsize=meta_font_size)
        if "Randomize memory-buffer content?" in metaData:
            if metaData["Randomize memory-buffer content?"]!=0:
                txt=f"{metaData['Buffer Mersenne Twister seed']}"
            else:
                txt="N/A"
        else:
            txt="N/A"
        text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 3.50,ypos,0,0]),figure_size_cm)
        figure.text(x=text_position_norm[0],y=text_position_norm[1],s=txt,fontsize=meta_font_size)
        text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 5.90,ypos,0,0]),figure_size_cm)
        figure.text(x=text_position_norm[0],y=text_position_norm[1],s="Rank PRNG Seed:",fontsize=meta_font_size)
        if "Number of tests with randomized rank order" in metaData:
            if metaData["Number of tests with randomized rank order"]!=0:
                if "Seed for rank-order randomization" in metaData:
                    if metaData["Seed for rank-order randomization"] is None:
                        txt="Indeterminate"
                    else:
                        txt=f"{metaData['Seed for rank-order randomization']}"
                elif metaData["Version"]=="2.1.15":
                    txt="1337"
                else:
                    txt="???"
            else:
                txt="N/A"
        else:
            txt="???"
        text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 9.30,ypos,0,0]),figure_size_cm)
        figure.text(x=text_position_norm[0],y=text_position_norm[1],s=txt,fontsize=meta_font_size)
        text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[11.40,ypos,0,0]),figure_size_cm)
        figure.text(x=text_position_norm[0],y=text_position_norm[1],s="Step PRNG Seed:",fontsize=meta_font_size)
        if "Randomize step order?" in metaData:
            if metaData["Randomize step order?"]!=0:
                if "Seed for step-order randomization" in metaData:
                    txt=f"{metaData['Seed for step-order randomization']}"
                else:
                    txt="???"
            else:
                txt="N/A"
        else:
            txt="???"
        text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[14.10,ypos,0,0]),figure_size_cm)
        figure.text(x=text_position_norm[0],y=text_position_norm[1],s=txt,fontsize=meta_font_size)
        nslow-=1
        ypos-=0.4
    # 6th Row
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 0.00,ypos,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s="Memory Allocator:",fontsize=meta_font_size)
    if "Allocator type" in metaData:
        txt=metaData["Allocator type"]
    elif metaData["Version"]=="2.0.0" or metaData["Version"]=="2.1.1" or metaData["Version"]=="2.1.2" or metaData["Version"]=="2.1.3" or metaData["Version"]=="2.1.4" or metaData["Version"]=="2.1.5" or metaData["Version"]=="2.1.6" or metaData["Version"]=="2.1.7" or metaData["Version"]=="2.1.8" or metaData["Version"]=="2.1.9" or metaData["Version"]=="2.1.10" or metaData["Version"]=="2.1.11" or metaData["Version"]=="2.1.12":
        txt="Pinned memory-map"
    else:
        txt="UNKNOWN"
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 3.50,ypos,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s=txt,fontsize=meta_font_size)
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[11.40,ypos,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s="# Multiple Buffers:",fontsize=meta_font_size)
    if "Use multiple memory buffers?" in metaData:
        if metaData["Use multiple memory buffers?"]:
            txt=f"{metaData['Number of multiple buffers']}"
        else:
            txt="None"
    else:
        txt="None"
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[14.10,ypos,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s=txt,fontsize=meta_font_size)
    ypos-=0.4
    # 7th & 8th Row
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 0.00,ypos,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s="Min. Value:",fontsize=meta_font_size)
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 3.50,ypos,0,0]),figure_size_cm)
    text1=figure.text(x=text_position_norm[0],y=text_position_norm[1],s=f"{SI_Prefix(timingData['Minimum time'])}s",fontsize=meta_font_size)
    text_position_norm=[text1.get_window_extent(renderer=renderer).x1/(figure.dpi*figure_size_in[0]),(metadata_position_cm[1]+ypos-0.4)/figure_size_cm[1]]
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s=f"{IEC_Binary_Prefix(metaData['Message size']/timingData['Minimum time'])}B/s",fontsize=meta_font_size,horizontalalignment="right")
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 5.90,ypos,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s="Average Value:",fontsize=meta_font_size)
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 9.30,ypos,0,0]),figure_size_cm)
    text2=figure.text(x=text_position_norm[0],y=text_position_norm[1],s=f"{SI_Prefix(timingData['Average time'])}s",fontsize=meta_font_size)
    text_position_norm=[text2.get_window_extent(renderer=renderer).x1/(figure.dpi*figure_size_in[0]),(metadata_position_cm[1]+ypos-0.4)/figure_size_cm[1]]
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s=f"{IEC_Binary_Prefix(metaData['Message size']/timingData['Average time'])}B/s",fontsize=meta_font_size,horizontalalignment="right")
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[11.40,ypos,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s="Max. Value:",fontsize=meta_font_size)
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[14.10,ypos,0,0]),figure_size_cm)
    text3=figure.text(x=text_position_norm[0],y=text_position_norm[1],s=f"{SI_Prefix(timingData['Maximum time'])}s",fontsize=meta_font_size)
    text_position_norm=[text3.get_window_extent(renderer=renderer).x1/(figure.dpi*figure_size_in[0]),(metadata_position_cm[1]+ypos-0.4)/figure_size_cm[1]]
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s=f"{IEC_Binary_Prefix(metaData['Message size']/timingData['Maximum time'])}B/s",fontsize=meta_font_size,horizontalalignment="right")
    ypos-=0.8
    # 9th Row +
    xpos=0.0
    if "Test MPI all-to-all?" in metaData and metaData["Test MPI all-to-all?"] != 0:
        n=numHosts
        text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 0.00,ypos,0,0]),figure_size_cm)
        figure.text(x=text_position_norm[0],y=text_position_norm[1],s="Min A2A Value:",fontsize=meta_font_size)
        text_position_norm=[text1.get_window_extent(renderer=renderer).x1/(figure.dpi*figure_size_in[0]),(metadata_position_cm[1]+ypos)/figure_size_cm[1]]
        figure.text(x=text_position_norm[0],y=text_position_norm[1],s=f"{SI_Prefix(timingData['Minimum all-to-all time'])}s",fontsize=meta_font_size,horizontalalignment="right")
        text_position_norm=[text1.get_window_extent(renderer=renderer).x1/(figure.dpi*figure_size_in[0]),(metadata_position_cm[1]+ypos-0.40)/figure_size_cm[1]]
        figure.text(x=text_position_norm[0],y=text_position_norm[1],s=f"{IEC_Binary_Prefix(metaData['Message size']*n*(n-1)/timingData['Minimum all-to-all time'])}B/s",fontsize=meta_font_size,horizontalalignment="right")
        text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 5.90,ypos,0,0]),figure_size_cm)
        figure.text(x=text_position_norm[0],y=text_position_norm[1],s="Average A2A Value:",fontsize=meta_font_size)
        text_position_norm=[text2.get_window_extent(renderer=renderer).x1/(figure.dpi*figure_size_in[0]),(metadata_position_cm[1]+ypos)/figure_size_cm[1]]
        text=figure.text(x=text_position_norm[0],y=text_position_norm[1],s=f"{SI_Prefix(timingData['Average all-to-all time'])}s",fontsize=meta_font_size,horizontalalignment="right")
        text_position_norm=[text2.get_window_extent(renderer=renderer).x1/(figure.dpi*figure_size_in[0]),(metadata_position_cm[1]+ypos-0.40)/figure_size_cm[1]]
        figure.text(x=text_position_norm[0],y=text_position_norm[1],s=f"{IEC_Binary_Prefix(metaData['Message size']*n*(n-1)/timingData['Average all-to-all time'])}B/s",fontsize=meta_font_size,horizontalalignment="right")
        text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[11.40,ypos,0,0]),figure_size_cm)
        figure.text(x=text_position_norm[0],y=text_position_norm[1],s="Max A2A Value:",fontsize=meta_font_size)
        text_position_norm=[text3.get_window_extent(renderer=renderer).x1/(figure.dpi*figure_size_in[0]),(metadata_position_cm[1]+ypos)/figure_size_cm[1]]
        text=figure.text(x=text_position_norm[0],y=text_position_norm[1],s=f"{SI_Prefix(timingData['Maximum all-to-all time'])}s",fontsize=meta_font_size,horizontalalignment="right")
        text_position_norm=[text3.get_window_extent(renderer=renderer).x1/(figure.dpi*figure_size_in[0]),(metadata_position_cm[1]+ypos-0.40)/figure_size_cm[1]]
        figure.text(x=text_position_norm[0],y=text_position_norm[1],s=f"{IEC_Binary_Prefix(metaData['Message size']*n*(n-1)/timingData['Maximum all-to-all time'])}B/s",fontsize=meta_font_size,horizontalalignment="right")
        nslow-=2
        ypos-=0.80
    if "Group processes according to hostname?" in metaData and metaData["Group processes according to hostname?"] != 0:
        text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 0.00,ypos,0,0]),figure_size_cm)
        figure.text(x=text_position_norm[0],y=text_position_norm[1],s="Grouping:",fontsize=meta_font_size)
        text_position_norm=[text1.get_window_extent(renderer=renderer).x1/(figure.dpi*figure_size_in[0]),(metadata_position_cm[1]+ypos)/figure_size_cm[1]]
        figure.text(x=text_position_norm[0],y=text_position_norm[1],s="Hostname",fontsize=meta_font_size,horizontalalignment="right")
        if "Do a bisection test?" in metaData and metaData["Do a bisection test?"] != 0:
            text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 5.90,ypos,0,0]),figure_size_cm)
            figure.text(x=text_position_norm[0],y=text_position_norm[1],s="Bisection Bandwidth:",fontsize=meta_font_size)
            text_position_norm=[text2.get_window_extent(renderer=renderer).x1/(figure.dpi*figure_size_in[0]),(metadata_position_cm[1]+ypos)/figure_size_cm[1]]
            figure.text(x=text_position_norm[0],y=text_position_norm[1],s=f"{IEC_Binary_Prefix(bisection_bandwidth)}B/s",fontsize=meta_font_size,horizontalalignment="right")
        nslow-=1
        ypos-=0.40
    elif "Do a bisection test?" in metaData and metaData["Do a bisection test?"] != 0:
        text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 0.00,ypos,0,0]),figure_size_cm)
        figure.text(x=text_position_norm[0],y=text_position_norm[1],s="Bisection Bandwidth:",fontsize=meta_font_size)
        text_position_norm=[text1.get_window_extent(renderer=renderer).x1/(figure.dpi*figure_size_in[0]),(metadata_position_cm[1]+ypos)/figure_size_cm[1]]
        figure.text(x=text_position_norm[0],y=text_position_norm[1],s=f"{IEC_Binary_Prefix(bisection_bandwidth)}B/s",fontsize=meta_font_size,horizontalalignment="right")
        nslow-=1
        ypos-=0.40
    elif "Test MPI all-to-all?" in metaData and metaData["Test MPI all-to-all?"] != 0:
        text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[ 0.00,ypos,0,0]),figure_size_cm)
        figure.text(x=text_position_norm[0],y=text_position_norm[1],s="A2A Speedup:",fontsize=meta_font_size)
        text_position_norm=[text1.get_window_extent(renderer=renderer).x1/(figure.dpi*figure_size_in[0]),(metadata_position_cm[1]+ypos)/figure_size_cm[1]]
        figure.text(x=text_position_norm[0],y=text_position_norm[1],s=f"{n*(n-1)*timingData['Average time']/timingData['Minimum all-to-all time']:10.3f}",fontsize=meta_font_size,horizontalalignment="right")
        nslow-=1
        ypos-=0.40
    ypos-=0.10;
    if not aggregated:
        if nslow > metaData["Number of serial retests"]:
            nslow=metaData["Number of serial retests"]
        if nslow>0 and timingData["Serial-retest origin hosts"][0] != timingData["Serial-retest destination hosts"][0]: # If false then no timing data was recorded
            if nslow==1:
                txt="Worst Connection Serially Retested:"
            else:
                txt="Worst Connections Serially Retested:"
            text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[0.00,ypos,0,0]),figure_size_cm)
            figure.text(x=text_position_norm[0],y=text_position_norm[1],s=txt,fontsize=meta_font_size)
            # Rows 8 to and including 12
            for i in range(nslow):
                text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[0.00,ypos-(i+1)*0.40,0,0]),figure_size_cm)
                figure.text(x=text_position_norm[0],
                            y=text_position_norm[1],
                            s=f"{hosts[timingData['Serial-retest origin hosts'][i]]} to {hosts[timingData['Serial-retest destination hosts'][i]]}: {SI_Prefix(timingData['Old serially retested timings'][i])}s ({IEC_Binary_Prefix(metaData['Message size']/timingData['Old serially retested timings'][i])}B/s) → {SI_Prefix(timingData['New serially retested timings'][i])}s ({IEC_Binary_Prefix(metaData['Message size']/timingData['New serially retested timings'][i])}B/s)",
                            fontsize=meta_font_size)
    if verbose:
        time2=time();print(f"Drawing text box took {time2-time1:.3f} s.");time1=time2;
    # Additional Information
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[0.00,-1.00,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s=f"LinkTest Version: {metaData['Version']}",fontsize=additional_font_size)
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[metadata_position_cm[2],-1.00,0,0]),figure_size_cm)
    text2=figure.text(x=text_position_norm[0],y=text_position_norm[1],s=f"LinkTest Python\u2013Report-Tool Version: {__version__}",fontsize=additional_font_size,horizontalalignment="right")
    if "Git hash" in metaData:
        txt=metaData["Git hash"]
    else:
        txt="UNKNOWN"
    text_position_norm=normalize_position(add_coordinates(metadata_position_cm,[0.00,-1.25,0,0]),figure_size_cm)
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s=f"LinkTest Git Hash: {txt}",fontsize=hash_font_size)
    text_position_norm=[text2.get_window_extent(renderer=renderer).x1/(figure.dpi*figure_size_in[0]),(metadata_position_cm[1]-1.25)/figure_size_cm[1]]
    figure.text(x=text_position_norm[0],y=text_position_norm[1],s=f"LinkTest Report Git Hash: {__git_hash__}",fontsize=hash_font_size,horizontalalignment="right")
    # Page Number
    include_Page_Number(pageNumberInfo,figure)
    if verbose:
        time2=time();print(f"Drawing footer took {time2-time1:.3f} s.");time1=time2;

    # Save Report
    pdf.savefig(figure,transparent=True)
    if verbose:
        time2=time();print(f"Saving timing report took {time2-time1:.3f} s.");

    if average_according_to_hostname:
        metaData=mData #Reset metaData

def pattern__report(pdf,
                    metaData,
                    timingData,
                    title_string,
                    cutoffs,
                    plot_range,
                    domain,
                    downsampling_factor_matrix_ticks,
                    verbose,
                    matrix_colormap,
                    histogram_colormap,
                    replace_slow,
                    correct_image_dimensions,
                    pageNumberInfo
                   ):

    # Position Data
    TTS_title_position_cm   =[ 1.5,28.5, 0.0, 0.0] #Anchor: Upper-Left Corner
    TTS_position_cm         =[ 4.0,15.5,12.0,12.0] #From Bottom-Left   Corner
    TST_title_position_cm   =[ 1.5,15.0, 0.0, 0.0] #Anchor: Upper-Left Corner
    TST_position_cm         =[ 4.0, 2.0,12.0,12.0] #From Bottom-Left   Corner
    max_matrix_ylabel_length=1.5
    title_font_size         =16
    meta_font_size          =8
    additional_font_size    =7
    hash_font_size          =5

    if verbose:
        time1=time()

    # Determine Data Range & Apply Cut-Offs
    cutoffs,minMaxTimings,minMaxPlotTimings,timings_clipped=determine_Cutoffs(timingData["Timings"],cutoffs,plot_range)

    # Determine Hosts
    numHosts=len(metaData["Hosts"])
    numDigitsLastHostRank=floor(log10(numHosts)).astype(int)+1
    hosts=[f'{host[:-1].decode("UTF-8")}{"." if len(host)!=0 else ""}{rank:0{numDigitsLastHostRank}d}' for rank,host in enumerate(metaData["Hosts"],start=0)]
    if domain: #Empty strings are false
        hosts=[host.replace(domain,"") for host in hosts]

    # Determine String Y-Label Maximum String Length
    max_matrix_ylabel_height=0.94*float(downsampling_factor_matrix_ticks)*TTS_position_cm[3]/numHosts #0.94 since 6% space minimum around label text (this accounts partially for poor bounding box estimation in renderer
    bbox_extent_1pt_cm=max_bbox_extend_1pt(hosts,True)
    font_scale_factor=auto_fit_fontsize(bbox_extent_1pt_cm[0],bbox_extent_1pt_cm[1],max_matrix_ylabel_length,max_matrix_ylabel_height)
    if font_scale_factor==1.0:
        down_sample_factor_y=ceil((numHosts-1)*bbox_extent_1pt_cm[1]/TTS_position_cm[3])
        if (numHosts-1)%down_sample_factor_y!=0:
            down_sample_factor_y=ceil(1.5*down_sample_factor_y).astype(int)
        else:
            down_sample_factor_y=down_sample_factor_y.astype(int)
    else:
        down_sample_factor_y=int(downsampling_factor_matrix_ticks)

    # Determine String X-Label Maximum String Size
    #################### ToDO ####################
    ##############################################
    ##############################################
    ##############################################
    ##############################################
    ##############################################
    ##############################################
    ##############################################
    ##############################################
    ##############################################
    ##############################################
    ##############################################
    ##############################################
    ##############################################
    ##############################################
    ##############################################
    ##############################################
    ##############################################

    # Normalize Position
    TTS_title_position_norm=normalize_position(TTS_title_position_cm,figure_size_cm)
    TTS_position_norm      =normalize_position(TTS_position_cm      ,figure_size_cm)
    TST_title_position_norm=normalize_position(TST_title_position_cm,figure_size_cm)
    TST_position_norm      =normalize_position(TST_position_cm      ,figure_size_cm)

    if verbose:
        time2=time();print(f"Processing meta data and hosts took {time2-time1:.3f} s.");time1=time2;

    # Initialize Figure
    figure_size_in=[size/cm_per_in for size in figure_size_cm]
    figure=Figure()
    figure.set_size_inches(figure_size_in[0],figure_size_in[1])
    figure.patch.set_visible(False)
    renderer=find_renderer(figure)

    #####################################
    # Draw Task-Task (Step) Matrix Plot #
    #####################################
    numSteps=timingData["Access Pattern"].max().astype(int)
    numDigitsLastStep=floor(log10(numSteps)).astype(int)+1
    # Draw TTS Title
    figure.text(x=TTS_title_position_norm[0],y=TTS_title_position_norm[1],s="Task-Task-Step Plot",fontsize=title_font_size,verticalalignment="top")
    # Configure Axis & Colourmap
    TTS_axis=figure.add_subplot(211)
    TTS_axis.set_frame_on(False)
    TTS_axis.xaxis.tick_top()
    TTS_axis.set_position(TTS_position_norm)
    cMap=get_cmap(matrix_colormap,2048)
    cMap.set_under("white")
    # Add colour-indexed image
    if correct_image_dimensions:
        TTS_Image=fixedPixelImage(axis=TTS_axis,data=timingData["Access Pattern"],cmap=cMap,vmin=1,vmax=numSteps)
    else:
        TTS_Image=TTS_axis.imshow(timingData["Access Pattern"],interpolation="none",rasterized=True,clip_on=False,cmap=cMap,vmin=1,vmax=numSteps)
    # Configure Ticks
    ticks=arange(0,numHosts,1)
    TTS_axis.set_xticks(append(ticks[0:-1:down_sample_factor_y],ticks[-1]))
    TTS_axis.set_xticklabels([f"{rank:0{numDigitsLastHostRank}d}" for rank in list(range(numHosts)[0:-1:down_sample_factor_y])+[numHosts-1]],rotation=90,fontsize=font_scale_factor)
    TTS_axis.set_yticks(append(ticks[0:-1:down_sample_factor_y],ticks[-1]))
    TTS_axis.set_yticklabels(append(hosts[0:-1:down_sample_factor_y],hosts[-1]),fontsize=font_scale_factor)
    TTS_axis.tick_params(axis=u"both",which=u"both",length=0,pad=1)
    if correct_image_dimensions:
        fixedPixelImage_shiftLabels(figure=figure,axis=TTS_axis)
    # Add Colourbar
    num_cticks=floor(numSteps/51).astype(int)
    cticks=list(range(1,numSteps,num_cticks if num_cticks>0 else 1))
    if (numSteps>51):
        if (cticks[-1]-cticks[-2])/(cticks[-2]-cticks[-3])>0.1:
            cticks.append(numSteps)
        else:
            cticks[-1]=numSteps;
    else:
        cticks.append(numSteps);
    colorbar_axis=figure.add_axes([TTS_axis.get_position().x1+0.01,TTS_axis.get_position().y0,0.02,TTS_axis.get_position().height])
    colorbar_axis.tick_params(labelsize=6)
    colorbar_axis.invert_yaxis()
    figure.colorbar(TTS_Image,cax=colorbar_axis,ticks=cticks,label='Step') #BUG: Vertical extent too large
    if verbose:
        time2=time();print(f"Drawing Task-Task (Step) TTS took {time2-time1:.3f} s.");time1=time2;

    #####################################
    # Draw Task-Step (Time) Matrix Plot #
    #####################################
    # Draw TTS Title
    figure.text(x=TST_title_position_norm[0],y=TST_title_position_norm[1],s="Task-Step-Timings Plot",fontsize=title_font_size,verticalalignment="top")
    # Sort Data
    TST=empty([numHosts,numSteps])
    TST[:]=nan
    for i in range(numHosts):
        for j in range(numHosts):
            if timingData["Access Pattern"][i,j].astype(int)==0:
                continue
            else:
                TST[i,timingData["Access Pattern"][i,j].astype(int)-1]=timingData["Timings"][i,j]
    # Configure Axis & Colourmap
    TST_axis=figure.add_subplot(212)
    TST_axis.set_frame_on(False)
    TST_axis.xaxis.tick_top()
    TST_axis.set_position(TST_position_norm)
    # Add colour-indexed image
    if correct_image_dimensions:
        TST_Image=fixedPixelImage(axis=TST_axis,data=TST,cmap=cMap,vmin=minMaxPlotTimings[0],vmax=minMaxPlotTimings[1])
    else:
        TST_Image=TST_axis.imshow(TST,interpolation="none",rasterized=True,clip_on=False,cmap=get_cmap(matrix_colormap,2048),vmin=minMaxPlotTimings[0],vmax=minMaxPlotTimings[1])
    # Configure Ticks
    TST_axis.set_aspect(numSteps/numHosts)
    xticks=arange(0,numSteps,1)
    yticks=arange(0,numHosts,1)
    TST_axis.set_xticks(append(xticks[0:-1:down_sample_factor_y],xticks[-1]))
    TST_axis.set_xticklabels([f"{tick:0{numDigitsLastStep}d}" for tick in list(range(1,numSteps+1,1)[0:-1:down_sample_factor_y])+[numSteps]],rotation=90,fontsize=font_scale_factor)
    TST_axis.set_yticks(append(yticks[0:-1:down_sample_factor_y],yticks[-1]))
    TST_axis.set_yticklabels(append(hosts[0:-1:down_sample_factor_y],hosts[-1]),fontsize=font_scale_factor)
    TST_axis.tick_params(axis=u"both",which=u"both",length=0,pad=1)
    # Add Colourbar
    if correct_image_dimensions:
        fixedPixelImage_shiftLabels(figure=figure,axis=TST_axis)
    cticks=linspace(minMaxPlotTimings[0],minMaxPlotTimings[1],num=25)
    clabels=[f"{SI_Prefix(tick)}s\n{IEC_Binary_Prefix(metaData['Message size']/tick)}B/s" for tick in cticks]
    colorbar_axis0=figure.add_axes([TST_axis.get_position().x1+0.01,TST_axis.get_position().y0,0.02,TST_axis.get_position().height])
    colorbar_axis0.tick_params(labelsize=6)
    figure.colorbar(TST_Image,cax=colorbar_axis0,ticks=cticks) #BUG: Vertical extent too large
    colorbar_axis0.set_yticklabels(clabels)
    colorbar_axis0.invert_yaxis()
    if verbose:
        time2=time();print(f"Drawing Task-Step (Timings) took {time2-time1:.3f} s.");time1=time2;

    # Page Number
    include_Page_Number(pageNumberInfo,figure)

    # Save Report
    pdf.savefig(figure,transparent=True)
    if verbose:
        time2=time();print(f"Saving pattern report took {time2-time1:.3f} s.");

def empty____report(pdf,pageNumberInfo):
    # Initialize Figure
    figure_size_in=[size/cm_per_in for size in figure_size_cm]
    figure=Figure()
    figure.set_size_inches(figure_size_in[0],figure_size_in[1])
    figure.patch.set_visible(False)
    renderer=find_renderer(figure)
    # Page Number
    include_Page_Number(pageNumberInfo,figure)
    # Save Report
    pdf.savefig(figure,transparent=True)

def linktest_report(metaData,
                    timingData,
                    output_report_filename,
                    title_string=[None],
                    cutoffs=[None,None],
                    plot_range=[None,None],
                    domain=[""],
                    downsampling_factor_matrix_ticks=[1],
                    verbose=False,
                    matrix_colormap=["gist_rainbow_r"],
                    histogram_colormap=["gist_rainbow_r"],
                    correct_image_dimensions=False,
                    replace_slow=False,
                    two_page=False,
                    average_according_to_hostname=False,
                    aggregate_bandwidth_according_to_hostname=False
                   ):

    if verbose:
        time1=time();

    # Check version
    if metaData["Version"]!="2.0.0" and metaData["Version"]!="2.1.1" and metaData["Version"]!="2.1.2" and metaData["Version"]!="2.1.4" and metaData["Version"]!="2.1.5" and metaData["Version"]!="2.1.6" and metaData["Version"]!="2.1.7" and metaData["Version"]!="2.1.8" and metaData["Version"]!="2.1.9" and metaData["Version"]!="2.1.10" and metaData["Version"]!="2.1.11" and metaData["Version"]!="2.1.12" and metaData["Version"]!="2.1.13" and metaData["Version"]!="2.1.14" and metaData["Version"]!="2.1.15" and metaData["Version"]!="2.1.16" and metaData["Version"]!="2.1.17" and metaData["Version"]!="2.1.18" and metaData["Version"]!="2.1.19":
        print(f"Potentially unsupported LinkTest version output detected.\nThe Supported LinkTest versions are 2.0.0 and 2.1.[1,2,4-19] while this file comes from {metaData['Version']}")
    if "Do a bisection test?" in metaData and metaData["Do a bisection test?"]!=0 and (metaData["Version"]=="2.0.0" or metaData["Version"]=="2.1.1" or metaData["Version"]=="2.1.2" or metaData["Version"]=="2.1.4" or metaData["Version"]=="2.1.5" or metaData["Version"]=="2.1.6" or metaData["Version"]=="2.1.7" or metaData["Version"]=="2.1.8" or metaData["Version"]=="2.1.9"):
        print("Bisection runs were not properly executed prior to LinkTest version 2.1.10. The Report may be faulty.");

    # Generate PDF
    with PdfPages(output_report_filename) as pdf:
        if isinstance(timingData, list): #If multiple iterations TimingData is a list
            # Generate a one/double-sided page report summarizing the results
            tData=timingData[0].copy()
            tData['Timings']=sum(data['Timings'] for data in timingData)/len(timingData)
            tData['Minimum time']=min(data['Minimum time'] for data in timingData)
            tData['Average time']=sum(data['Average time'] for data in timingData)/len(timingData)
            tData['Maximum time']=max(data['Maximum time'] for data in timingData)
            tData['End Time']=timingData[-1]['End Time']
            if "Test MPI all-to-all?" in metaData and metaData["Test MPI all-to-all?"] != 0:
                tData['Minimum all-to-all time']=min(data['Minimum all-to-all time'] for data in timingData)
                tData['Average all-to-all time']=sum(data['Average all-to-all time'] for data in timingData)/len(timingData)
                tData['Maximum all-to-all time']=max(data['Maximum all-to-all time'] for data in timingData)
            pageNumberInfo=[1,len(timingData)+1];
            if two_page:
                pageNumberInfo[1]*=2
            timing___report(pdf,
                            metaData,
                            tData,
                            title_string,
                            cutoffs,
                            plot_range,
                            domain,
                            downsampling_factor_matrix_ticks,
                            verbose,
                            matrix_colormap,
                            histogram_colormap,
                            False,
                            correct_image_dimensions,
                            pageNumberInfo,
                            aggregated=True,
                            average_according_to_hostname=average_according_to_hostname,
                            aggregate_bandwidth_according_to_hostname=aggregate_bandwidth_according_to_hostname
                           );
            pageNumberInfo[0]+=1;
            if two_page:
                empty____report(pdf,pageNumberInfo);
                pageNumberInfo[0]+=1;
            # Generate a double-sided page for each iteration
            k=0;
            for tData in timingData:
                if verbose:
                    time0=time()
                title_string=f"Permutation {k+1}"
                timing___report(pdf,
                                metaData,
                                tData,
                                title_string,
                                cutoffs,
                                plot_range,
                                domain,
                                downsampling_factor_matrix_ticks,
                                verbose,
                                matrix_colormap,
                                histogram_colormap,
                                replace_slow,
                                correct_image_dimensions,
                                pageNumberInfo,
                                aggregated=False,
                                average_according_to_hostname=average_according_to_hostname,
                                aggregate_bandwidth_according_to_hostname=aggregate_bandwidth_according_to_hostname
                               );
                pageNumberInfo[0]+=1;
                if two_page:
                    pattern__report(pdf,
                                    metaData,
                                    tData,
                                    title_string,
                                    cutoffs,
                                    plot_range,
                                    domain,
                                    downsampling_factor_matrix_ticks,
                                    verbose,
                                    matrix_colormap,
                                    histogram_colormap,
                                    replace_slow,
                                    correct_image_dimensions,
                                    pageNumberInfo
                                   );
                    pageNumberInfo[0]+=1;
                if k==0: #For further permutations the seed for them cannot be necessarily established from the PRNG state
                    metaData['Seed for rank-order randomization']=None
                if verbose:
                    print(f"Generating report for iteration {k} took {time()-time0:.3f} s.");
                k+=1
        else:
            if verbose:
                time0=time()
            pageNumberInfo=[1,1];
            if two_page:
                pageNumberInfo[1]*=2
            timing___report(pdf,
                            metaData,
                            timingData,
                            title_string,
                            cutoffs,
                            plot_range,
                            domain,
                            downsampling_factor_matrix_ticks,
                            verbose,
                            matrix_colormap,
                            histogram_colormap,
                            replace_slow,
                            correct_image_dimensions,
                            pageNumberInfo,
                            aggregated=False,
                            average_according_to_hostname=average_according_to_hostname,
                            aggregate_bandwidth_according_to_hostname=aggregate_bandwidth_according_to_hostname
                           );
            if two_page:
                pageNumberInfo[0]+=1;
                pattern__report(pdf,
                                metaData,
                                timingData,
                                title_string,
                                cutoffs,
                                plot_range,
                                domain,
                                downsampling_factor_matrix_ticks,
                                verbose,
                                matrix_colormap,
                                histogram_colormap,
                                replace_slow,
                                correct_image_dimensions,
                                pageNumberInfo
                               );
            if verbose:
                    print(f"Generating report took {time()-time0:.3f} s.");
        # Set Metadata Information
        d=pdf.infodict()
        d['Title'       ] = title_string
        d['Author'      ] = f"LinkTest Python Report Tool - {__version__}, Forschungszentrum Jülich GmbH"
        d['Subject'     ] = "LinkTest Report"
        d['Keywords'    ] = "LinkTest"
        d['CreationDate'] = datetime.today()
        d['ModDate'     ] = datetime.today()

def main():
    parser=ArgumentParser(prog="linktest_report",
                          description=f"LinkTest Python Report Tool Version {__version__}.\n\nThis function generates a MatPlotLib based report on A4 paper which includes an indexed-colour image of the LinkTest timing data, a histogram of said data, a textual overview of the metadata and the serially retested timings for at most the five worst connections.\n\nThis script has been tested to work with Python version 3.8.5. Other versions may work as well.",
                          epilog="Provided as is by the Forschungszentrum Jülich GmbH. Copyright 2008-2022.\nPlease see the license under which this software is provided in the file LICENSE.\n This file was enclosed with the distribution of this software."
                         )
    parser.add_argument("-i",
                        "--input",
                        nargs=1,
                        required=True,
                        help="Input SION file for analysis."
                       )
    parser.add_argument("-o",
                        "--output",
                        nargs=1,
                        required=True,
                        help="Output PDF report filename. The suffix must be \".pdf\"."
                       )
    parser.add_argument("-t",
                        "--title",
                        nargs=1,
                        required=False,
                        default=[None],
                        help="Report title, default is input filename."
                       )
    parser.add_argument("-d",
                        "--domain",
                        nargs=1,
                        required=False,
                        default=[""],
                        help="Domain string to remove from host names."
                       )
    parser.add_argument("--min-clip-time",
                        nargs=1,
                        required=False,
                        default=[None],
                        type=float,
                        help="All recorded main average LinkTest times below this value are replaced with this value before plotting. This value is in seconds."
                       )
    parser.add_argument("--max-clip-time",
                        nargs=1,
                        required=False,
                        default=[None],
                        type=float,
                        help="All recorded main average LinkTest times above this value are replaced with this value before plotting. This value is in seconds."
                       )
    parser.add_argument("--min-plot-time",
                        nargs=1,
                        required=False,
                        default=[None],
                        type=float,
                        help="Restricts plotting to the specified minimum time. This value is in seconds. Values excluded this way are missing from the histogram."
                       )
    parser.add_argument("--max-plot-time",
                        nargs=1,
                        required=False,
                        default=[None],
                        type=float,
                        help="Restricts plotting to the specified maximum time. This value is in seconds. Values excluded this way are missing from the histogram."
                       )
    parser.add_argument("--downsampling-factor",
                        nargs=1,
                        required=False,
                        default=[1],
                        help="Downsampling factor for indexed-image ticks."
                       )
    parser.add_argument("--verbose",
                        required=False,
                        action='store_true',
                        default=False,
                        help="If set prints runtime timing information."
                       )
    parser.add_argument("--colormap-indexed-image",
                        nargs=1,
                        required=False,
                        default=["gist_rainbow_r"],
                        help="Indexed-image colormap name. For options see MatPlotLib colormaps."
                       )
    parser.add_argument("--colormap-histogram",
                        nargs=1,
                        required=False,
                        default=["gist_rainbow_r"], #Formally: RdYlGn_r
                        help="Histogram colormap name. For options see MatPlotLib colormaps."
                       )
    parser.add_argument("--replace-slowest",
                        required=False,
                        action='store_true',
                        default=False,
                        help="If set replaces timings from the main test before further analysis with those from the serial retest if they are smaller."
                       )
    parser.add_argument("--correct-image-dimensions",
                        required=False,
                        action='store_true',
                        default=False,
                        help="If set generates correctly sized colour-indexed images. This is a workaround for a MatPlotLib bug."
                       )
    parser.add_argument("--two-page",
                        required=False,
                        action='store_true',
                        default=False,
                        help="If set generates a second page showing in which step certain times were measured and the times aggregated according to the step."
                       )
    parser.add_argument("--average-according-to-hostname",
                        required=False,
                        action='store_true',
                        default=False,
                        help="If set causes timings to be averaged on a per-hostname basis. The resultant timing information is removed from the plots."
                       )
    parser.add_argument("--aggregate-bandwidth-according-to-hostname",
                        required=False,
                        action='store_true',
                        default=False,
                        help="If set causes bandwidths to be aggregated on a per-hostname basis. The resultant timing information is removed from the plots."
                       )

    # Parse Input Arguments
    args=parser.parse_args()

    # Test if input filename is a string
    if not isinstance(args.input[0],str):
        print(f"ERROR: Input {args.input[0]} does not appear to be a string.")
        exit(-1)

    # Test if input file exists
    if not isfile(args.input[0]):
        print(f"ERROR: Input {args.input[0]} does not appear to point to a SION file.")
        exit(-1)

    # Test if output filename is a string
    if not isinstance(args.output[0],str):
        print(f"ERROR: Output {args.output[0]} does not appear to be a string.")
        exit(-1)

    # Test if output filename specifies a PDF
    if not args.output[0].endswith(".pdf"):
        print(f"ERROR: Output {args.output[0]} does not appear to specify a PDF file. Only PDF files are supported")
        exit(-1)

    # Test if the title string is a string
    if not isinstance(args.title[0],str):
        if args.title[0] is not None:
            print("WARNING: The title string is not a string. Using input SION filename instead.")
        args.title[0]=basename(args.input[0])

    # Begin Timing
    if args.verbose:
        time0=time()

    # Load Data
    metaData,timingData=readSION(args.input[0])
    if args.verbose:
        print(f"Loading data took {time()-time0:.3f} s.");

    if args.average_according_to_hostname and args.aggregate_bandwidth_according_to_hostname:
        print(f"ERROR: --average-according-to-hostname and --aggregate-bandwidth-according-to-hostname cannot be used together")
        exit(-1)

    # Create Report
    linktest_report(metaData=metaData,
                    timingData=timingData,
                    output_report_filename=args.output[0],
                    title_string=args.title[0],
                    domain=args.domain[0],
                    cutoffs=[args.min_clip_time[0],args.max_clip_time[0]],
                    plot_range=[args.min_plot_time[0], args.max_plot_time[0]],
                    downsampling_factor_matrix_ticks=args.downsampling_factor[0],
                    verbose=args.verbose,
                    matrix_colormap=args.colormap_indexed_image[0],
                    histogram_colormap=args.colormap_histogram[0],
                    correct_image_dimensions=args.correct_image_dimensions,
                    replace_slow=args.replace_slowest,
                    two_page=args.two_page,
                    average_according_to_hostname=args.average_according_to_hostname,
                    aggregate_bandwidth_according_to_hostname=args.aggregate_bandwidth_according_to_hostname
                   )
    if args.verbose:
        print(f"DONE! Total time required: {time()-time0:.3f} s.");
