#/****************************************************************************
#**  LinkTest                                                               **
#*****************************************************************************
#**  Copyright (c) 2008-2022                                                **
#**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
#**                                                                         **
#**  See the file COPYRIGHT in the package base directory for details       **
#****************************************************************************/
import os
import subprocess
import codecs
from setuptools import setup, Extension

# Hard-coded LinkTest version for older pip installers
link_ver="2.1.19"
# Hard-coded LinkTest git hash for older pip installers
link_hash="UNKNOWN"
link_hash_short="UNKNOWN"

def read(rel_path):
	here = os.path.abspath(os.path.dirname(__file__))
	with codecs.open(os.path.join(here, rel_path), 'r') as fp:
		return fp.read()

def get_git_revision_hash() -> str:
	try:
		return subprocess.check_output(['git', 'rev-parse', 'HEAD']).decode('ascii').strip()
	except:
		return link_hash

def get_git_revision_short_hash() -> str:
	try:
		return subprocess.check_output(['git', 'rev-parse', '--short', 'HEAD']).decode('ascii').strip()
	except:
		return link_hash_short

def get_version(rel_path):
	try:
		for line in read(rel_path).splitlines():
			if line.startswith('#define LINKTEST_VERSION_PATCHLEVEL'):
				linktest_version_patch=line.split()[2]
				return f"{linktest_version}.{linktest_version_sub}.{linktest_version_patch}"
			if line.startswith('#define LINKTEST_VERSION_SUB'):
				linktest_version_sub=line.split()[2]
				continue
			if line.startswith('#define LINKTEST_VERSION'):
				linktest_version=line.split()[2]
		else:
			raise RuntimeError("Unable to determine version string.")
	except:
		return link_ver

packages = [ "linktest" ]
sionconfig = os.environ.get("SIONCONFIG", "sionconfig")
sionlib_compile_args = subprocess.run(
	[sionconfig, "--ser", "--cflags", "--c"],
	stdout=subprocess.PIPE,
	encoding="utf-8"
).stdout.split()
sionlib_link_args = subprocess.run(
	[sionconfig, "--ser", "--libs", "--c"],
	stdout=subprocess.PIPE,
	encoding="utf-8"
).stdout.split()

ext_modules=[
	Extension("linktest/reader",
		sources = ["linktest/linktest.c",
		           "linktest/linktest_helper.c",
		           "linktest/linktest_2_0_0.c",
		           "linktest/linktest_2_1_1.c",
		           "linktest/linktest_2_1_2.c",
		           "linktest/linktest_2_1_4.c",
		           "linktest/linktest_2_1_5.c",
		           "linktest/linktest_2_1_6.c",
		           "linktest/linktest_2_1_7.c",
		           "linktest/linktest_2_1_8.c",
		           "linktest/linktest_2_1_9.c",
		           "linktest/linktest_2_1_10.c",
		           "linktest/linktest_2_1_11.c",
		           "linktest/linktest_2_1_12.c",
		           "linktest/linktest_2_1_13.c",
		           "linktest/linktest_2_1_14.c",
		           "linktest/linktest_2_1_15.c",
		           "linktest/linktest_2_1_16.c",
		           "linktest/linktest_2_1_17.c",
		           "linktest/linktest_2_1_18.c",
		           "linktest/linktest_2_1_19.c"],
		include_dirs=['linktest'],
		extra_compile_args = sionlib_compile_args, #Add `+['-DDEBUG_OUTPUT']` to turn on debug output
		extra_link_args = sionlib_link_args
	)
]

setup(
	name="LinkTest",
	version=get_version("../benchmark/config.h"),
	githash=get_git_revision_hash(),
	author="Juelich Supercomputing Centre (JSC), Forschungszentrum Juelich (FZJ)",
	author_email="sc@fz-juelich.de",
	description='A module to load LinkTest SION files and scripts to create reports.',
	keywords="LinkTest, SION",
	license="Custom, based on BSD 3. See COPYRIGHT.",
	url="https://www.fz-juelich.de/ias/jsc/EN/Expertise/Support/Software/LinkTest/_node.html",
	packages=packages,
	ext_modules=ext_modules,
	entry_points={
		"console_scripts": [
			"linktest-report=linktest.report:main"
		]
	},
	install_requires=['numpy','matplotlib'],
	classifiers=[
		'Development Status :: 5 - Production/Stable',
		'Intended Audience :: Information Technology',
		'Topic :: System :: Benchmark',
		'Topic :: System :: Monitoring',
		'Natural Language :: English',
		'License :: CUSTOM :: MODIFIED BSD',
		'Operating System :: POSIX :: Linux',
		'Programming Language :: C',
		'Programming Language :: Python',
		f'Git Hash :: {get_git_revision_hash()}'
	]
)
